:- module(pac_etc, [expand_etc/5, etc/6]).

:- use_module(pac(op)).
:- use_module(pac(reduce)).
:- use_module(pac(basic)).
:- use_module(pac(meta)).
:- use_module(pac('expand-pac')).

:- discontiguous etc/5.		% [2017/04/08]
:- discontiguous etc/6.		% [2017/04/12]

:- op(1000, xfy, &).
:- op(700, xfx, :=).
:- op(600, yfx, ..).

			/**************************
			*     some tiny helpers   *
			**************************/
%
free_variables(E, E0, Vs):- term_variables(E, U),
	term_variables(E0, V),
	subtract_memq(V, U, NewVs),
	subtract_memq(U, NewVs, Vs).

%
subtract_memq([], _, []).
subtract_memq([X|Xs], Y, Z):- memq(X, Y), !,
	subtract_memq(Xs, Y, Z).
subtract_memq([X|Xs], Y, [X|Z]):- subtract_memq(Xs, Y, Z).
%
pre_expand_arg(F, M, G, Vs, P, Q):-
	(	nonvar(F) -> expand_arg(F, M, G, P, Q),
					 term_variables(G, Vs)
	;	Q = P,
		G = F,
		Vs = [G]
	).

			/********************
			*     expand_etc    *
			********************/

%%  expand_etc(+Pac:pac, +M:atom, -G:pred, -P:list, -Q:list) is det
%
%   True if Pac is converted to a reference G to an expanded
%	predicate in P (minus Q)  at a module M.
%	Currently the following macro are expanded:
%
%	phrase	---	hybrid regular expressions in DCG phrases
%	for	---	control predicate 'for'
%	foldl	---	foldl   ( 'foldl' is builtin in SWI-Prolog)
%	foldr	---	foldr
%	repeat	---	repeat
%   iterate --- repeat action until final.
%   while	--- while statement
%	until	--- until statement
%	sed	---	unix-like sed command based on the hybrid regualar expressions.
%	flip	---	generates a variant predicates with arguments flipping.

expand_etc(E, M, G, P, Q):- E=..[K|Args],
	once(etc(K, Args, M, G, P, Q)).

%%  expand_etc(+F:term, A:list, +M:atom, -G:pred, +P:list, -Q:list) is det
%
%   True if G is unified with a fresh predicate and helper predicates
%   are created in P (minus Q) at a module M such that callng G is equivalent
%   to applying F to the argument list A.  F is reserved name only.

% etc(call)//N and etc(apply)//N were removed.

% ?- apply(pred([A,B,C]:- plus(A,B,C)), [1,2,X]).
% ?- iff(member(X, [3,1,2]),  member(X, [1,2,3])).
% ?- call(iff(member(X, [3,3,1,2])), member(X, [1,2,3])).
% ?- M1 = member(X, [3,3,1,2]), M2 = member(X, [1,2,3]), call(iff(M1, M2)).
% ?- M1 = member(X, [3,3,1,2]), M2 = member(X, [1,2,3]), iff(M1, M2).
%
etc(iff, [C0, C1], M, G, P, Q) :- !,
    	expand_arg(C0, M, D0, P, P0),
		expand_arg(C1, M, D1, P0, Q),
		G =	\+( (D0,\+D1) ; (D1, \+D0)).
etc(iff, [C], M, meta:iff(D), P, Q):-
    	expand_arg(C, M, D, P, Q).
etc(iff, [], _, meta:iff, P, P).

%
etc(unless, [C, B], M, (Not_C -> B0; true), P, Q):-
		expand_arg(\+(C), M, Not_C, P, P0),
		expand_arg(B, M, B0, P0, Q).
%
etc(when, [C, B], M, (C0 -> B0; true), P, Q):-
	expand_arg(C, M, C0, P, P0),
	expand_arg(B, M, B0, P0, Q).

		/************************
		*     eval/eval_term    *
		************************/
% Compile an expression to a goal.
etc(eval, [E, V], M, G, P, Q) :- pac:expand_exp(E, [], =, V, M, G, P, Q).
%
etc(eval_term, [E, V], M, G, P, Q) :- pac:args_inside_out(E, E0),
	pac:expand_exp(E0, [], =, V, M, G, P, Q).

			/****************
			*     phrase    *
			****************/

% ?- etc_phrase([(a,b)], user, G, P, []).
% ?- etc_phrase([w("a")], user, G, P, []).
etc(phrase, [X|As], _, G, P, P):- var(X), !, complete_args(phrase, [X|As], G).
etc(phrase, [X|As], M, G, P, Q):- !, etc_phrase([X|As], M, G, P, Q).
%
etc_phrase([X|As], M, G, P, Q):- term_variables(X, H),
	phrase_to_pred(X, M, [U, V]:- Body, P, P0),
	expand_core(pred(H, [U,V]:- Body), M, G0, P0, Q),
	complete_args(G0, As, G).

			/*************
			*     sed    *
			*************/

% ?- phrase(sed(s/ "e"/ "yyzz"), `ae`, R), basic:smash(R).
% ?- phrase(sed(sl/ "e"/ "yyzz"), `ae`, R), basic:smash(R).
% ?- phrase(sed(s/ "a" / =([])), `abacad`, R).
% ?- sed(pred([X,[X|Y], Y]), `abc`, A).
% ?- pac_word:let_sed(X, (a/"."/"x")), call(X, `abc`, Y).
% ?- pac_word:let_sed(X, (b/"."/"x")), call(X, `abc`, Y).
% ?- pac_word:let_sed(X, (s/"."/"x")), call(X, `abc`, Y).
% ?- pac_word:let_sed(X, (s/".."/[])), call(X, `abc`, Y).
% ?- pac_word:let_sed(S, "a">>"b"), call(S, `abcb`, R).
% ?- sed(a/"b"/"c", `abcb`, R), basic:smash(R).
% ?- sed(a/"b+"/"c", `abcb`, R), basic:smash(R).
% ?- sed(b/"b"/"c", `abc`, R), basic:smash(R).
% ?- pac_etc:etc(sed, ["a">>"b", `abcb`, R], user, G, L, []).
% ?- pac_etc:etc(sed, [f], user, G, L, []).

etc(sed, [S|R], _, G, P, P):- var(S), !,
	complete_args(pac_etc:sed(S), R, G).
etc(sed, [S|R], M, G, P, Q):-  pac:expand_sed(S, [F, W, A]), !,
	pac:expand_recognize_act(F, W, A, M, G0, P, Q),
	complete_args(G0, R, G).
etc(sed, [S|R], M, G, P, Q):-
	expand_arg(S, M, S0, P, Q),
	complete_args(pac_etc:sed(S0), R, G).
%
etc(d_sed, [S|R], _, G, P, P):- var(S), !,
	complete_args(pac_etc:d_sed(S), R, G).
etc(d_sed, [S|R], M, G, P, Q):-  pac:expand_sed(S, [F, W, A]), !,
	pac:expand_recognize_act(F, W, A, M, G0, P, Q),
	complete_args(G0, R, G).
etc(d_sed, [S|R], M, G, P, Q):- expand_arg(S, M, S0, P, Q),
	complete_args(pac_etc:d_sed(S0), R, G).

:- meta_predicate sed(:,?,?,?).
sed(G, X, Y) :- sed(G, Y, X, []).
%
sed(G, [U|V], X, Y) :- call(G, U, X, Z), !, sed(G, V, Z, Y).
sed(G, [U|V], [U|X], Y) :-!, sed(G, V, X, Y).
sed(G, [U|V], acc(A, [U|X]), Y) :-!, sed(G, V, acc(A, X), Y).
sed(_, acc(A), acc(A, []),  []).
sed(_, [], [], []).

%  D-list version of sed
% ?- d_sed(pred([[X|A], A, [X|Y], Y]), `abc`, A).
:- meta_predicate d_sed(:,?,?).
:- meta_predicate d_sed(:,?,?,?,?).

d_sed(G, X, Y) :- d_sed(G, Y, [], X, []).
%
d_sed(G, X, Y, A, B):- call(G, X, X0, A, A0), !,
		   d_sed(G, X0, Y, A0, B).
d_sed(G, [U|X], Y, [U|A], B):- d_sed(G, X, Y, A, B).
d_sed(G, [U|X], Y, acc(M, [U|A]), B):- d_sed(G, X, Y, acc(M, A), B).
d_sed(_, X, X, [], []).
d_sed(_, [acc(M)|X], X, acc(M,[]), []).

		/***************************************
		*     Term rewriting system (Naive)    *
		***************************************/
% ?- module(pac).
% ?- pac_etc:etc(trs, [[a=b]], user, G,  P, []),
%	maplist(assert, P), !,	call(G, a, Out).
% ?- pac_etc:etc_trs(trs, [(A,B)=(A,C) :- B=C], user,G,  P, []),
%	maplist(assert, P), call(G, (a, b), X).

etc(trs, [R|As], M, G, P, Q):-
	etc_trs(_Trs, R, M, G0, P, Q),
 	complete_args(G0, As, G).
%
etc_trs(Trs, R, M, G, P, Q):-
	(R = Vs^Rules  ->	true
	;	Vs = [],
		Rules = R
	),
	maplist(make_rewrite_pred(Basic_rule_name), Rules, Preds),
	list_to_ampersand(Preds, Amps),
	expand_core(rec(Basic_rule_name, Vs, Amps), M, G0, P, P0),
	expand_core(
		rec(Trs, Vs,
			([X, Y]:-	call(G0, X, X0), !,
						call(Trs, X0, Y))
		    &
			([X, X])
		   ),
			M, G, P0, Q).

%
make_rewrite_pred(_, A = B, [A,B]):-!.
make_rewrite_pred(R, (A = B :- Eqs), [A,B]:- G0) :-
	make_rewrite_cond(Eqs, R, G0).

%
make_rewrite_cond((X, Y), R, (X0, Y0)):-
	make_rewrite_cond(X, R, X0),
	make_rewrite_cond(Y, R, Y0).
make_rewrite_cond((X; Y), R, (X0; Y0)):-
	make_rewrite_cond(X, R, X0),
	make_rewrite_cond(Y, R, Y0).
make_rewrite_cond(U=V, R, call(R, U, V)).
make_rewrite_cond(G, _, G).

% ?- pac_etc:list_to_ampersand([a,b,c], R).
list_to_ampersand([], []).
list_to_ampersand([X], X).
list_to_ampersand([X,Y|Z], &(X, U)):- list_to_ampersand([Y|Z], U).
%
etc(head_sed, [E|As], M, H, P, Q)	:-
    pac:expand_sed(E, [F, W, A]),
	pac_word:expand_head_sed(F, W, A, M, G0, P, P0),
	pac:expand_core(pred(F, [In, Out]:- once(call(G0, Out, In, []); Out=In)),
			     M, H0, P0, Q),
	complete_args(H0, As, H).
			/***************
			*     foldl    *
			***************/
% ?- foldl(plus, [1,2,3,4,5,6,7,8,9,10], 0, R).
% ?- foldl(plus, [1, 2, 3], 0, R).
% ?- foldl(flip([1,3,2], plus), [1,2,3,4,5,6,7,8,9], 0, R).
% ?- foldl(flip([1,3,2], pred([X, Y, [X|Y]])), [1,2,3,4,5,6,7,8,9], R, []).
% ?- foldl(flip([1,3,2], pred([X, U, [X|U]])), [a,b,c], R, []).
%@ R = [a, b, c].
etc(foldl, [F|Args], _, Foldl, P, P):- var(F),!,
	complete_args(foldl, [F|Args], Foldl).
etc(foldl, [F|Args], M, G, P, Q):-
	expand_arg(F, M, F0, P, P0),
	term_variables(F0, Vs),
	expand_core(
		rec(R, Vs,
			[[], X, X]
		   &
			([[A|As], X, Y]:-
				call(F0, A, X, X0),
				call(R, As, X0, Y))),
		 M, G0, P0, Q),
	complete_args(G0, Args, G).
etc(foldl, [], _, foldl, P, P).

			/***************
			*     foldr    *
			***************/
% ?- foldr(pred([X, U, [X|U]]), [a,b,c], [], R).
% ?- foldr(plus, [1, 2, 3], R, 0).
% ?- F = plus, foldr(F, [1, 2, 3], R, 0).
% ?- foldr(pred([X, Y, Z]:- plus(X, Z, Y)), [1, 2, 3], R, 0).
% ?- foldr(pred([X, Y, Z]:- flip([1,3,2], plus, X, Y, Z)), [1, 2, 3], R, 0).
% ?- foldr(pred([X, Y, Z]:- call(flip([1,3,2], plus), X, Y, Z)), [1, 2, 3], R, 0).
% ?- call( pred([A, B]:- foldl(flip([1,3,2], plus), [1, 2, 3], B, A)), R, 0).
% ?- foldr(pred([X, Y, Z]:- plus(X, Z, Y)), [1, 2, 3], R, 0).
etc(foldr, [F|Args],_, meta:Foldr, P, P):- var(F),
	complete_args(foldr, [F|Args], Foldr).
etc(foldr, [F|Args], M, G, P, Q):-
	expand_arg(F, M, F0, P, P0),
	term_variables(F0, Vs),
	expand_core(
		rec(R, Vs,
			[[], X, X]
		  &	([[A|As], X, Y]:-
				call(F0, A, X0, Y),
				call(R, As, X, X0))),
		 M, G0, P0, Q),
	complete_args(G0, Args, G).
etc(foldr, [], _, foldr, P, P).


		/*************
		*     for    *
		*************/
% ?- for(1..3, writeln).
% ?- time(for(..(1,10), pred([X]:- write(X)))).
% ?- time(for(..(1, 5+5), pred([X]:- write(X)))).
% ?- for(..(1,10), pred([X]:- write(X))).
% ?- for(1-10, write).
% ?- Sum=sum(0), profile(
%	for( ..(1,100_000_000),
%		pred(Sum, [I]:- ( arg(1, Sum, S), Si is S+I, setarg(1, Sum, Si))))).
%@ Sum = sum(5000000050000000).
% ?- Sum=sum(0), time(
%	for( ..(1,100_000_000),
%		pred(Sum, [I]:- ( arg(1, Sum, S), Si is S+I, setarg(1, Sum, Si))))).
% ?- pac:show(for(1..10, pred([X]:- write(X)))).
% ?- pac:show(for(1..10, write)).
% ?- F = write, A=1, B=10, for(A..B, F).
% ?- F = write, A=1, B=10, pac:show(for(A..B, F)).

etc(for, [Iexp, F], M, G, P, Q):-!, int_intervalval_exp(Iexp, I, J),
	etc_for([I, J, F], M, G, P, Q).
%
int_intervalval_exp(E, I, J):- var(E), !, E = ..(I,J).
int_intervalval_exp(E, I, J):- once(E= ..(L, R); E= -(L, R)),
	(	ground(L) -> I is  L
	;	I = L
	),
	(	ground(R) -> J is R
	;	J = R
	).
%
etc_for([I, J, F], _, meta:for(I, J, F), P, P):- var(F), !.
etc_for([I, J, F], M, G, P, Q):-
	expand_arg(F, M, F0, P, P0),
	term_variables(F0, Vs),
	(	integer(J)
	->	Plist=[A],
		Qlist = [I],
		Rlist = [A0],
		Check = (A>J)
	;	Plist=[A,B],
		Qlist=[I,J],
		Rlist = [A0, B],
		Check = (A>B)
	),
	CALL =..[call, R|Rlist],
	expand_core(
		rec(R, Vs,
				(Plist:- Check, !)
			&	(Plist:- call(F, A), A0 is A + 1, CALL)
		   ),
		 M, G0, P0, Q),
	complete_args(G0, Qlist, G).


		/****************
		*     repeat    *
		****************/
% ?- repeat(between(1, 3, I), writeln(I)).
% ?- repeat(3, writeln(hello)).
% ?- repeat(between(1, 5, J), writeln(J)).
% ?- repeat(0, write(.)).
% ?- repeat(100, write(.)).
% ?- repeat(10-8, write(.)).
% ?- repeat(10 mod 3, write(.)).
% ?- R is 10//2, G= write(.), repeat(R, G).

etc(repeat, As, M, G, P, Q):- etc_repeat(As, M, G, P, Q), !.

%
etc_repeat([Rep, Goal], _, G, P, P):- ( var(Rep); var(Goal) ), !,
	G = meta:repeat(Rep, Goal).
etc_repeat([Rep, Goal], M, G, P, Q):- repeat_cond(Rep, Cond), !,
	expand_goal((Cond, Goal, fail; true), M, G, P, Q).
%
repeat_cond(N, between(1, N0, _)):- simple_int_exp(N), N0 is N.
repeat_cond(X, X).
%
simple_int_exp(N):- integer(N).
simple_int_exp(E):- ground(E),
	functor(E, F, _),
	memberchk(F, [+,-,^,mod, //]).

% % NOT FAST !! (Copy_term seems better).
% etc(foreach, [Gen, Con, Vs, Ws], M, G, P, Q):-
% 	pac:expand_core(pred([Vs, Ws]:- Con), M, F, P, P0),
% 	G0 =
% 	(	Stash = '$STASH'(_),
% 		nb_setarg(1, Stash, Ws),
% 		(	call(Gen),
% 			arg(1, Stash, U),
% 			once(call(F, Vs, U)),
% 			nb_setarg(1, Stash, U),
% 			fail
% 		;   arg(1, Stash, Ws)
% 		)
% 	),
% 	expand_goal(G0, M, G, P0, Q).


% ?- show( fold(I, between(1, 10, I), plus, 0, S)).
% ?- time( fold(I, between(1, 100000, I), plus, 0, S)).
% ?- N = 3, numlist(1, N, Ns), time(fold(M, append(M, _, Ns), pred([X, Y, Z]:- append(X, Y, Z)), [], S)).
% ?- N = 3, numlist(1, N, Ns), time(fold(M, append(M, _, Ns), append, [], S)).
% ?- use_module(pac(basic)).
% ?- N = 3, numlist(1, N, Ns), time(fold(M, member(M, Ns), cons, [], S)).
% ?- N = 3, numlist(1, N, Ns), time(fold(M, member(M, Ns), append, [], S)).
% ?- N = 3, A = append, E = member(M, Ns), numlist(1, N, Ns), time(fold(M, member(M, Ns), append, [], S)).
%
etc(fold, As, M, G, P, Q):-!, etc_fold(As, M, G, P, Q).

etc_fold([Arg, Gen, Act|As], _, G, P, P):- (var(Gen), var(Act)), !,
	complete_args(meta:fold(Arg, Gen, Act), As, G).
etc_fold([Arg, Gen, Act, X, Y], M, G, P, Q):-!,
	expand_arg(Act, M, Act0, P, P0),
	expand_goal(Gen, M, Gen0, P0, P1),
	expand_goal((	Acc = '$acc'(X),
					(	Gen0,
						arg(1, Acc, U),
						call(Act0, Arg, U, V),
						nb_setarg(1, Acc, V),
						fail
					;	arg(1, Acc, Y)
					)
				),
				M, G, P1, Q).

% ?- A = acc(0), fold(M, member(M, [1,2,3]), pred([I, X]:- (arg(1, X, V), U is I + V, nb_setarg(1, X, U))),  A).
etc_fold([Arg, Gen, Act, X], M, G, P, Q):-!,
	expand_arg(Act, M, Act0, P, P0),
	expand_goal(Gen, M, Gen0, P0, P1),
	expand_goal((	(	Gen0,
						call(Act0, Arg, X),
						fail
					;	true
					)
				),
				M, G, P1, Q).

% ?- fold(M, member(M, [1,2,3]), writeln).
% ?- fold(M, member(M, [1,2,3]), X^(Y is X*X, writeln(Y))).
% ?- Gen = member(M, [1,2,2]), Con = writeln, fold(M, Gen, Con).
% ?- fold(M, member(M, [1,2,3]), I^(J is I*I, writeln(J))).
% ?- N is 10^8, time(fold(J, between(1, N, J), X^(X=X))).
% ?- N is 10^9, time(fold(J, between(1, N, J), X^(X=X))).
%@ % 2,000,000,001 inferences, 49.470 CPU in 49.519 seconds (100% CPU, 40428767 Lips)

etc_fold([X, Gen, X^Act], M, G, P, Q):-!,
	expand_goal(Gen, M, Gen0, P, P1),
	expand_goal((	Gen0,
					Act,
					fail
				;	true
				),
				M, G, P1, Q).
etc_fold([X, Gen, Act], M, G, P, Q):-
	expand_arg(Act, M, Act0, P, P0),
	expand_goal(Gen, M, Gen0, P0, P1),
	expand_goal((	Gen0,
					call(Act0, X),
					fail
				;	true
				),
				M, G, P1, Q).


		/*******************************
		*     unary/binary  foldnum    *
		*******************************/
% ?- foldnum(plus, 1-10, 0, X).
% ?- foldnum(pred([X,Y,Z]:- Z is X*Y), 1-4, 1, R).
% ?- foldnum(pred([X,Y,Z]:- Z is X*Y), 1-1000, 1, R).
% ?- F = plus, foldnum(F, 1-100, 0, R).
% ?- N=100,  functor(A, #, N),
%	forall(between(1, N, I), nb_setarg(I, A, I)),
%	foldnum(pred(A, ([J, C, D]:- arg(J, A, Aj), D is C * Aj) ),
%			1 - N, 1, S).

etc(foldnum, [F|As], _, meta:G, P, P):- var(F), !,
	complete_args(foldnum, [F|As], G).
etc(foldnum, [F, IntExp|As], M, G, P, Q):-
	int_intervalval_exp(IntExp, I, J),
	expand_arg(F, M, F0, P, P0),
	term_variables([I, J, F0], Vs),
	expand_core(
		rec(R, Vs,
			([A, U, U]:- A>J, !)
			   &
			([A, U, V]:- call(F, A, U, U0),
				A0 is A + 1,
				call(R, A0, U0, V))),
		 M, G0, P0, Q),
	complete_args(G0, [I|As], G).

		/****************************
		*     fold_paths_of_term    *
		****************************/
% ?- fold_paths_of_term(pred([A,[A|B], B]), f(1,2), X, []).
etc(fold_paths_of_term, [F|As], _, meta:G, P, P):- var(F), !,
	complete_args(fold_paths_of_term, [F|As], G).
etc(fold_paths_of_term, [F|Args], M, G, P, Q):-
   	expand_arg(F, M, F0, P, P0),
	term_variables(F0, Vs),
	expand_core(mrec(Vs, [
		_Entry = pred([T, X, Y]:- call(Fold_4, [[T]], [], X, Y)),
		Fold_4 = pred(
			 &(	([[], _, X, X]:-!),
				 ([[Ts|L], Path, X, Y]:-
						(	Ts==[]
						->	call(Fold_4, L, Path, X, Y)
						;	Ts=[T|Rs],
							call(Fold_5, T, [Rs|L], Path,  X, Y)
						)))),
		Fold_5 = pred(
			  &( ([T, Ls, Path, X, Y]:- atomic(T), !,
						call(F, [T|Path], X, Xtp),
						call(Fold_4, Ls, Path, Xtp, Y) ),
			 	 ([T, Ls, Path, X, Y]:- T=..[Ft, At|As],
						call(Fold_5, At, [As|Ls], [Ft|Path], X, Y)
						)))
		]),
		M, G0, P0, Q),
	complete_args(G0, Args, G).

		/*******************
		*     fold_args    *
		*******************/
% ?- fold_args(plus, f(1,2,3,4), 0, S).
% ?- F=plus, fold_args(F, f(1,2,3,4), 0, S).
% ?- fold_args(pred([X, Y, Z]:- Z is X + Y), f(1,2,3,4), 0, S).
etc(fold_args, [F|As], _, meta:G, P, P):- var(F), !,
	complete_args(fold_args, [F|As], G).
etc(fold_args, [F|As], M, G, P, Q):-
   	expand_arg(F, M, F0, P, P0),
	term_variables(F0, Vs),
	expand_core(
		mrec(Vs, [	_Entry	= pred(( [V, A, B]:-
										functor(V, _, Nv),
										call(Fold_args, 1, Nv, V, A, B))),
					Fold_args = pred(
							  ([I, Nv, _, A, A]:- I > Nv, !)
								&
							  ([I, Nv, V, A, B]:-
									arg(I, V, Vi),
									call(F, Vi, A, Ai),
									I1 is I + 1,
									call(Fold_args, I1, Nv, V, Ai, B)))
				  ]),
				M, G0, P0, Q),
	complete_args(G0, As, G).

		/*****************
		*     mapterm    *
		*****************/

% ?- mapterm(=, f(a, b), Out).
%@ Out = f(a, b).
% ?- mapterm(pred([a,b]), f(a,a), R).
%@ R = f(b, b).
% ?- mapterm(mapterm(pred([a,b])), f(g(a,a), h(a,a)), R).
%@ R = f(g(b, b), h(b, b)).
% ?- mapterm(pred([A, [A,A]]), f(a,b), Out).
%@ Out = f([a, a], [b, b]).
% ?- F = pred([A, [A,A]]), mapterm(F, f(a,b), Out).
%@ F = update_link:'pac#16',
%@ Out = f([a, a], [b, b]).
%@ F = update_link:'pac#13',
%@ Out = f([a, a], [b, b]).

% etc(mapterm, [F|As], _, meta:G, P, P):- var(F), !,
% 	complete_args(mapterm, [F|As], G).
% etc(mapterm, [F|As], M, G, P, Q):-
%    	expand_arg(F, M, F0, P, P0),
% 	term_variables(F0, Vs),
% 	expand_core(
% 		mrec(Vs, [	_Entry	= pred(( [A, B]:-
% 										functor(A, Fa, Na),
% 										functor(B, Fa, Na),
% 										call(Mapterm, Na, A, B))),
% 					Mapterm = pred( ([0, _, _]:- !)
% 								&
% 							  ([I, A, B]:-
% 									arg(I, A, Ai),
% 									arg(I, B, Bi),
% 									call(F, Ai, Bi),
% 									J is I - 1,
% 									call(Mapterm, J, A, B)))
% 				  ]),
% 				M, G0, P0, Q),
% 	complete_args(G0, As, G).

% ?- mapterm_rec(=, f(a, b), Out).
%@ Out = f(a, b).
% ?- mapterm_rec(pred([a, b]), f(a, b), Out).
%@ Out = f(b, b).
% ?- mapterm_rec(pred([a, b]), f(g(a), h(a,b)), Out).
%@ Out = f(g(b), h(b, b)).
% ?- show(mapterm_rec(pred([a, b]), f(g(a), h(a,b)), Out)).

etc(mapterm_rec, [F|As], _, meta:G, P, P):- var(F), !,
	complete_args(mapterm_rec, [F|As], G).
etc(mapterm_rec, [F|As], M, G, P, Q):-
   	expand_arg(F, M, F0, P, P0),
	term_variables(F0, Vs),
	expand_core(
		mrec(Vs, [	Entry	= pred(( [A, B]:-
										functor(A, Fa, Na),
										functor(B, Fa, Na),
										call(Mapterm_rec, Na, A, B))),
					Mapterm_rec = pred( ([0, _, _]:- !)
								&
							  ([I, A, B]:-
									arg(I, A, Ai),
									arg(I, B, Bi),
									(	call(F0, Ai, Bi)-> true
									;	atomic(Ai) -> Bi = Ai
									;	call(Entry, Ai, Bi)
									),
									J is I - 1,
									call(Mapterm_rec, J, A, B)))
				  ]),
				M, G0, P0, Q),
	complete_args(G0, As, G).

		/***************************
		*     recursive maplist    *
		***************************/

% ?- show(maplist_rec(plus(1), [0,1,2], Out)).
% ?- N = 1000, K=1000, numlist(1, N, Ns),  length(Ks, K),
%	maplist(=(Ns), Ks), time(maplist_rec(plus(1), Ks, Out)).

etc(maplist_rec, [F|As], _, meta:G, P, P):- var(F), !,
	complete_args(maplist_rec, [F|As], G).
etc(maplist_rec, [F|As], M, G, P, Q):-
   	expand_arg(F, M, F0, P, P0),
	term_variables(F0, Vs),
	expand_core(
		mrec(Vs, [	Main = pred(
						[[], []]
					&	([[X|Xs], [Y|Ys]]:-
							(	X = [_|_]      % for fast is_list(X)
							-> 	call(Main, X, Y)
							;	call(F0, X, Y)
							),
							call(Main, Xs, Ys))
					)
				  ]),
				M, G0, P0, Q),
	complete_args(G0, As, G).


		/*****************
		*     maprows    *
		*****************/
% ?-  qcompile(zdd(zdd)), module(zdd).
%   (inner_prod is missing.)
% ?-  maprows(zdd:inner_prod(f(1,2)), m(f(1,2), f(3,4)), B).
% ?-  maprows(zdd:inner_prod(f(1,2)), m(f(1,2,3), f(3,4,5)), B).
etc(maprows, [F|As], _, meta:G, P, P):- var(F), !,
	complete_args(maprows, [F|As], G).
etc(maprows, [F|As], M, G, P, Q):-
	expand_arg(F, M, F0, P, P0),
	term_variables(F0, Vs),
	expand_core(
		mrec(Vs, [	_Entry	= pred(([A, A]:- atom(A), !)
								   &
								   ( [A, B]:- arg(1, A, A1),
										functor(A1, Fa, Na),
										functor(B, Fa, Na),
										call(Maprows, Na, A, B))),
					Maprows = pred( ([0, _, _]:- !)
								&
							  ([I, A, B]:- arg(I, B, Bi),
									call(F, I, A, Bi),
									J is I - 1,
									call(Maprows, J, A, B)))
				  ]),
				M, G0, P0, Q),
	complete_args(G0, As, G).

		/*****************
		*     mapargs    *
		*****************/
% now a SWI library.
% ?- Y = f(1, 2), mapargs(=, count(2, 1, 1), f(a,b), Y).
% ?- Y = f(1, 2), mapargs(=, count(1, 2, 2), f(a,b), Y).
% ?- Y = f(1, 2), mapargs(=, count(1, 1, 2), f(a,b), Y).
% etc(mapargs, [F|As], _, meta:G, P, P):- var(F), !,
% 	complete_args(mapargs, [F|As], G).
% etc(mapargs, [F|As], M, G, P, Q):-
% 	expand_arg(F, M, F0, P, P0),
% 	term_variables(F0, Vs),
% 	expand_core( mrec(Vs,
% 		 [	_Entry = pred(
% 				( [count(N, I, J), A, B]:-
% 					call(Mapargs, N, I, J, A, B))),
% 			Mapargs = pred(
% 				([0, _, _, _, _]:-!)
% 			 &	([N, I, J, A, B]:-
% 					arg(I, A, Ai),
% 					call(F, Ai, Bj),
% 					setarg(J, B, Bj),
% 					N1 is N-1,
% 					I1 is I+1,
% 					J1 is J+1,
% 					call(Mapargs, N1, I1, J1, A, B)))
% 		]),
% 		M, G0, P0, Q),
% 	complete_args(G0, As, G).


		/*****************************
		*     while/until/iterate    *
		*****************************/

% ?- until( pred([s(I, _)]:- I>10),
%	pred(([s(I, X), s(J, Y)]:- J is I+1, Y is J*X, writeln(Y))), s(1,1),  R).
% ?- let(S, pred(([s(I, X), s(J, Y)]:- J is I+1, Y is J*X, writeln(Y)))),
%	let(Fin, pred([s(I, _)]:- I>10)),
%   until(Fin, S, s(1,1), R).
etc(until, [Fin, S|Args], M, meta:G, P, Q):- (var(Fin); var(S)), !,
	expand_arg(Fin, M, Fin0, P, P0),
	expand_arg(S, M, S0, P0, Q),
	complete_args(until, [Fin0, S0|Args], G).
etc(until, [Fin, S|Args], M, G, P, Q):-
	pre_expand_arg(Fin, M, F0, VsF, P, P_),
	pre_expand_arg(S, M, S0, VsS, P_, P0),
	union(VsF, VsS, Vs),
	expand_core(
		rec(Rec, Vs,
			([X, Y]:-
				call(S0, X, X0),
				(	call(F0, X0) ->	Y = X0
				;	call(Rec, X0, Y)
				)
			)
		   ),
		 M, G0, P0, Q),
	complete_args(G0, Args, G).


% ?- A = 5,
%	iterate( pred(A,
%			( [ s(X, Y), s(X0, Y0) ] :- X<A, !, X0 is X+1,  Y0 is X*Y )
%		 & 	[ U, stop(U)] ) ,
%			s(1,1),  R).

% ?- let(F,  iterate(pred(A,
%			( [ s(X, Y), s(X0, Y0) ] :- X<A, !, X0 is X+1,  Y0 is X*Y  )
%		&	[ U, stop(U)] ))),
%	A = 5,
%	call(F, s(1,1),  R).

etc(iterate, [S|Args], M, meta:G, P, Q):- var(S), !,
	expand_arg(S, M, S0, P, Q),
	complete_args(iterate, [S0|Args], G).
etc(iterate, [S|Args], M, G, P, Q):-
	pre_expand_arg(S, M, S0, Vs, P, P0),
	expand_core(
		rec(Rec, Vs,
			(	[stop(X), X]:-!		)	&
			(   [X, Y]:- call(S0, X, X0), !, call(Rec, X0, Y) )),
			 M, G0, P0, Q),
	complete_args(G0, Args, G).

% ?- while(pred([s(X)]:- X < 10), pred(([s(X),s(Y)]:- writeln(X),
% Y is X + 1)), s(0), R).
% ?- let(Fin, pred([s(X)]:- X < 10)),
%	let(S, pred(([s(X),s(Y)]:- writeln(X), Y is X + 1))),
%	while(Fin, S, s(0), R).
etc(while, [Fin, S|Args], M, meta:G, P, Q):- (var(Fin); var(S)), !,
	expand_arg(Fin, M, Fin0, P, P0),
	expand_arg(S, M, S0, P0, Q),
	complete_args(while, [Fin0, S0|Args], G).
etc(while, [Fin, S|Args], M, G, P, Q):-
	pre_expand_arg(Fin, M, F0, VsF, P, P_),
	pre_expand_arg(S, M, S0, VsS, P_, P0),
	union(VsF, VsS, Vs),
	expand_core(
		rec(Rec, Vs,
			([X, Y]:- (		call(F0, X)
					  ->	call(S0, X, X0),
							call(Rec, X0, Y)
					  ;		Y = X
					  )
			)
		   ),
		 M, G0, P0, Q),
	complete_args(G0, Args, G).


		/**************
		*     flip    *
		**************/

% flip([j1, j2, ..., Jn], F)  is defined as G such that
%    G(a1, a2, ..., an) if F(aj1, aj2, ..., ajn).

% swap the last two arguments.
% ?- flip([2,1], is, 1+1, R).
% ?- flip([2,1], flip([2,1], is), R, 1+1).
% ?- flip([2,1], flip([2,1], flip([2,1],is)), 1+2, R).
% ?- flip([2,1], string_codes, `abc`, R).
% ?- flip([1,3,2], append, [a,b], R, [c,d]).
% ?- P = [1,3,2], F=append, flip(P, F, [a,b], R, [c,d]).
% ?- show(flip([1,3,2], append, [a,b], R, [c,d])).
% ?- let(F, flip([2,1], flip([2,1], flip([2,1],is)))), call(F, 1+2, R).
% ?- let(F, flip([2,1], flip([2,1], is))), flip([2,1], F, 1+2, R).
% ?- let(F, flip([2,1], flip([2,1], is))), L= [2,1], flip(L, F, 1+2, R).
% ?- foldl(pred([X, Y, Z]:- plus(X, Z, Y), [1, 2, 3])), 0, R).
% ?- foldl(plus, [1, 2, 3], 0, R).
% ?- flip([2,1], foldl(plus, [1, 2, 3]), 0, R).
% ?- flip([2,1], foldl(plus, [1, 2, 3]), R, 0).
% ?- F = plus, flip([2, 1], foldl(F, [1, 2, 3]), R, 0).
% ?- let(F, foldl(G, [1, 2, 3])), G = plus, flip([2,1], F, R, 0).
% ?- flip([2,1], plus(5), X, 3).

etc(flip, Args, M, G, P, Q):- etc_flip(Args, M, G, P, Q).
%
etc_flip([Permu, F|Args], M, meta:G, P, Q):- (var(Permu); var(F)), !,
	expand_arg(F, M, F0, P, Q),
	complete_args(flip(Permu, F0), Args, G).
etc_flip([Permu, F|Args], M, G, P, Q):- expand_arg(F, M, F1, P, P0),
 	term_variables(F1, Vs),
	length(Permu, Len),
	length(In, Len),
	flip_list(Permu, In, Out),
 	complete_args(F1, Out, Call),
	expand_core(pred(Vs, In:- Call), M, G0, P0, Q),
	complete_args(G0, Args, G1),
	slim_goal(G1, G).

		/***************
		*     xargs    *
		***************/

etc(xargs, Args, M, G, P, Q):-  etc_xargs(Args, M, G, P, Q).
%
etc_xargs([], M, G, P, P):- complete_args(M:call, [], G).
etc_xargs([F|Args], M, G, P, P):- var(F), !,
	complete_args(M:F, Args, G).
etc_xargs([F|Args], M, G, P, Q):-
	(	F = (Head:- Body)
	->	pac:expand_goal(Body, M, G0, P, Q),
		pac:slim_exp_goal((pac:partial_args_match(Head, Args), G0), G)
	;	(	F = (_Head -> _Body)
		->	pac:expand_exp(F, Args, call, _V , M, G, P, Q)
		; 	Args = [A1,A2|Rest],
			etc_xargs([[A1, A2|Rest] :- apply(F, [A2, A1|Rest])], M, G, P, Q)
		)
	).

		/*****************
		*     cputime    *
		*****************/

% ?- pac_etc:etc(cputime, [3, A=b, T], user, G, P, []).
% ?- nopac(meta:cputime(3000000, (b_setval(a,b),b_getval(a,B)), T)).
% ?- pac_etc:cputime(3000000, (b_setval(a,b),b_getval(a,B)), T).

etc(cputime, [], _,  meta:cputime, P, P):-!.
etc(cputime, [N], _,  meta:cputime(N), P, P):-!.
etc(cputime, [N, Goal|Args], _, meta:G, P, P):-	var(Goal),!,
	complete_args(cputime(N, Goal), Args, G).
etc(cputime, [Count, Goal|Args], M, G, P, Q):-
	goal_to_pred(Goal, M, Vs:-Goal0, P, P1),
	expand_core(
		mrec(Vs,
			[	_Top = pred(([N, T]:- writeln("running compiled cputime/3 ...  "),
										   call(Repeat, N, 0.00, T))),
			 Repeat	 = pred(	( [0, T, T])
							&	( [N, T, T0]:-  succ(N0, N),
										call(Cputime, S),
										T1 is T + S,
										call(Repeat, N0, T1, T0))),
			Cputime  = pred(( [T]:-  statistics(cputime, T0),
										call(Goal0),
										statistics(cputime, T1),
										T is T1-T0))
			]), M, G0, P1, Q),
	complete_args(G0, [Count|Args], G).


		/*****************
		*     collect    *
		*****************/
% ?- collect(integer, [1,a,2,b], X).
etc(collect, [], _, basic:collect, P, P):- !.  % collect/3 is a meta-predicate
etc(collect, [F|Args], _, basic:Collect, P, P):- var(F), !,
	Collect=..[collect, F|Args].
etc(collect, [F], M, G, P, Q):- !,
	etc_collect(F, M, G, P, Q).
etc(collect, [F, X], M, G, P, Q):- !,
	etc_collect(F, M, G0, P, Q),
	complete_args(G0, [X], G).
etc(collect, [F, X, V], M, G, P, Q):- !,
	etc_collect(F, M, G0, P, Q),
	complete_args(G0, [X, V], G).
%
etc_collect(F, M, G, P, Q):- !,
	expand_arg(F, M, F0, P, P0),
	term_variables(F0, F, Vs),
	expand_core(
		rec(R, Vs,
			( [[A|As], S]  :-
				( call(F0, A)
				->  S = [A|S0]
				;   S = S0
				),
				call(R, As, S0) )
		  &	( [[], []] :- true  )
		   ),
		 M, G, P0, Q).

		/**********************
		*     maplist_opp     *
		**********************/

% ?- maplist_opp([=, =, =], a, X).
% ?- Fs = [=, =, =], maplist_opp(Fs, a, X).
% ?- maplist_opp([plus(1), plus(2), plus(3)], 0, X).
% ?- maplist_opp([pred([X,Y]:-plus(X,10,Y)),
%				pred([X,Y]:-plus(X,20,Y))], 1,  U).

etc(maplist_opp, [L|R], M, meta:Call_maplist_opp, P, Q):-
    map_arg_list(L, M, L0, P, Q),
    Call_maplist_opp=..[maplist_opp, L0|R].

% ?-pac_etc:map_arg_list([a,b,c], [], L, P, []).
map_arg_list(A, _, A, P, P):-var(A),!.
map_arg_list([A|As], M, [A0|As0], P, Q):-
    expand_arg(A, M, A0, P, P0),
    map_arg_list(As, M, As0, P0, Q).
map_arg_list([], _, [],  P, P).

		/**********************
		*     term_rewrite    *
		**********************/


% ?- show(term_rewrite(pred([a,b]), {*,+,-}, a+b, X)).
% ?- term_rewrite(pred([a,b]), \+ (*), a+b, X).
% ?- term_rewrite(pred([a,b]), \+ (*), a+b*c, X).
% ?- term_rewrite(pred([a,b]), \+ (*), a*c+b*c, X).
% ?- call(term_rewrite(pred([a,b]), \+ (*)), a*c+b*c, X).
%
etc(term_rewrite, [R, E|Args], M, Call, P, Q):- ground(R), ground(E), !,
	misc:expand_sgn_brace(E, E0),
	expand_arg(R, M, R0, P, P0),
	expand_core(
		pred(([X, Y]:- reduce:subtree(E0, X, Y, S, S0),
				  nonvar(S),
				  call(R0, S, S0))),
		M, Reduce_one, P0, P1),
	expand_core(
		rec(Reduce,
				([X, Y]:-
				  call(Reduce_one, X, X0),
				  !,
				  call(Reduce, X0, Y))
		   &	[X, X]),
		M, G, P1, Q),
	complete_args(G, Args,  Call).
etc(term_rewrite, [R, E|Args], _, Call, P, P):-
	complete_args(reduce:term_rewrite(R, E), Args, Call).


%	A termplate for etc plug-in
%
% etc(pacmap, [F|Args], _, Maplist, P, P):- var(F), !,
% 	Maplist=..[maplist, F|Args].
% etc(pacmap, [F|Args], M, G, P, Q):-
% 	expand_arg(F, M, F0, P, P0),
% 	free_variables(F0, F, Vs),
% 	length(Args, N),
% 	length(EmptyLists, N),
% 	maplist(=([]), EmptyLists),
% 	length(Params, N),
% 	maplist(cons, Cars, Cdrs, Params),
% 	Call1 =..[call, F0|Cars],
% 	Call2 =..[call, R|Cdrs],
% 	expand_core(
% 		rec(R, Vs,
% 			EmptyLists &
% 			( Params :- Call1, Call2 )
% 		   ),
% 		 M, G0, P0, Q),
% 	complete_args(G0, Args, G).
% etc(pacmap, [], _, maplist, P, P).   % maplist is a builtin.

%@ true.
