:- module(cil, [unify/2,
				find_key/3,
				close_btree/1,
				distribute_constr/2,
				region_constr_of_leaves/4
			   ]).
:- use_module([
		library(lists),
		library(sort),
		library(ordsets)]).

% for test.
% ?- qcompile(util('ptq-fragment')), module(cil).


	/****************************************************
	*            CIL attributed variables version       *
	*            based on attr_unify_hook/2.            *
	****************************************************/

% for short.
% ?- put_attr(con(total), X).
put_attr(X, Attr):- put_attr(X, cil, Attr).
get_attr(X, Attr):- get_attr(X, cil, Attr).

% After Jan's advice.
attr_unify_hook(V, Y):-
	(	get_attr(Y, cil, A)
	->	attr_unify(V, A, Y)
	;   attr_unify(V, Y)
	).

%% attr_unify(+X, +Y) is det.
%
%  Apply attribute X to Y.
attr_unify(btree(X), Y):- is_btree(Y), !, unify(X, Y).
attr_unify(con(Rgn, _), Y):- is_btree(Y), !,
						  distribute_constr(Y, Rgn).
attr_unify(_, _).

%% attr_unify(+X, +Y, -Z) is det.
%
%  Merge attributes of X and Y, and save the merged
%  attribute as that of Z in case it is necessary.
%  Note that here merge includes unification.

attr_unify(btree(X), btree(Y), _):- unify(X, Y).
attr_unify(btree(X), con(Rgn, D), Z):-
	distribute_constr(X, Rgn),
	put_attr(Z, con(Rgn, D), Z),
	subsume(X, D).
attr_unify(con(Rgn, Y), btree(D), Z):-
	distribute_constr(D, Rgn),
	put_attr(Z, con(Rgn, Y)),
	subsume(D, Y).
attr_unify(con(Rgn, Y), con(Rgn0, Y0), Z):-
	meet_region(Rgn, Rgn0, Rgn1),
	add_only_new(Y, Y0, Y1),
	put_attr(Z, con(Rgn1, Y1)),
	unify(Y, Y0).

	/****************************************************
	*   Unification over feature sructures				*
	****************************************************/

% Kernel Of CIL  by K. Mukai 01-DEC-85
%
% Originally coded in 1985 at ICOT.
%
% Revised around 2016-11-13.
%
% This is a unifier over feature structures rewritten on top of
% attributed variables in SWI-7.
%
%
% key		:=	<prolog ground term>
%
% btree		:=	[]			(void tree)
%			|	t(key, zterm, btree, btree)
%
% zterm		:=	btree
%			|	<prolog variable>
%			|	<prolog atomic term>
%			|	<prolog atom>(zterm, ..., zterm)


	/***************************************
	*   Unifier over feature structues.    *
	***************************************/

% For test.
% ?- module(cil).
% ?- X.a=1.
% ?- X.a=1, X=Y, Y.a=V.
% ?- X=Y, X.a=1, Y.a=V.
% ?- X.a=1, Y.a=V, X=Y.
% ?- X.a #= Y.b, X={b:1}, X=Y, Y.a=V.
% ?- X.a = X.b, X={b:1}.
% ?- X={a:2}, Y={b:1}, X=Y.
% ?- X=f({a:2}), Y=f({b:1}), X=Y, f(A)= X, B = A.b.
% ?- X=f({a:2}), Y=f({b:1}), X=Y.

is_btree(t(_,_,_,_)).
is_btree({}).
%
apply_constr(X, con(I,_)):- distribute_constr(X, I).

%% distribute_constr(+X, +Rgn) is det.
%  Distribute a region constraint Rgn down to the leaves of X.
%
distribute_constr(X, Rgn):- var(X), !,
	(	get_attr(X, A)
	-> 	(	A = con(RgnA, D)
		-> 	meet_region(RgnA, Rgn, RgnA0),
			put_attr(X, con(RgnA0, D))
		;	A = btree(A0),
			distribute_constr(A0, Rgn)
		)
	;	put_attr(X, con(Rgn, []))
	).
distribute_constr({}, _):-!.
distribute_constr(t(K,_,L,R), Rgn):-
	has_member(Rgn, K),
	meet_region(Rgn, lower(K), LowK),
	meet_region(Rgn, upper(K), UppK),
	distribute_constr(L, LowK),
	distribute_constr(R, UppK).

%% region_constr_of_leaves(+X, +Rgn, +P, -Q) is det.
%   The list of region constraints on leaves of X with Rgn
%	propagated down  is a  difference list P by Q.
%

region_constr_of_leaves(X, Rgn, P, Q):- child_of_node(X, K, _, L, R), !,
	meet_region(lower(K), Rgn, RgnL),
	region_constr_of_leaves(L, RgnL, P, P0),
	meet_region(upper(K), Rgn, RgnR),
	region_constr_of_leaves(R, RgnR, P0, Q).
region_constr_of_leaves(X, Rgn, [put_attr(X, cil, con(Rgn, []))|P], P):- var(X), !.
region_constr_of_leaves(_, _, P, P).

%%  unify(+X, +Y) is det.
%   unify X with Y.
%	Unification dispatcher.

unify(X, Y):- (attvar(X); attvar(Y)), !, X=Y.
unify(X, Y):- (var(X); var(Y)), !, X=Y.
unify(X, Y):- is_btree(X), is_btree(Y),	!, subsume(X, Y).
unify(X, X).	% cool !

%% subsume(+X:btree, +Y:btree) is det.
%	Merge btree X into btree Y.
%   subsume(X,Y) and subsume(Y,X) implies X=Y.
%
subsume(X, Y):-attvar(X), !,
		get_attr(X, A),
		(	A=btree(T)
		->	subsume(T, Y)
		;	A=con(Rgn, S),
			add_only_new([Y], S, S0),   % sounds natural !
			put_attr(X, con(Rgn, S0))
		).
subsume(X, Y):-var(X), !,
		put_attr(X, con(total, [Y])).
subsume(t(K,V,L,R), Y):-
		find_key(Y, K, U),
		unify(V, U),
		subsume(L, Y),
		subsume(R, Y).
subsume({}, _).

%% child_of_node(X, K, V, L, R) is det.
% True if X is a nonterminal node with key K,
%	value V, left child L, and right child R.

child_of_node(X, K, V, L, R):- var(X), !,
					   get_attr(X, A),
					   A=btree(t(K, V, L, R)).
child_of_node(t(K, V, L, R), K, V, L, R).

%%%
% inspect_node(X, A):-var(X), (get_attr(X, A); A=var(X)), !.
% inspect_node(X, X).

%% weak_subsume(+X:btree, +Y:btree, -Z:btree) is det.
%	True if  Z.a = V iff one of the following holds.
% 	1. Y.a = V, and V is instantiated.
%	2. Y.a = V, and X.a is not instantiated.
%	3. X.a = V, V is instantiated, and Y.a uninstantiated.

% ?- module(cil).
% ?- cil:weak_subsume({a:2}, {a:1}, Z).
% ?- cil:weak_subsume({a:2, b:1, c:3}, {a:1, b:X, c:5, d:6}, Z).
weak_subsume(X, Y, Z):-attvar(X), !,
		get_attr(X, A),
		(	A=btree(T)
		->	weak_subsume(T, Y, Z)
		;	Z = Y
		).
weak_subsume(X, Y, Y):-var(X), !.
weak_subsume(t(K,V,L,R), Y, Z):-
		find_key(Y, K, U),
		(unify(V, U); true),
		!,
		weak_subsume(L, Y, Z),
		weak_subsume(R, Y, Z).
weak_subsume({}, Y, Y).

% :- module(cil).
%@ true.
%@ true.

%% close_btree(+X) is det.
% Close all leaves of X with [] (void tree)
%
% ?- {a:1}=X, close_btree(X), X={b:1}.
% ?- {a:1}=X, close_btree(X).

close_btree(X):- child_of_node(X, _, _, L, R), !,
				 close_btree(L),
				 close_btree(R).
close_btree({}):-!.
close_btree(t(_, _, L, R)):-close_btree(L),
		close_btree(R).

% ?- [util('ptq-fragment')].
%% is_equal_btree(+X:btree, +Y:btree) is det.
%  True if closed X is equaivalent to closed Y as open dict.

% ?-is_equal_btree({a:1}, {a:1}).
% ?-is_equal_btree({a:1, b:2}, {b:2, a:1}).
% ?-is_equal_btree({a:f(2)}, {a:f(1)}).			% false
% ?-is_equal_btree({a:f({b:2})}, {a:f({b:3})}).	% false
is_equal_btree(X, Y):- is_equal_by_stack([X],[Y]).

%
is_equal_by_stack(X, Y):- skip_leaves(X, X0),
					skip_leaves(Y, Y0),
					is_equal_pair(X0, Y0).
%
is_equal_pair([],[]).
is_equal_pair([pair(K, V)|L],[pair(K, U)|M]):-
	is_equal_arg(V, U),
	is_equal_by_stack(L, M).
%
is_equal_arg(X, Y):- var(X), var(Y), X==Y, !.
is_equal_arg(X, Y):-
	(	child_of_node(X, K, V, L, R)
	->	(	child_of_node(Y, K0, V0, L0, R0)
		->  is_equal_by_stack(	[L,	 pair(K, V),   R],
								[L0, pair(K0, V0), R0])
		;	false
		)
	;	(	child_of_node(Y, _, _, _, _)
		->  false
		;   is_equal_non_btree(X, Y)
		)
	).
%
is_equal_non_btree(X, Y):- (var(X); var(Y)), !, X==Y.
is_equal_non_btree(X, Y):- (atomic(X); atomic(Y)), !, X==Y.
is_equal_non_btree(X, Y):- X=..[F|Xs],
				  Y=..[F|Ys],
				  maplist(is_equal_arg, Xs, Ys).
%
skip_leaves([], []).
skip_leaves([X|Xs], U):- child_of_node(X, K, V, L, R),!,
		skip_leaves([L, pair(K,V), R|Xs], U).
skip_leaves([X|Xs], U):- (var(X); X==[]), !,
		skip_leaves(Xs, U).
skip_leaves(X, X).

%% role(+K, ?X, ?V) is det.
%	unify V with the value  of the key K in the btree X
%   so that X.K = V

user:role(K, X, V):- when(ground(K), find_key(X, K, U)),
 				  unify(V, U).

%
min(X, Y, X):- X@<Y, !.
min(_, Y, Y).

max(X, Y, Y):- X@<Y, !.
max(X, _, X).

%% find_key(+B, +K,  -V):- det.
% True if B has a key K with value V, i.e., B.K = V.
% This is a primitive of this library.

% ?- find_key({a:1}, a, V, R).
% ?- find_key(X, a, V).
% ?- find_key(t(k, 1, L, R), k, V).
find_key(B, K, V):-
		(	var(B)
		-> 	(	get_attr(B, U)
		   	->	(	U=btree(U0)
		   		->	find_key(U0, K, V)
		   		;	insert_key(U, K, V, B)
		   		)
		   	;	T = t(K, V, L, R),
		   		put_attr(L, con(lower(K),[])),
		   		put_attr(R, con(upper(K),[])),
		   		put_attr(B, btree(T))
		   	)
		;	B=t(J,U,L,R),
			(   J==K
			->  V=U
			;   (	J @< K
				->	find_key(R, K, V)
				;	find_key(L, K, V)
				)
			)
		).

%
find_key_list([], _, _).
find_key_list([Y|Ys], K, V):- find_key(Y, K, V),
	find_key_list(Ys, K, V).

%% insert_key(+con(Rgn, Ys), K, V, N) is det.
%  Helper predicate for find_key/3 above.
%	True if N is a variable with an attribute t(K, V, L, R)
%	such that L and R have the region constraint con(Rgn, Ys)
%	making all members of Ys having a subtree t(K, V, _, _).
%   i.e, N.K=V, and Y.K=V for all  Y in Ys.

insert_key(con(Rgn, Ys), K, V, N):-
	has_member(Rgn, K),
	meet_region(Rgn, lower(K), RgnL),
	meet_region(Rgn, upper(K), RgnR),
	put_attr(L, con(RgnL, Ys)),
	put_attr(R, con(RgnR, Ys)),
	put_attr(N, btree(t(K, V, L, R))),
	find_key_list(Ys, K, V).

%% has_member(+R, +M) is det.
%	True if M falls in region R.
has_member(total, _).
has_member(seg(L,R), X):- L@<X, X@<R.
has_member(lower(U), X):- X@<U.
has_member(upper(U), X):- U@<X.

%% meet_region(+X, +Y) is det.
%  True if the intersection of X and Y is not empty.

meet_region(X, Y):- meet_region(X, Y, Z), Z\==[].

%% meet_region(+X, +Y, -Z) is det.
%
%  True if Z is the intersection of regions X and Y.
%  This is a primitive engine of this open dict library.

% ?- meet_region(lower(a), lower(b), X).
% ?- meet_region(seg(a, p), seg(b, z), X).
% ?- meet_region(seg(a, z), seg(b, z), X).
meet_region(total, X, X).
meet_region(X, total, X).
meet_region([], _, []).
meet_region(_, [], []).
meet_region(lower(A), lower(B), lower(C)):- min(A, B, C).
meet_region(lower(A), upper(B), R):-
	(	B @< A -> R = seg(B, A)
	;   R = []
	).
meet_region(lower(C), seg(A,B), R):-
	(	C @=< A -> R = []
	;	R = seg(A, D),
		min(B,C,D)
	).
meet_region(upper(A), upper(B), upper(C)):- max(A, B, C).
meet_region(upper(A), lower(B), R):-
	(	A @< B -> R = seg(A, B)
	;   R = []
	).
meet_region(upper(C), seg(A, B), R):-
	(	C @>= B
	->	R = []
	;   R = seg(D, B),
		max(A, C, D)
	).
meet_region(seg(A, B), lower(C), R):-
	(	C @=< A -> R = []
	;	R = seg(A, D),
		min(B,C,D)
	).
meet_region(seg(A, B), upper(C), R):-
	(	C @>= B
	->	R = []
	;   R = seg(D, B),
		max(A, C, D)
	).
meet_region(seg(A, B), seg(C, D), M):-
	max(A,C,L),
	min(B,D,R),
	(	L @>= R -> M = []
	;	M = seg(L, R)
	).

% % contain(+X, +Y) is det.
%  True if region X contains region Y, i.e. Y is a subset of X.
% ?- contain(lower(i), lower(j)).
% ?- contain(lower(j), lower(j)).

contain(total,_).
contain(_, []).
contain(lower(A), lower(B)) :- B@=<A.
contain(lower(A), seg(_,B)) :- B@=<A.
contain(upper(A), upper(B)) :- A@=<B.
contain(upper(A), seg(B,_)) :- A@=<B.
contain(seg(A,B), seg(A0, B0)) :- A@=<A0, B0@=<B.

%
add_only_new([], X, X):- !.
add_only_new([X|Y], Z, U):- memq(X, Z), !,
		add_only_new(Y, Z, U).
add_only_new([X|Y], Z, [X|U]):-
		add_only_new(Y, Z, U).
%
memq(X, [Y|_]):- X==Y, !.
memq(X, [_|Y]):- memq(X, Y).
