:- module(coalgebra, [show_am/1, regex_to_pdf/2, regex_to_pdf/3,
		      gv/1]).

% ?- gv(a).
% ?- gv("ab").

% ?- show_am(a+b).
% ?- show_am("([a-d])*[b-e]").
% ?- show_am("c|a|b|c").
% ?- show_am("cabc").
% ?- show_am("(ca)|(bc)").
% ?- show_am("(ca)+(bc)").
% ?- time(show_am("^(.*)/([^/]*)")).
%@ % 3,848 inferences, 0.001 CPU in 0.341 seconds (0% CPU, 5786466 Lips)
%@ true .
% ?- time(show_am("^(.*)/([^/]*)$")).
%@ % 12,850 inferences, 0.001 CPU in 0.349 seconds (0% CPU, 9518519 Lips)
%@ true .


:- use_module(library(lists)).
:- use_module(pac(basic)).
:- use_module(pac('expand-pac')).
:- use_module(pac('expand-word')).
:- use_module(util(misc)).
:- use_module(util(math)).
:- use_module(util('emacs-handler')).
:- use_module(util(meta2)).
:- use_module(util(file)).
:- use_module(util('env-dict')).
:- use_module(util(obj)).
:- use_module(util(web)).

term_expansion --> pac:expand_pac.
:- use_module(pac(op)).

% :- set_prolog_flag(unknown, fail).
% :- expand_file_name('~', X), log(X). % empty string.
:-op(750, yfx, &).

signature([(^)/2,(+)/2,(\)/2,(&)/2,(#)/2,(-)/1,(*)/1,reverse/1]).

%  'coa' stands for coalgebra; 'sts', state transition system;
%  am, automaton;  dfam, deterministic finite automaton
% ?- trace, coalgebra:compose_am((a+b)^(c+d), X).

% ?- qcompile(util(coalgebra)).
% ?- module(coalgebra).
% ?- am(a, X).
% ?- am(a+b, X).
% ?- am(a^b, X).
% ?- compose_am(*(a), X).

compose_am(E, C):- symbols(E, S),
	preprocess_complement(S, E, E0),
	am(E0, C0),
	remove_dead_states(C0, C).

dfam --> call, det_am, minimum_am, refresh_am.

% ?- listing(am).

am(empty,A1):-!,ampty_am(A1) .
am(epsilon,A1):-!,epsilon_am(A1) .
am(A+B,A1):-!,(am(A,A2),am(B,A3)),dfam(disj_am(A2,A3),A1) .
am(A&B,A1):-!,(am(A,A2),am(B,A3)),dfam(conj_am(A2,A3),A1) .
am(A^B,A1):-!,(am(A,A2),am(B,A3)),dfam(concat_am(A2,A3),A1) .
am(*A,A1):-!,am(A,A2),dfam(star_am(A2),A1) .
am((A\B),A1):-!,(am(A,A2),am(B,A3)),dfam(subtr_am(A2,A3),A1) .
am(#(A,B),A1):-!,(am(A,A2),am(B,A3)),dfam(xor_am(A2,A3),A1) .
am(reverse(A),A1):-!,am(A,A2),dfam(reverse_am(A2),A1) .
am(X,am([(1,[(X,[2])]),(2,[])],1,[2])):-! .


%% Standard construction of automata including
%  Kleene closure and set difference

empty_am(am([ (1,[])],1, [])).

epsilon_am(am([ (1,[])],1, [1])).

conj_am(X, Y, Z):- prod_am(conj, X, Y, Z).

disj_am(X, Y, Z):- prod_am(disj, X, Y, Z).

concat_am(am(C0,I0,F0), Am1, Am):-
	length(C0, L),
	L1 is L+1,
	refresh_am(Am1, am(C1, I1, F1), L1, _),
	product(F0, [I1], E),
	append(C0, C1, H),
	join_am(E, am(H,I0,F1), Am).

star_am(am(C,I,F), Am):-
	product(F, [I], E),
	join_am(E, am(C,I,[I|F]), Am).

subtr_am(X, Y, Z):- prod_am(subtr, X, Y, Z).

xor_am(X, Y, Z):- prod_am(xor, X, Y, Z).

reverse_am(Am0, Am) :-
	am_sts(Am0, sts(S0, I, F)),
	maplist(reverse_triple, S0, S),
	sts_coa(S, C),
	length(C, L), N is L+1,
	product([N], F, E),
	join_am(E, am([(N, [])|C], N, [I]), Am).

join_am(R, am(C0, I0, F0), am(C1, I0, F1)):- join_link(R, C0, C1),
	foldl(propagate_final_state, R, F0, F1).

%
prod_am(FinalPred, am(C0, I0, F0), am(C1, I1, F1), am(C, I, F)):-
	coalgebra([(I0, I1)], product_coa(p(C0, C1)), C2),
	refresh_coa(C2, C, 1, Assoc),
	assoc(Assoc, (I0, I1), I),
	maplist(fst, C2, S),
	call(FinalPred, S, F0, F1, F2),
	maplist(assoc(Assoc), F2, F).

conj(S, F0, F1, F):- collect(conj(F0, F1), S, F).

conj(F0, F1, (X, Y)):- memberchk(X, F0), memberchk(Y, F1).

disj(S, F0, F1, F):- collect(disj(F0, F1), S, F).

disj(F0, F1, (X, Y)):- (memberchk(X, F0);  memberchk(Y, F1)), !.

subtr(S, F0, F1, F):- collect(subtr(F0, F1), S, F).

subtr(F0, F1, (X, Y)):- (memberchk(X, F0),  \+ memberchk(Y, F1)), !.

xor(S, F0, F1, F):- collect(xor(F0, F1), S, F).

xor(F0, F1, (X, Y)):-
	(  memberchk(X, F0),  \+ memberchk(Y, F1)
	;  \+ memberchk(X, F0),  memberchk(Y, F1)
	),
	!.

product_coa(X, Y, Z, U):-product_coa(X, Y, Z),
	union_image(Z, U).

product_coa(_, (0, 0), []):-!.
product_coa(p(_, C1), (0, X1), Y):- !, memberchk((X1, Y1), C1),
	maplist(pairing_fst(0), Y1, Y).
product_coa(p(C0, _), (X0, 0), Y):- !, memberchk((X0, Y0), C0),
	maplist(pairing_snd(0), Y0, Y).
product_coa(p(C0, C1), _, []):- (C0 = []; C1=[]), !.
product_coa(p(C0, C1), (X0, X1), Y):-
	memberchk((X0, Y0), C0),
	memberchk((X1, Y1), C1),
	product_coa0(Y0, Y1, Y).

product_coa0([], R1, R):-!, maplist(pairing_fst(0), R1, R).
product_coa0(R0, [], R):-!, maplist(pairing_snd(0), R0, R).
product_coa0([(X,[Y0])|R0], [(X,[Y1])|R1], [(X, [(Y0, Y1)])|R]):-!,
	product_coa0(R0, R1, R).
product_coa0([(X0,[Y0])|R0], [(X1,[Y1])|R1], [(X0, [(Y0, 0)])|R]):-
	X0 @< X1, !, product_coa0(R0, [(X1,[Y1])|R1], R).
product_coa0([(X0,[Y0])|R0], [(X1,[Y1])|R1], [(X1, [(0, Y1)])|R]):-
	product_coa0([(X0,[Y0])|R0], R1, R).

%% make deterministic
det_am(am(C, I, F), Am):-
	pow_coa(I, C, D0),
	maplist(fst, D0, S0),
        collect(meet(F), S0, F0),
	refresh_dfam(am(D0, [I], F0), Am).

%%% coalgebra/3 (generalized sub_coa, inspired by `F-coalgebra')
coalgebra(X, Y, Z) :- coalgebra(X, Y, [], Z0), sort(Z0, Z).

coalgebra([], _, X, X).
coalgebra([N|Ns], Generator, Y, Z) :- call(Generator, N, U, Sucs), !,
	Y1 = [(N, U) | Y],
	new_nodes(Sucs, Y1, D1),
	union(D1, Ns, Ms),
	coalgebra(Ms, Generator, Y1, Z).

%% Note that automata are coalgebras !
%  sub_coa(X, C, Z, U).
% X: subset of domain(C).
% C: coalgebra
% Z: restriction of U.
% U: the minimum subcoalgebra of C over X.
%
% ?- sub_coa([1], comma_set([(1,[(a,1),(b,1)]), (2,[(a, 1), (b,2)])]), [], X).
% X = [ (1, [ (a, 1), (b, 1)])]
% ?- sub_coa([[1]],  pow_merge(eq_set([ 1 = [ (a,[2]), (b,[1,2]) ], 2 = [(a,[1]),(b,[1])]])), X).
% X = [ ([1, 2], [ (a, [1, 2]), (b, [1, 2])]), ([2], [ (a, [1]), (b, [1])]), ([1], [ (a, [2]), (b, [1, 2])])]

sub_coa(X, Y, Z) :- sub_coa(X, Y, [], Z0), sort(Z0, Z).

sub_coa([], _, X, X).
sub_coa([N|Ns], Generator, Y, Z) :- call(Generator, N, U), !,
	Y1 = [(N, U) | Y],
	maplist(snd, U, D),
	new_nodes(D, Y1, D1),
	union(D1, Ns, Ms),
	sub_coa(Ms, Generator, Y1, Z).

% ?- sts_coa(1, [(1,a,2), (1, b, 1), (2, a, 1)], P).
% ?- sts_coa(1, [(1,a,2), (1, b, 1), (1, b, 2), (2, a, 1)], P).
% P = [ ([1, 2], [ (a, [1, 2]), (b, [1, 2])]), ([2], [ (a, [1])]), ([1], [ (a, [2]), (b, [1, 2])])]
% ?- sts_coa([(1,a,2),(1, a, 1), (1, b, 1), (2, a, 1)], Ts), pow_coa(1, Ts, P).
% P = [ ([1], [ (a, [1, 2]), (b, [1])]), ([1, 2], [ (a, [1, 2]), (b, [1])])]
% ?- sts_coa([(1,a,2), (1, b, 1), (1, b, 2)], P).
% P = [ (1, [ (a, [2]), (b, [1, 2])]), (2, [])]

sts_coa(S0, Ts, P) :- sts_coa(Ts, C),
	sub_coa([[S0]], pow_merge(comma_set(C)), P).

sts_coa(Ts, C):- maplist(triple_pair, Ts, Ts0),
	foldl(extend_coa, Ts0, [], C0),
	range_of_triples(Ts, F),
	maplist(pair_with_empty, F, Ps),
	merge_coa(Ps, C0, C).

% ?- sts_am(sts([(1, a, 1), (1, a, 2), (1, b, 2), (2, a, 3)], 1, [1]), D).
% D = coa([ (1, []), (2, [ (a, 1)]), (3, [ (a, 3), (b, 2)]), (4, [ (a, 3), (b, 2)]), (5, [ (a, 4), (b, 2)])], 5, [3, 4, 5])

sts_am(sts(T, I, F), Am):- sts_coa(T, C), det_am(am(C, I, F), Am).

%% power automata

pow_coa(S0, Ts, P) :- coalgebra([[S0]], pow_image(comma_set(Ts)), P).

pow_image(Gen, Xs, R, Ys):- maplist(Gen, Xs, R0),
	foldl(foldl(extend(contract_merge)), R0, [], R),
	maplist(snd, R, Ys).

pow_merge(A, X, R):- maplist(A, X, R0),
	foldl(foldl(extend(contract_merge)), R0, [], R).

new_nodes([], _, []).
new_nodes([N|Ns], L, Ms) :- member((N, _), L)
	->  new_nodes(Ns, L, Ms)
	;   Ms=[N|Ms0], new_nodes(Ns, L, Ms0).

% ?- extend_coa((a, [(1,[2])]), [(a,[(2,[4])])], X).
% X = [ (a, [ (1, [2]), (2, [4])])]

merge_coa(X, Y, Z):- foldl(extend_coa, X, Y, Z).

extend_coa(X, Y, Z):- extend(merge_indexed_family_of_sets, X, Y, Z).

% ?- merge_indexed_family_of_sets([(1,[2]), (2,[3])], [(2,[4])], X).
% X = [ (1, [2]), (2, [4, 3])]

merge_indexed_family_of_sets(X, Y, Z):- foldl(extend(contract_merge),X, Y, Z).

extend(F, (X, S0), [(X, S1)|R], [(X, S2)|R]):-!, call(F, S0, S1, S2).
extend(F, (X, S0), [(Y, S1)|R0], [(Y, S1)|R]):-  X@>Y, !,
	extend(F, (X, S0), R0, R).
extend(F, (X, S0), R, [(X, S)|R]):- call(F, S0, [], S).


% ?- coalgebra:contract_merge([1,3,6], [2, 5, 7], R).
%@ R = [1, 2, 3, 5, 6, 7].

%  Remark.
% ?- union([1,3,6], [2, 5, 7], R).
%@ R = [1, 3, 6, 2, 5, 7].
% ?- union([1,3,6], [2, 3,  5, 7], R).
%@ R = [1, 6, 2, 3, 5, 7].

contract_merge([], X, X):-!.
contract_merge(X, [], X):-!.
contract_merge([X|R], [X|S], [X|T]):- !, contract_merge(R, S, T).
contract_merge([X|R], [Y|S], [X|T]):- X@<Y, !, contract_merge(R, [Y|S], T).
contract_merge([X|R], [Y|S], [Y|T]):- contract_merge([X|R], S, T).


% ?- coalgebra:contract_insert(3, [2, 5, 7], R).
%@ R = [2, 3, 5, 7].

contract_insert(X, [], [X]):-!.
contract_insert(X, [X|S], [X|S]):- !.
contract_insert(X, [Y|S], [X, Y|S]):- X@<Y, !.
contract_insert(X, [Y|S], [Y|R]):- contract_insert(X, S, R).


%%%%  refresh coalgebra by renaming states
refresh_dfam(A, B) :- refresh_dfam(A, B, 1,  _).

refresh_dfam(am(A,I,F), am(B,J,G), N, A_list) :-
	make_assoc(A, N, A_list),
	maplist(refresh_state_dfam(A_list), A, B),
	maplist(assoc(A_list), [I|F], [J|G]).


refresh_state_dfam(A, (X, S), (Y, T)):- assoc(A, X, Y),
	maplist(refresh_right_dfam(A), S, T).

refresh_right_dfam(Assoc, (A, S), (A, [S1])) :- assoc(Assoc, S, S1).

%%%%
refresh_am(X, Y):- refresh_am(X, Y, 1, _).

refresh_am(am(A, B, C), am(A0, B0, C0), N, Assoc):-
	refresh_coa(A, A0, N, Assoc),
	maplist(assoc(Assoc),[B|C], [B0|C0]).

refresh_coa(A, B) :- refresh_coa(A, B, 1,  _).

refresh_coa(A, B, N0) :- refresh_coa(A, B, N0, _).

refresh_coa(A, B, N, A_list) :- make_assoc(A, N, A_list),
	maplist(refresh_state(A_list), A, B).

refresh_state(A, (X, S), (Y, T)):- assoc(A, X, Y),
	maplist(refresh_pair(A), S, T).

refresh_pair(Assoc, (A, S), (A, S1)) :- maplist(assoc(Assoc), S, S1).

% ?- join_link([(a, b)], [(a, []), (b, [(1, [a])])], C).
% C = [ (a, [ (1, [a, b])]), (b, [ (1, [a, b])])]
% ?- join_link([ (a, b), (b, a)], [ (a, []), (b, [ (1, [a])])], C).
% C = [ (a, [ (1, [a])]), (b, [ (1, [a])])]
% ?- join_link([ (a, b), (b, a)], [ (a, [(1,[b])]), (b, [ (1, [a])])], C).
% C = [ (a, [ (1, [a, b])]), (b, [ (1, [a, b])])]

join_link(R, C0, C):- foldl(extend_link, R, C0, C).

extend_link(P) --> extend_link_right(P),  extend_link_left(P).

extend_link_right((X,Y), C0, C):- maplist(extend_link_right(X, Y), C0, C).

extend_link_right(X, Y, (A, F0), (A, F)):- maplist(associate(X, Y), F0, F).

associate(X, Y, (A, S0), (A, S)):- memberchk(X, S0)
	-> contract_merge([Y], S0, S)
	;  S = S0.

extend_link_left((X, Y), C0, C):- memberchk((Y, F), C0),
	extend_coa((X, F), C0, C).

propagate_final_state((X, Y), F0, F1):- memberchk(Y, F0), !, union([X], F0, F1).
propagate_final_state(_, F, F).

%%% automata state minimization
% ?- coalgebra:minimum_am(am([ (1, [ (a, [2])]), (2, [ (b, [3])]), (3, [])], 1, [3]), A).
% A = am([ (1,[ (a,[2])]), (2,[ (b,[3])]), (3,[])],1,[3]).

% ?- qcompile(util(coalgebra)), module(coalgebra).
% ?- listing(hybrid_regex_html/3).

% ?- call(coalgebra:hybrid_regex_html,*char(alnum)+ +char(digit),svg,_25662).
% ?- call(coalgebra:hybrid_regex_html,*char(alnum)+ +char(digit),pdf,_25662).

minimum_am(am(C0, I, Final), Am):-
    C = [(0,[])|C0],
	maplist(fst, C, S),
	pairs(S, P0),
	maplist(ordered_pair(0), Final, DeadStateConflict0),
	basic_conflicts(P0, [], Final, P1, Q),
	append(DeadStateConflict0, Q, Q1),
	remove_conflicts(P1, Q1,  C, P, _),
	maplist(pred([X,[X]]), S, Bs0),
	union_find(P, Bs0, Bs1),
	remove_dead_state(Bs1, Bs),
	quotient_am(am(C0,I,Final), Bs, Am).

% ?- basic_conflicts([(1,2),(2,3)], [], [3], X, Y).
% X = [ (1, 2)], Y = [ (2, 3)]

basic_conflicts([],[], _, [],[]).
basic_conflicts([(X,Y)|P0], Q0, F, P, [(X,Y)|Q]):-
	(  memberchk(X, F)
	-> \+ memberchk(Y, F)
	;  memberchk(Y, F)
	),
	!,
	basic_conflicts(P0, Q0, F, P, Q).
basic_conflicts([A|P0], Q0, F, [A|P], Q):- basic_conflicts(P0, Q0, F, P, Q).

remove_conflicts(P0, Q0, C, P, Q):- select((X,Y), P0, P1),
	conflict(X, Y, Q0, C),
	!,
	remove_conflicts(P1, [(X,Y)|Q0], C, P, Q).
remove_conflicts(P, Q,  _, P, Q).

conflict(X, Y, Q, C):- memberchk((X, F), C),
	memberchk((Y, G), C),
	conflict(F, G, Q).

conflict(F, G, Q):- select((X,S1), F, F1), select((X,S2), G, G1), !,
	(  pair_member(S1, S2, Q) -> true ;  conflict(F1, G1, Q) ).
conflict(F, G,Q):-(member((_X,S),F); member((_X,S),G)),
	member(Y, S), pair_member0(Y, 0, Q).

pair_member(V, W, Q):- member(X, V), member(Y, W), pair_member0(X, Y, Q).

pair_member0(X, Y, [(X, Y)|_]).
pair_member0(X, Y, [(Y, X)|_]).
pair_member0(X, Y, [_|Q]):- pair_member0(X, Y, Q).

% ?- union_find([(a,b),(x,y), (x,x), (y, z), (b,c)], [], R).
%@ R = [[a,b,c],[x,y,z]].
% ?- union_find([(a-b),(x-y), (x-x), (y-z), (b-c)], [], R).
%@ R = [[a,b,c],[x,y,z]].

union_find([],X,X).
union_find([P|R],C,D):- (P = (X,Y); P = (X-Y)), !,
	union_find(X, Y, C, C1),
	union_find(R,C1,D).

%
union_find(X,Y,Z,U):-find_cluster(X,Z,C,Z0),
	(memberchk(Y, C) -> U=[C|Z0]
	; find_cluster(Y, Z0, C0, Z1),
	  append(C,C0, C1),
	  U=[C1|Z1]
	).

%
remove_dead_state([],[]).
remove_dead_state([[0|X]|R],[X|R]):-!.
remove_dead_state([X|R],[X|S]):-remove_dead_state(R,S).

%% post processing

remove_dead_states(am(M0,I,F), am(M, I, F)):- 	select((X, Ps), M0, M1),
	\+ memberchk(X, F),
	dead_state(X, Ps),
	!,
	maplist(delete_target(X), M1, M).
remove_dead_states(A, A).

delete_target(X, (Y,Ps), (Y,Qs)):- maplist(delete_target0(X), Ps, Qs).

delete_target0(X, (Y,Ps), (Y,Qs)):-select(X, Ps, Qs), !.
delete_target0(_, Z, Z).

dead_state(X, Ps):- forall(member((_,Xs), Ps), (Xs ==[]; Xs=[X])).

% quotient dfam

quotient_am(am(C0,I0,F0), Bs, am(C, I, F)):-
	maplist(quotient_state(Bs), C0, C1),
        sort(C1, C2),
	remove_duplicates(C2, C),
	quotient_map(Bs, I0, I),
	maplist(quotient_map(Bs), F0, F1),
	sort(F1, F).

remove_duplicates([], []).
remove_duplicates([(X,Y),(X,_)|C0], C):-!, remove_duplicates([(X,Y)|C0], C).
remove_duplicates([(X,Y)|C0], [(X,Y)|C]):- remove_duplicates(C0, C).

quotient_map(Bs, X, A):- member(B, Bs), memberchk(X, B), B=[A|_], !.

quotient_state(Bs, (X, S), (Y, T)):- quotient_map(Bs, X, Y),
	maplist(quotient_snd(Bs), S, T).

quotient_snd(Bs, (A, X), (A, Y)):- maplist(quotient_map(Bs), X, Y0),
	sort(Y0, Y).

%
compare_two_regex(A:B, C):- regex_compare(C, A, B).

%
compare_am(R0, R, E0, E):- compose_am(E0, am(C0, I0, F0)),
	compose_am(E, am(C, I, F)),
	coalgebra([(I0, I)], product_coa(p(C0, C)), D),
	compare_am(R0, R, D, F0, F).

%
compare_am(R, R, [], _, _):-!.
compare_am(R0, R, [((X,Y),_)|D], F, G):-
	(    memberchk(X, F)
	-> (  memberchk(Y, G)
	   -> R1=R0
	   ;  subtract(R0, [=, <], R1)
	   )
	; ( memberchk(Y, G)
	   -> subtract(R0, [=, >], R1)
	   ;  R1=R0
	  )
	),
        compare_am(R1, R, D, F, G).

%
order_boole(=, R, true):- memberchk(=, R).
order_boole(<, [<], true).
order_boole(>, [>], true).
order_boole(>=, R, true):- memberchk(>, R); memberchk(=, R).
order_boole(=<, R, true):- memberchk(<, R); memberchk(=, R).
order_boole(_, _, false):- memberchk(<, _R); memberchk(=, _R).

%%%%  tiny routines (selector, etc.) %%%%
%%

&(X, Y, Z):- append(X, Y, Z).

% ?- find_cluster(a, [[a,b],[c,d]], C, X).
find_cluster(X,[],[X],[]):-!.
find_cluster(X,[Y|Z],Y,Z):- memberchk(X,Y),!.
find_cluster(X,[Y|Z],U,[Y|V]):- find_cluster(X,Z,U,V).


% ?- field_of_triples([(c, 2, d), (b,1,a)], X).
% X = [a, b, c, d]
field_of_triples(Ts, F):- unzip(Ts, L, G),
	maplist(snd, G, R),
	append(L, R, F0),
	sort(F0, F).

range_of_triples(Ts, R):- maplist(third, Ts, R0), sort(R0, R).

make_assoc([],_,[]).
make_assoc([(X,_)|R], N, [(X,N)|S] ):- N1 is N+1, make_assoc(R, N1, S).

assoc(A, X, Y):- memberchk((X,Y), A).

meet(X, Y):- member(A, X), memberchk(A, Y), !.

third((_,_,X), X).

triple_pair((X,A,Y), (X,[(A,[Y])])).

pair_with_empty(X, (X, [])).

reverse_triple((X,A,Y), (Y,A,X)).

eq_set( F, X, Y) :- member(X = Y, F).

comma_set(F, X, Y) :- memberchk((X, Y), F), !.
comma_set(_, _, []).

pairing_fst(A, (X, [Y]), (X, [(A,Y)])).

pairing_snd(A, (X, [Y]), (X, [(Y,A)])).

union_image(X, Y):- foldl(union_snd, X, [], Y).

union_snd((_, X), A, B):- union(X, A, B).

%;; (setq module-query  "qcompile(util(coalgebra)), module(coalgebra).")
% ?- qcompile(util(coalgebra)), module(coalgebra).

% directory for temporary files  .pdf .png .bb  etc
am_dir_path(D):- getenv(home_html_root, C),
		atomics_to_string([C, automata], /, D).
%
am_file_name(am).
%
compose_sts(E, Sts):- compose_am(E, Am0), am_sts(Am0, Sts).

am_sts(am(M0,X,Y), sts(M,X,Y)):-
	act_tree(coalgebra:list_triple, M0, M).

list_triple([Y, A, X], (X, A, Y)).

%%%
% ?- coalgebra:arrows([(b, g, c), (a, f, b), (c, h, d), (d, k, b), (c, c, c)]).
arrows(Arrows):-autviz_jpg(arrows2dot, Arrows).

% ?- qcompile(util(coalgebra)).
% ?- module(coalgebra).

% ?- trace, gv(a).
% ?- gv(a+b+c+d+e).

gv(R) :- compose_sts(R, Aut), autviz(Aut).

autviz(Aut):-autviz(coalgebra:aut2dot, Aut).

%
autviz(Pred, Aut):-
	absolute_file_name(tmp('DOTTEMP.dot'), DOT),
	absolute_file_name(tmp('DOTTEMP.'), M),
	file(DOT, write, call(Pred, Aut)),
	qshell(dot(-'T'(ps2), -o(M+ps), M+dot);
			ps2pdf(M+ps, M+pdf);
			open(-a('Preview'), M+pdf)).

autviz_jpg(Pred, Aut):-
	absolute_file_name(tmp('DOTTEMP.dot'), DOT),
	absolute_file_name(tmp('DOTTEMP.'), M),
	file(DOT, write, call(Pred, Aut)),
	qshell(dot(-'T'(jpg), -o(M+jpg), M+dot); open(-a('Preview'), M+jpg)).

states(M, S):- 	maplist(fst, M, L0),
	maplist(third, M, L1),
	union(L0, L1, S0),
	sort(S0, S).

automaton(R):- compose_sts(R, sts(M,Ini,Fin)),
	states(M, S),
	format("Regular expression = ~w~n",[R]),
	format("Initial state = ~w~n",[Ini]),
	format("List of final states = ~w~n",[Fin]),
	format("List of states = ~w~n",[S]),
	format("State transitions:~n"),
	(  member((A,B,C), M),
	   format("    ~w ----- ~w -----> ~w~n",[A,B,C]), fail
	;
	   true
	).

%
aut2dot(sts(M,Ini,Fin)):-arrows2dot([(dummy, '', Ini)|M], Fin).

arrows2dot(Arrows):-arrows2dot(Arrows, []).

arrows2dot(Arrows, Fin):-
     format("digraph g {~n",[]),
     format("rankdir=LR;~n",[]),
     format("dummy [shape = none, label = \"\"];~n",[]),
     final_in_dot(Fin),
     move_in_dot(Arrows),
     format("}~n",[]).
		/******************************************************
		*     convert a  regex to automata  in a pdf file.    *
		******************************************************/

% ?- regex_coalgebra(".*", X).
% ?- coalgebra:show_am(".").
% ?- coalgebra:show_am("[a-zA-B]******hello").
% ?- coalgebra:show_am("a*").
% ?- coalgebra:show_am(".*").
% ?- coalgebra:show_am("a*b*c").
% ?- coalgebra:show_am("([^abc]+[abc])*").
% ?- coalgebra:show_am("[st][0-9][0-9][0-9][0-9][0-9][0-9]*").
% ?- coalgebra:show_am("(a|b|c)*").
% ?- coalgebra:show_am(a\a).
% ?- coalgebra:show_am((*(*(a) + b + *(a) + b + *(a)))).
% ?- pac:show_am((*(*(a) + b + *(a) + b + *(a)))).
% ?- coalgebra:show_am(* char(alnum)).

show_am(Regex) :- regex_to_pdf(Regex, PDF),	qshell(open(PDF)).

% ?- regex_to_pdf("a", "~/Desktop/deldel", PDF).
regex_to_pdf(Regex, PDF) :- once(regex_to_pdf(Regex, 'DOTTEMP', PDF)).
%
regex_to_pdf(Regex, Base, PDF):- once(regex_am(Regex, coa(A, I))),
	am_finals(coa(A, I), F),
	coalgebra_triples(A, B),
	automaton_quasi_string(am(B, I, F), Quasi_String),
	expand_file_name(Base, [M]),
	atomics_to_string([M, ".dot"], DOT),
	atomics_to_string([M, ".pdf"], PDF),
	file(DOT, write, smash(Quasi_String)),
	qshell(dot(-'T'(pdf), -o(PDF), DOT)).

			/***********************************
			*     handling character escape    *
			***********************************/

automaton_quasi_string(am(A, Ini, F),
	      [		"digraph g {\n",
			"rankdir=LR;\n",
			IniName,
			Label,
			Finals,
			Moves,
			"}\n"
	      ])  :-
	number_string(Ini, IniName),
	atomics_to_string([" [label= \"start( =", IniName, ")\"];\n"], Label),
	maplist(triple_elim_dot, A,  Moves),
	maplist( pred([X, [Y, " [shape = doublecircle];\n"]] :-
		number_string(X, Y)),
		F,
		Finals).

% ?- trace, coalgebra:triple_elim_dot((1, dot([97-98, 100-103]), 2), R).
%@ R = ["1", "->", "2", " [label = \"", [91, 97, 93], "\"];\n"] .

triple_elim_dot((X, A, Y), [X0, "->", Y0, " [label = ", B, "];\n"])
		:-
		 number_string(X, X0),
		 number_string(Y, Y0),
		 adjacent_interval(A, OA, []),
		 maplist(interval_exp, OA, A0),
 		 term_string(A0, B0, [quoted(false)]),
		 term_string(B0, B, [nl(false)]).

% ?- coalgebra:simplify_interval([inf-97, 99-sup], X).
%@ X = [\=(b)] .
% ?- coalgebra:simplify_interval([inf-97, 100-sup], X).
%@ X = [=<(a), >=(d)] .

%
interval_exp(x(X, _), Y):- atom_string(X, Y).
% interval_exp(x(X), Y):- atom_string(X, Y).
interval_exp(X, Y):- interval_code_char(X, Y), !.
interval_exp((inf-sup)\W,  !=(W0))	:- code_to_char_x(W, W0).
interval_exp(X\W,  Y\W0):- interval_code_char(X, Y),
	code_to_char_x(W, W0).

%
interval_code_char(x(X, _), Y):- atom_string(X, Y).
interval_code_char(inf-sup, '.').
interval_code_char(inf-A, =<(A0)):- code_to_char_x(A, A0).
interval_code_char(A-sup, >=(A0)):- code_to_char_x(A, A0).
interval_code_char(A-A, A0):- code_to_char_x(A, A0).
interval_code_char(A-B, A0-B0):- code_to_char_x(A, A0),
	code_to_char_x(B, B0).

%
code_to_char_x(x(X, _), Y):- !, atom_string(X, Y).
code_to_char_x(X, Y):- char_code(Y0, X),
	convert_char(Y0, Y).
%
printable(A):- member(T, [csym, period, punct, prolog_symbol]),
	char_type(A, T),
	!.
%
convert_char('\n', "\\n").
convert_char('\t', "\\t").
convert_char('\s', "\\ ").
convert_char('.', "\\.").
convert_char(A, A):- printable(A), !.
convert_char(A, B):- char_code(A, C),
	number_string(C, B0),
	string_concat("code", B0, B).

%
coalgebra_triples([], []).
coalgebra_triples([A-G|R], Triples):-
	make_triples(A, G, T),
	coalgebra_triples(R, TR),
	append(T, TR, Triples).

%
make_triples(_, [], []).
make_triples(A, [B-C|R],[ (A, B, C)|Ts]) :- make_triples(A, R, Ts).
make_triples(A, [_|R], Ts):- make_triples(A, R, Ts).
%

final_in_dot([]).
final_in_dot([X|Y]):-format("~w [shape = doublecircle];~n",[X]),
	final_in_dot(Y).

move_in_dot([]).
move_in_dot([(X,A,Y)|Z]):-edge_in_dot(X,A,Y), move_in_dot(Z).

edge_in_dot(X,A,Y):- format("\"~w\" -> \"~w\" [label = \"~w\"];~n",[X,Y,A]).

% main predicate
% generete an automaton for the given regular expression.

reg2html(X,T,URL):-
	am_dir_path(D),				% T: output type
	am_file_name(F),
	phrase((counter(update), obj_get([count(C)])),
	       [directory(D), counter_name(atm_cur_id)],  _),
	format(codes(OutHtml), "~w/~w~w.html", [D, F, C]),
	reg2html(F, D, C, X, T, Body),
	flatten(["<html><body>\n", Body, "</body></html>"], H1),
	atom_codes(OutNameAtom, OutHtml),
	create_file(OutNameAtom, H1),
	expand_cgi_path(OutHtml, URL).

reg2html(F,D,C,R,T,H) :-  compose_sts(R, sts(M,Ini,Fin)),
	states(M, S),
	H1 = (format_codes(`<p>Regular expression = ~w</p>`,[R]) &
	format_codes(`<p>Initial state = ~w</p>`,[Ini]) &
	format_codes(`<p>Final states = ~w</p>`,[Fin]) &
	format_codes(`<p>All states = ~w</p>`,[S]) &
	format_codes(`<p>State Transitions:</p>~n`,[])),
%
	atomic_list_concat([D, '/', F, C, '.dot'], DotName),
	file(DotName, write, coalgebra:aut2dot(sts(M,Ini,Fin))),
	formatForHtml(F,D,C,T,Format,Args),
	print_moves(M, M1),
	format_codes_list(H1 & format_codes(Format,Args) & M1, L, []),
	!,
	append(L, H).


% ?- apply3(coalgebra:hybrid_regex_html, [*char(alnum)+ +char(digit),svg], R).

% char class automata  [2014/06/02]
hybrid_regex_html(X, T, URL):-	% T: output type; svg/pdf
	am_dir_path(D),
	am_file_name(F),
	phrase((counter(update), obj:obj_get([count(C)])),
	       [directory(D), counter_name(atm_cur_id)],  _),
	format(string(HtmlLoc), "~w/~w~w.html", [D, F, C]),
	hybrid_regex_html(F, D, C, X, T, Body),
	smash(Body, BodyStr),
	atomic_list_concat(["<html><body>\n", BodyStr, "</body></html>"], HtmlTag),
	create_file(HtmlLoc, HtmlTag),
    getenv(host_html_root, R),
	atomic_list_concat([R, /, automata, /, F, C, '.html'], URL).


% ?- smash0([2, b, c], X).
%@ Correct to: "smash([2,b,c],X)"? yes
%@ X = "\u0002bc" .
hybrid_regex_html(F, D, C, R, T, H) :-
	regex_am(R, coa(A, Ini)),
	am_finals(coa(A, _), Fin),
	length(A, Num),
 	coalgebra_triples(A, M),
 	automaton_quasi_string(am(M, Ini, Fin), Quasi_String),
	maplist(pred([U-_, U]), A, S),
	H1 = (format_codes(`<p>Regular expression = ~w</p>`,[R]) &
	format_codes(`<p> The number of states = ~d</p>`,[Num]) &
	format_codes(`<p> Initial state = ~w</p>`,[Ini]) &
	format_codes(`<p> Final states = ~w</p>`,[Fin]) &
	format_codes(`<p> All states = ~w</p>`,[S]) &
	format_codes(`<p> State Transitions:</p>~n`,[])),
	atomic_list_concat([D, '/', F, C, '.dot'], DotName),
 	file(DotName, write, smash(Quasi_String)),
	formatForHtml(F,D,C,T,Format,Args),			% call sh from inside
	hybrid_print_moves(M, M1),
	format_codes_list(H1 & format_codes(Format,Args) & M1, L, []),
	!,
	append(L, H).

%
format_codes_list(X&Y, L, M):- !, format_codes_list(X, L, L0),	format_codes_list(Y, L0, M).
format_codes_list([], L, L).
format_codes_list([X|Y], [[X|Y]|L], L).
format_codes_list(P, [V|L], L):- call(P, V).

%
option_table(pdf, ps2, ps, pdf).
option_table(png, 'png:gd', png, png).
option_table(X, X, X, X).

formatForHtml(F,D,C,X,Format,[F,C,X]) :-
        img_frame(Format),
        atomic_list_concat([D, (/), F, C], Base),
	once(option_table(X, Opt, Ext0, Ext1)),
     (  X == pdf
     -> Com = ps2pdf(`-sOutputFile=`+ Base+ `.`+ Ext1,
		     Base+ `.`+ Ext0)
     ;  Com = 'DUMMY=1'
     ),
         once(qshell(dot(-'T'(Opt), Base+ `.` + dot,
		  -o(Base + `.` + Ext0)); Com)).

img_frame(X) :- flatten([ `<p><div `,
	  `id='diagram' `,
	  `style='border : solid 2px #ff0000; `,
	  `width : 1600px; `,			% was 600px
	  `height : 500px; `,			% was 300px
	  `overflow : auto; '><br/>`,
	  `<img src="~w~w.~w"/>`,
	  `</div></p>~n`
	 ], X).

% ?- coalgebra:hybrid_print_moves([(1, [97-97], 2)], R).

hybrid_print_moves(M, H):-
	maplist(pred([(X, A, Y), (X, B, Y)]:-
	       maplist(interval_code_char, A, B)),
		M, M0),
	print_moves(M0, H).

%
print_moves(M, H):-
	maplist(print_moves_x, M, T),
	flatten(T, T1),
	flatten(["<pre>\n", T1, "</pre>\n"], H).

print_moves_x((X,A,Y), H):-
	format_codes(`   ~w----~w--->~w~n` , [X,A,Y], H).

%  ?- symbols(a + b, X).
%  ?- symbols(a + b, X)
%  X = [a, b].

symbols(A, S) :- collect_symbols(A, S0), sort(S0, S).

symbols_(A, [A]):- atomic(A), !.
symbols_(E, S):- E =..[Op|As],
		length(As,L),
	        signature(Sig),
		memberchk(Op/L, Sig),
		maplist(symbols_, As, Bs),
		append(Bs, S).
symbols_(A = [A]).

% ?- pac:expand_exp('`'([]), a, V, [], G, P, []).

% ?- trace, pac: expand_kind_rule(s, [], [], epsilon = '`'([]), Y).
%@ Y =  (s(epsilon, []):-!) .

% ?- spy(ekind).
%@ % Spy point on pac:ekind/0
%@ true.



collect_symbols(epsilon,[]):-! .
collect_symbols(empty,[]):-! .
collect_symbols(E,A1):-(E=..[Op|As],length(As,L),L>0,signature(Sig),memberchk(Op/L,Sig)),!,maplist(collect_symbols,As,A2),append(A2,A1) .
collect_symbols(A,[A]):-! .


% ?- coalgebra:ya_symbols(a, X).
% ?- listing(preprocess_complement).

preprocess_complement(A, X, Y):-
	( A== []
	->	U = epsilon
	;	vector_term(+, A, A0),
		U= *(A0)
	),
	elim_unary_minus(U, X, Y).

%
elim_unary_minus(A, -(B), A\C):- !, elim_unary_minus(A, B, C).
elim_unary_minus(A, B, C):- compound(B), B =..[Op|Bs],
		length(Bs,L),
		L>0,
	        signature(Sig),
		memberchk(Op/L, Sig),
		!,
		maplist(elim_unary_minus(A), Bs, Cs),
		C =..[Op|Cs].
elim_unary_minus(_, B, B).

% ?- gv(a).

% ?- vector_term(+, [a,b,c], V).
% ?- coalgebra:elim_unary_minus(epsilon, a, X).
%@ X = a.
% ?- coalgebra:elim_unary_minus(universe, -(a + (-b)), X).
%@ X = (universe\a+ (universe\b)).

complement_subtract(U, -X, U\X).

is_unary_minus_term(-(_)).
