:- module(ejockey, []).

:- use_module(pac(basic)).
:- use_module(pac(reduce)).
:- use_module(pac(meta)).
:- use_module(pac('pac-listing')).
:- use_module(util(misc)).
:- use_module(util('work-command')).
:- use_module(util(snippets)).
:- use_module(util('term-string')).
:- use_module(util(file)).
:- use_module(util('prolog-elisp')).
:- use_module(util('emacs-handler')).
:- use_module(util('swap-args')).
:- use_module(util(obj)).
:- use_module(pac(op)).
:- op(1200, xfx, -->>).

flip(F, X, Y):- call(F, Y, X).

:- discontiguous handle/3.  % [2016/01/28]

term_expansion --> pac:expand_pac.

			/***********************
			*     Emacs Handler    *
			***********************/

%
handle([halt]) --> {halt}.
%
handle([count, paragraph]) -->region,
	  paragraph,
	  remove([]),
	  length,
	  number_codes.

		/*************************
		*     sed-like usage.    *
		*************************/

% Example.
% wl("(..)") >> ["hello"]
% abcd

% Example.
% (w(".*", A), wl("b+")) >> pred(A, [A])
% aaabbcccbdddbeee

handle([sed|Optional]) --> region,
	   phrase((wl("[^\n]*", SedCommandText), wl("\n+"))),
		{	herbrand(_, SedCommandText, SedPhrase),
			let_sed(Sed, SedPhrase)
		},
		call(Sed),
		optional_overwrite(Optional).

%! handle(+Action:[let, sed, Global]) is det.
%	nb_setval(sed, S), where S is the sed phrase,
%   which is converted from the region.
handle([let, sed, Global]) --> region, current(Region),
		{	herbrand(_, Region, SedPhrase),
			let_sed(Sed, SedPhrase),
			nb_setval(Global, Sed)
		},
		clear.

%! handle(+Action:[apply, sed, S|Optional]) is det.
%	Apply the sed action stored in global S to
%	the region. Optional buffer action is append/overwrite
%	like for other handles.

handle([apply, sed, S|Optional]) --> region,
	{	nb_getval(S, Sed) },
	call(Sed),
	optional_overwrite(Optional).
%
handle([one, line]) --> region,
		sed(wl("[\n\s\t]+") >> "\s").
%
handle([pldoc, action]) -->
	region,
	split,
	maplist(insert_plus_action),
	insert("\n"),
	overwrite.

%
insert_plus_action --> "%c", w("[\s\t]*"),
					   "handle(",
					   peek(X, ["%%\thandle(+Action:", X]).
insert_plus_action --> [].

%!	handle(+Action:[meta, handle]) is det.
%  Interprete the first line of the region
%  as the handle call with the rest of the region
%  as an argument.

handle([meta, handle]) --> region,
	handle_parse_eval(_, _),
	peek(R, ["===>\n", R]).

%!	handle(+Action:[meta, handle, overwrite]) is det.
%   `Overwrite' version  of handle([meta, handle]).

handle([meta, handle, overwrite]) --> region,
	handle_parse_eval(Tag_codes, _),
	peek(R, [Tag_codes,".\n", R]),
	overwrite.

handle_parse_eval(Tag, Rest, X, Y) :-
	 append(Tag, [0'., 0'\n | Rest], X), !,
	 string_codes(S, Tag),
	 term_string(H,  S),
 	 expand_arg(H, [], H0, Aux, []),
	 maplist(assert, Aux),
	 meta_handle(H0, ejockey, modify_handle, [Rest, Y]).

%
modify_handle(overwrite(X, X), _, [], true).
modify_handle(overwrite, _, [X, X], true).
modify_handle(handle(U), M, [X, Y], BodyH):-
	clause(M:handle(U, X, Y), BodyH),
	!.
modify_handle(handle(U, X, Y), M, [], BodyH):-
	clause(M:handle(U, X, Y), BodyH),
	!.

%!	handle(+Action:[load, buffer]) is det.
%   Reload the file at the current buffer,
%   dropping the suffix "<..>" if exists.
handle([load, buffer]) -->
	{	load_buffer(Name),
		message([Name, " reconsulted."])}.

load_buffer(Y) :-
	lisp(list('default-directory', 'buffer-name'()), [X, Y]),
	atomics_to_string([X, /, Y], Z),
	(	exists_file(Z) -> Z0 = Z
	;	sub_string(Z, _, _, 1, ">"),
		sub_string(Z, J, 1, _, "<"),
		sub_string(Z, 0, J, _, Z0)
	),
    load_files([Z0]).

%
handle([collect, functors]) --> region,
 	swap_args:elem_list(X, []),
	{	swap_args:collect_functors(X,  Y, []),
		sort(Y, Y0),
		insert(",\n", Y0, Y1)
	},
	peek(["[", Y1, "]"]).
%
handle([set, functors, list]) --> region,
	herbrand,
	current(Sgn_list),
	{ nb_setval(functors_list, Sgn_list) },
	peek("*functors_list set*\n").

%
handle([insert, last, arg]) --> region,
	{ nb_getval(functors_list, Sgn),
	  (		nb_current(arg_name, Arg), Arg \== [] -> true
	  ;		Arg = "State"
	  ),
	  string_concat(" ", Arg, Arg0)
	},
	swap_args:elem_list(X, []),
	peek(X),
	swap_args:edit_elem_list(swap_args:insert_last_arg, Sgn, Arg0),
	overwrite.

handle([set, string, V]) --> region,
	pred([X, Y]:- string_codes(Y, X)),	% flip(string_codes),
	current(String),
	{ nb_setval(V, String)}.


expand_tilda(X, Y):- expand_file_name(X, [Y|_]).

%!	handle(+Action:[remove, comment]) is det.
%  Remove comments in the region.

handle([remove, comment]) --> region,
							  remove_comment,
							  overwrite.

handle([luatex, region])--> region,
	current(R),
	peek([]),
	{
		expand_tilda("~/tmp/deldel.tex", TeXFile),
		expand_tilda("~/tmp/preamble.tex", Preamble),
		Fs=	[text("\\RequirePackage{luatex85}\n"),
			text("\\documentclass{ltjsarticle}\n"),
			text("\\usepackage[hiragino-pron,jis2004]{luatexja-preset}\n"),
			file(Preamble),
			text("\\begin{document}\n"),
			codes(R),
			text("\\end{document}\n") ],
			assemble(Fs, TeXFile)
	},
	{	expand_tilda("~/tmp", TMP),
		qshell(	cd(TMP) ;
				lualatex("deldel") ;
				open(-a("Preview"), "deldel.pdf")
			)

	}.

%!	handle(+Action:[compile, pac, region]) is det.
%  Compile pac clauses in the region.
% ==
% ?- ejockey:handle([compile, pac, region], `a.\na.\n`, R).
% ==
% ?- ejockey:handle([compile,pac,region, dbg], `:-betrs(a).\n:-etrs.\n`, R).
% ?- ejockey:handle([compile,pac,region, dbg],
%	`f:= [g/0-h].\n`, R).
% ?- ejockey:handle([cpr], `a:-b.\n`, R), smash(,R).
% ?- ejockey:handle([cpr], `a:-b(pred([x])).\n`, R), smash(R).
% ?- ejockey:handle([cpr], `:-bekind(f, []).\n:-ekind.\n`, R), smash(R).
% ?- ejockey:handle([cpr],  `:-bekind(f, []).\na=b.\n:-ekind.\n`, R),
%	basic:smash(R).
% ?- ejockey:handle([cpr],  `:-bekind(f, []).\na=b.\n:-ekind.\n`, R),
%	basic:smash(R).

% for short.
handle([cpr|Optional])-->handle([compile,pac,region|Optional]).

handle([compile, pac, region|Optional]) --> region,
	pred([X, Y]:- string_codes(Y, X)),	% flip(string_codes),
	string_to_terms,
	compile_terms_to_qstring,
	optional_overwrite(Optional).

%
optional_overwrite(Optional) -->
	{ partial_match(Optional, overwrite) }, !,
	overwrite.
optional_overwrite(_) --> [].

%
handle([region, string]) --> region, escape_codes_for_string.
%
escape_codes_for_string(X, [0'"|Y]):-	 %'
	escape_codes_for_string_(X, Y).
%
escape_codes_for_string_([], [0'"]):-!. %'
escape_codes_for_string_([X|P], Q):-
	(	X = 0'\n -> Q = [0'\\, 0'n|Q0]		%'
	;	X = 0'\t -> Q = [0'\\, 0't|Q0]		%'
	;	X = 0'" -> Q = [0'", 0'"|Q0]		%"
	;	X = 0'\\ -> Q = [0'\\, 0'\\|Q0]		%'
	;	Q = [X|Q0]
	),
	escape_codes_for_string_(P, Q0).
%
handle([region, atom]) --> region, escape_codes_for_atom.

escape_codes_for_atom(X, [0'\'|Y]):-  %'
	escape_codes_for_atom_(X, Y).
%
escape_codes_for_atom_([], [0'\']):-!. %'
escape_codes_for_atom_([X|P], Q):-
	(	X = 0'\n -> Q = [0'\\, 0'n|Q0]		%'
	;	X = 0'\t -> Q = [0'\\, 0't|Q0]		%'
	;	X = 0'\' -> Q = [0'\', 0'\'|Q0]		%'
	;	X = 0'\\ -> Q = [0'\\, 0'\\|Q0]		%'
	;	Q = [X|Q0]
	),
	escape_codes_for_atom_(P, Q0).

handle([region, bq]) --> region, escape_codes_for_BQ.

escape_codes_for_BQ(X, [0'\`|Y]):-  %'
	escape_codes_for_BQ_(X, Y).
%
escape_codes_for_BQ_([], [0'\`]):-!. %'
escape_codes_for_BQ_([X|P], Q):-
	(	X = 0'\n -> Q = [0'\\, 0'n|Q0]		%'
	;	X = 0'\t -> Q = [0'\\, 0't|Q0]		%'
	;	X = 0'\` -> Q = [0'\`, 0'\`|Q0]		%'
	;	X = 0'\\ -> Q = [0'\\, 0'\\|Q0]		%'
	;	Q = [X|Q0]
	),
	escape_codes_for_BQ_(P, Q0).




%!	handle(+Action:[compile, pac, generic]) is det.
%  Compile pac clauses in the region.
%

% ?- ejockey:handle([compile, pac, generic], `~/local/lib/pacpl7/a.pl`, R).
handle([compile, pac, generic]) -->
	region,
	split,
	maplist(trim_white),
	remove([]),
	maplist(pred([X, F]:- string_codes(F,  X))),
	(	maplist(compile_pac_generic),
		peek([]),
		{message("done")},
		!
	;	peek([]),
		{message("Error ! file not exists ?")}
	).

% ?- ejockey:compile_pac_generic("~/local/lib/pacpl7/a.pl", X).
% ?- ejockey:compile_pac_generic("~/local/lib/pacpl7/a.pac", X).
% ?- ejockey:compile_pac_generic("~/local/lib/pacpl7/a", X).
compile_pac_generic(In, Files):-
	expand_file_name(In, Files),
	maplist(compile_pac_generic, Files).
%
compile_pac_generic(Src):-
	once(determine_source(Src, Src0)),
	setup_call_cleanup(open(Src0, read, SX, [encoding(utf8)]),
		stream_parse_pac_terms(SX, Xs, []),
		close(SX)),
	compile_terms_to_qstring(Xs, QuasiText),
	determine_target(Src0, Target),
	setup_call_cleanup(open(Target, write, SY, [encoding(utf8)]),
		write_qstring(QuasiText, SY),
		close(SY)).

%
determine_source(X, X):- string_drop_suffix(X, ".pl", X0), !,
	string_concat(X0, ".pac", X1),
	\+exists_file(X1).
determine_source(X, X):- string_end_with(X, ".pac"), !,
	exists_file(X).
determine_source(X, Y):- member(M, [".pac", ".pl"]),
	string_concat(X, M, Y),
	exists_file(Y),
	!.
%
determine_target(X, Y):- string_drop_suffix(X, ".pac", Z), !,
	string_concat(Z, ".pl", Y).
determine_target(X, X):- string_drop_suffix(X, ".pl", Y),
	modify_file_name(Y, 0, '.pac', Keep),
	rename_file(X, Keep).

% ?- ejockey:string_drop_suffix("abcd", "cd", Y).
string_end_with(X, S):- sub_string(X, _, _, 0, S).

% ?- ejockey:string_drop_suffix("abcd", "cd", Y).
string_drop_suffix(X, S, Y):- sub_string(X, J, L, 0, S),
	sub_string(X, 0, J, L, Y).
%
ignore_pac_term(term_expansion-->expand_pac).

%
stream_parse_pac_terms(Stream, U, Q) :-
		stream_term_string(Eqs, X, Stream),
		(	X == end_of_file	->	U = Q
		;	ignore_pac_term(X)	->  stream_parse_pac_terms(Stream,  U, Q)
		;  	(	is_backquote_begin(X, X0),
				set_prolog_flag(back_quotes, symbol_char)
			; 	is_backquote_end(X, X0),
				set_prolog_flag(back_quotes, codes)
			; 	X0 = X
			),
			!,
			U = [X0-Eqs|P],
			stream_parse_pac_terms(Stream, P, Q)
		).

%
compile_terms_to_qstring(Xs, QuasiText):-
	compile_pac(Xs, P, []),
	pred_grouping(P, Blocks),
	maplist(maplist(pred([P, pair(X0, H0)]:-
							 clause_to_string(P, X0, H0))),
			Blocks, ExpandedBlocks),
	maplist(pred(([Block, [Cs,"\n", Hs]]:-
				maplist(pred([pair(U,V), U, V]), Block, Cs, Hs))),
			ExpandedBlocks, QuasiText).

%
write_qstring([A|B], S):- write_qstring(A, S),
						  write_qstring(B, S).
write_qstring([], _).
write_qstring(A, S):- write(S, A).


%!	handle(+Action:[compile, predicate]) is det.
%  Compile a pac clause at the region.
handle([compile, predicate|Flag]) --> region,
	herbrand(web, Eqs),
	current(X),
	{ pac_listing:compile_pred_word(X, Eqs, H0, R0) },
	peek([R0, "\n", H0, ".\n"]),
	overwrite(Flag).

%	select_phrase(partial_match(Flag, overwrite), overwrite, =).

%!	handle(+Action:[expand, pac]) is det.
%  Expand pac clauses.

handle([expand, pac]) --> region,
	pred([X, Y]:- string_codes(Y, X)),
	parse_pac_terms,
	pred([Xs, P]:- compile_pac(Xs, P, [])),
	pred_grouping,
	maplist(maplist(pred([P, pair(X0, H0)]:-
							 pac_listing:clause_to_string(P, X0, H0)))),
	maplist(pred(([Block, [Cs,"\n", Hs]]:-
				maplist(pred([pair(U,V), U, V]), Block, Cs, Hs)))).

%
parse_pac_terms(X, Pacs):-
	setup_call_cleanup(
		open_string(X, Stream),
		stream_parse_pac_terms(Stream, Pacs, []),
		close(Stream)).

%!	handle(+Action:[expand, pac, overwrite]) is det.
% Expand pac clauses, and rewrite the region with them.

handle([expand, pac, overwrite]) -->
	handle([expand,pac]),
	overwrite.

%!	handle(+Action:[neighbor, string]) is det.
%  Return the string on which the cursor is.

handle([neighbor, string]) --> {neighbor_string("[","]", X)},
							   peek(X).

%!	handle(+Action:[collect, identifiers]) is det.
%   Collect Prolog identifiers.

handle([collect, identifiers]) --> region,
	collect_tokens(prolog_indentifier),
	maplist(herbrand),
	sort,
	insert("\n").

%!	handle(+Action:[collect, keywords]) is det.
%   Collect keywords.

handle([collect, keywords]) --> region,
	collect_tokens(keyword),
	maplist(herbrand),
	sort,
	insert("\n").

%!	handle(+Action:[copy, handle]) is det.
%   Copy the current head of a clause, and insert it
%   after modifying before the line.

handle([copy, head])-->
	 {	line_get(Obj),
		obj_get([line(Line)], Obj),
		string_codes(Line, Codes),
		phrase(( wl("[\s\t]*"),
				 w(".*", Head),
				 w("[\s\t]*((-->)|(:-))")),
			   Codes, _)
		},
	 peek(["%%\t", Head, " is det.\n%\n%\n", Line]),
	 overwrite.

%!	handle(+Action:[swap, args, +Name, +I, +J]) is det.
%   Swap I-th argument with J-th one in for all terms
%	with functor Name.

handle([swap, args, N, I, J]) -->
	{ atom_number(I, I0),
	  atom_number(J, J0) },
	region,
	pred([N, I0, J0],
		 ([X, Y]:-
			 swap_args:swap_args_of(N, I0, J0, X, Y))),
	overwrite.

%!	handle(+Action:[handle, kbd]) is det.
%   ask LISP to eval global-set-key

handle([handle, kbd])  --> region,
	pred(([S, []] :- term_codes(E, S),
		arg(1, E, K),
		arg(2, E, P),
		global_set_kbd(K, P))).

%!	handle(+Action:[dired]) is det.
%   Activate emacs dired with a prompt.

handle([dired])  --> dired.

%!	handle(+Action:[dired, mark, swipl]) is det.
%   Mark all *.pl files in DIRED.

handle([dired, mark, swipl]) --> {dired_mark_swipl}.

%!	handle(+Action:[swap, args, X]) is det.
%	Swap the first two arguments.
handle([swap, args, X]) --> region,
	swap_args_of(X, 1, 2),
	overwrite.

%!	handle(+Action:[numbering, paragraphs]) is det.
%  Numbering paragraphs.
handle([numbering, paragraphs])  --> region,
	paragraph,
	remove([]),
	pred(([X, Y]:- length(X, N),
	      numlist(1, N, Ns),
	      maplist(pred([I, P,[I0, ". ", P]]:-
			  number_string(I, I0)),
		      Ns, X, Y))),
	insert("\n"),
	overwrite.


%!	handle(+Action:[help]) is det.
%   Show all handle commands.
% handle([help]) --> {setof(H, P^Q^R^clause(handle(H,P,Q), R), S),
% 		 sort(S, S1),
% 		 maplist(pred(([X, Y]:-
% 				numbervars(X,0,_),
% 				write_to_chars(X, Z, []),
% 				string_codes(Z, Y))),
% 			S1, S2),
% 		 insert_tab_nl(4, 4, S2, R)
% 		},
% 		peek(R).

%!	handle(+Action:[handle, list]) is det.
%   Set Lisp variables  handle-list to the list of all handles.


handle([handle, list]) --> region,
    {findall(H, clause(handle(H,_,_), _), S),
	 sort(S, S1),
	 maplist(pred(([H, H0] :-
			   foldl(pred([H, H0], ([X, ['VAR'|L], L]:- var(X))
					 &     [X, [X|L], L]
				     ),
				     H, H0, []))),
		 S1, S2),
	remove([], S2, S3),
	maplist(pred(([X, Y]:- atomics_to_string(X, " ", Y0),
				  atomics_to_string(["[", Y0, "] "], Y))),
				S3, Commands)
%	,	List =..[list|Commands]
%	,	elisp:lisp(setq('handle-list', List))
%	,	elisp:lisp(message("variable 'handle-list' has been set."))
	},
	peek(Commands),
	overwrite.

%!	handle(+Action:[reload, buffer, file]) is det.
%   Reload the current buffer file.
handle([reload, buffer, file]) -->
	{ call_lisp('buffer-file-name'(), [value(File)]),
  	  string_codes(S, File),
	  unload_file(S),
	  load_files(S, [silent(true)]) }.

%!	handle(+Action:[edit]) is det.
%	Generate edit commnad ?- edit(X) from the symbol at the current point.
handle([edit])--> { prolog_symbol_at_point(X) },
			  peek(["\n\n% ?- edit(", X, ").\n"]).

%!	handle(+Action:[take, memo]) is det.
%  Take a time-stamped memo, and display it
%  in the emacs window.
handle([take, memo]) --> region,
 	{   lisp('time-stamp', Dir),
		getinfo_string("date +%Y-%m-%d", Today),
		atomics_to_string([Dir, Today], File0),
		expand_file_name(File0, [File|_]),   % File is of type atom.
		pshell(touch(File))
	},
	pred([Today, File],
		 ( [ Note, []]
		   :- file:push_to_file(
					 basic:smash(["[", Today, "]\n",
								  Note,
								  "\n"]),
							File))),
	{	atom_string(File, File0),
		lisp('find-file'(File0))
	}.

%!	handle(+Action:[clipboard]) is det.
% 	Save the clipboard as a note.
% handle([clipboard]) -->
% 	{getinfo_codes("pbpaste", D)},
% 	peek(D),
% 	handle([snippet]).

%!	handle(+Action:[choose, working, directory]) is det.
%	Choose a folder, and save its name to a lisp variable.

handle([choose, working, directory]) -->
	{	choose_folder(X),
		set_string(working_directory, X)
	},
	peek(X).

%!	handle(+Action:[choose, folder]) is det.
%   Choose a folder.
handle([choose, folder]) --> {choose_folder(X)}, peek(X).

%%	handle(+Action:[choose, file]) is det.
%	Choose a file.
handle([choose, file]) --> {choose_file(X)}, peek(X).

%!	handle(+Action:[choose, file, open]) is det.
%	Choose a file and opent it.
handle([choose, file, open]) -->
	{	choose_file(X),
		term_string(X, Y),
		pshell(open(Y))
	},
	peek(X).

% ?- atomic_list_concat([a, ' b'], C).

%!	handle(+Action:[choose, folder, open]) is det.
%	Choose a folder, and open it.
handle([choose, folder, open]) -->
	{	choose_folder(X),
		term_string(X, Y),
		pshell(open(Y))
	},
	peek(X).

		/********************
		*     sort lines    *
		********************/

%!	handle(+Action:[sort, lines]) is det.
%   Sort lines.
handle([sort, lines]) --> region,
						  split,
						  sort,
						  insert("\n"),
						  overwrite.

%%	handle(+Action:[trim, sort, lines]) is det.
%   Trim leading white codes and sort the trimed lines.

handle([trim, sort, lines]) --> region,
						  split,
						  remove([]),
						  maplist(phrase(wl("[\s\t]*"))),
						  sort,
						  maplist(pred([L, [L,"\n"]])),
						  overwrite.

% ?- trace, nopac(expand_basic(maplist(phrase(wl("[\s\t]*"))), user, G, P, [])).

		/************************
		*     shell in buffer   *
		************************/

%!	handle(+Action:[shell]) is det.
%	Run shell commands in the region.

handle([shell]) --> set_mark_region,
	region,
	split,
	remove([]),
	insert(" ; "),
        {
	    tmp_file_stream(utf8, File, Stream),
	    close(Stream)
	},
	peek(X, ["( ", X, " ) > ", File]),
	smash,
	pred(File, ([Shell_in_string, Codes]:-
			shell(Shell_in_string),
			read_file_to_codes(File, Codes, [tail([]), encoding(utf8)])
		   )
	    ),
        {
	    delete_file(File)
	}.

%%	handle(+Action:[shell, dakuten]) is det.
%	Run shell commnad in the region, and convert
%   the Japanese ligature to the normal form.
handle([shell, dakuten]) --> handle([shell]), dakuten_convert.

%!	handle(+Action:[message]) is det.
%   Run message command of Lisp.

handle([message]) --> region, pred([M, []]:- message(M)).

		/*********************
		*     prolog/lisp    *
		*********************/

%!	handle(+Action:[prolog]) is det.
%	Run a Prolog goal.
%
handle([prolog])--> handle([prolog, paragraph]).

%!	handle(+Action:[prolog, paragraph]) is det.
%	Run a Prolog goal.

handle([prolog, paragraph]) --> set_mark_region,
	region,
	remove_leading_comment_chars,
	(	eh:solve_bind_once,
		peek(X, ["\n", X,"\n"])
	;	peek("\nfail")
	).

%!	handle(+Action:[prolog, region]) is det.
%	Run a prolog goal.
%
handle([prolog, region])    --> region,
	remove_leading_comment_chars,
	(	eh:solve_bind_once,
		peek(X, ["\n", X,"\n"])
	;	peek("\nfail")
	).

%!	handle(+Action:[comment, region]) is det.
%	Put the comment symbol to each line of the region.
handle([comment, region])  --> region,
	split,
	maplist(comment),
	insert("\n"),
	overwrite.

%!	handle(+Action:[uncomment, region]) is det.
%	Remove the comment symbol of each line of the region.
handle([uncomment, region])  --> region,
	split,
	maplist(uncomment),
	insert("\n"),
	overwrite.

% Handle call with:  elisp value string
% (append (list 1 2 3) (list 4 5 6))
% => (1 2 3 4 5 6)
%
handle([lisp | Keys]) -->
	(	{ apropos_chk(Keys, paragraph) }
	->  set_mark_region
	;	[]
	),
	region,
	current(X),
	{ handle_lisp(X, Keys, Out) },
	peek(Out).

%!	handle(+Action:[enum]) is det.
%   Make a LaTeX enumerate environment from
%	the items in the region.
handle([enum]) --> region, snippets:environment(enumerate), overwrite.

%!	handle(+Action:[eit]) is det.
%   Make a LaTeX eitemize environment from
%	the items in the region.
handle([eit])  --> region, snippets:environment(itemize), overwrite.

%!	handle(+Action:[html,tag,code]) is det.
%   Put "<code>" and "</code>" around the region.
handle([html,tag,code])  --> region,
			 peek(X, ["<code> ", X, " </code>"]),
			 overwrite.

%!	handle(+Action:[ignore, goal|R]) is det.
%	Run the region as a goal igonoring errors.
handle([ignore, goal|R]) --> region_or_line(R),
	herbrand(_),
	pred([X, []]:- ignore(X)).

%!	handle(+Action:[solve,  goal|R]) is det.
%	Run the region as a goal catching errors.
handle([solve,  goal|R]) -->  region_or_line(R),
	herbrand(_),
	pred(B, [X, ["\n", R, "\n"]]:- catch_once(X, B, R)).

%!	handle(+Action:[shift, region]) is det.
%	Insert tab before the each line of the region.
%	Insert tab before the each line of the region.
%	Insert tab before the each line of the region.

handle([shift, region]) -->  region,
	split,
	maplist(pred([X, ['\t'|X]])),
	insert('\n'),
	overwrite.

%!	handle(+Action:[insert, before, A]) is det.
%	Insert A before the each line of the region.
handle([insert, before, A]) -->  region,
	split,
	maplist(pred(A, [X, [A|X]])),
	insert('\n'),
	overwrite.

		/****************
		*     indent    *
		****************/

%!	handle(+Action:[tab, region]) is det.
%!	handle(+Action:[tab, region, N]) is det.
%   insert tab.
%!	handle(+Action:[indent, region]) is det.
%!	handle(+Action:[indent, region, N]) is det.
%   Indent the region.

handle([tab, region])	-->  indent_region(0'\t, 1). %'
handle([tab, region, N])-->  { atom_number(N, N0) },
	indent_region(0'\t, N0).		%'
handle([indent, region])--> indent_region(0'\s, 4).		%'
handle([indent, region, N]) -->  { atom_number(N, N0) },
	indent_region(0'\s, N0).					%'

%!	handle(+Action:[header]) is det.
%	Put framed header without centering.
%
handle([header]) -->  region,
	split,
	maplist(remove_trailing_white),
	maplist(detab),
	remove_enveloping_nulls,
	pred(Max, ([X, X]:- maplist(length,X, L), poly:list_max(L, Max))),
	{ Width is Max + 7,
	  length(Hr, Width),
	  maplist(=("*"), Hr)
	},
	maplist(fill_trailing_spaces(Width)),
	maplist(pred([X, ["\t*", X, "*\n"]])),
	peek(Body,	["\t/", Hr, "*\n",
				Body,
				"\t*", Hr, "/\n"]),
	overwrite.

%
fill_trailing_spaces(Width, X, [X, Y]):-
	length(X, L),
	J is Width - L,
	length(Y, J),
	maplist(=("\s"), Y).

% ?- ejockey:remove_enveloping_nulls([[], a, [], b,[]], R).
%@ R = [a, [], b] .
remove_enveloping_nulls -->[[]], remove_enveloping_nulls.
remove_enveloping_nulls -->[], remove_trailing_nulls.
%
remove_trailing_nulls(X, []):- nulls(X), !.
remove_trailing_nulls([X|R], [X|S]):-
	remove_trailing_nulls(R, S).

%
nulls([[]|R]):- nulls(R).
nulls([]).

% ?- ejockey:remove_trailing_white(X, `abc   `, Y).
% ?- trace, ejockey:remove_trailing_white(`abc   `, Y).
% ?- trace, ejockey:remove_trailing_white(`a`, Y).

handle([trim, trailing, white]) --> region, remove_trailing_white(X), peek(X).

remove_trailing_white(X) --> w(".*", X), wl("[\s\t]*$").

remove_trailing_white --> w(".*", X), wl("[\s\t]*$"), peek(X).

%!	handle(+Action:[detab]) is det.
%	Removing tab from each line of the region.
%
handle([detab]) --> region, detab, overwrite.
%
detab_spaces(`\s\s\s\s`).
%
detab(X, Y):- detab(X, Y, []).

% ?- ejockey:detab(`\ta\t\b`, X, []).
detab([0'\t|R], X, Y):- !, detab_spaces(S),   %'
	append(S, X0, X),
	detab(R, X0, Y).
detab([A|R], [A|X], Y):-detab(R, X, Y).
detab([], X, X).


%!	handle(+Action:[shift, frame|R]) is det.
%	Make a framed header wiht centering.
handle([shift, frame|Optional]) --> region_or_line(Optional),
	split,
	maplist(trim_white),
	remove_enveloping_nulls,
	pred(Max, ([X, X]:- maplist(length,X, L), poly:list_max(L, Max))),
	{ Width is Max + 11 },
	maplist(pred([Width, Max], ( [A, B]:-
				length(B, Width),
				N is (Width - Max) div 2,
				length(L, N),
			       	append([ ['*'], L, A, R, ['*']], B),
				maplist(=('\s'), L),
			        maplist(=('\s'), R)))),
	{	length(Top, Width),
		length(Bottom, Width) ,
		Top = ['/'|L0],
		append(L1, ['/'], Bottom),
		maplist(=('*'), L0),
		maplist(=('*'), L1)
	},
	maplist(pred([X, ['\t\t', X, '\n']])),
	peek(Body, ['\t\t', Top, '\n',
				Body,
				'\t\t', Bottom, '\n']),
	overwrite.

%!	handle(+Action:[single, quote]) is det.
%	Convert the region to a comma list of
%	lines in the region added with single quotation
%	marks.

handle([single, quote]) --> region,
	split,
	remove([]),
	maplist(html:single_quote),
	insert(',\n'),
	overwrite.

%!	handle(+Action:[double, quote]) is det.
% 	Convert the region to a comma list of lines
%	in the region added with double quotation marks.
handle([double, quote]) --> region,
	split,
	remove([]),
	maplist(html:double_quote),
	insert(',\n'),
	overwrite.

%!	handle(+Action:[region, standalone]) is det.
%	Copy region to copyboad; lualatex it with standalone class.
handle([region, standalone])  -->
	{ call_lisp(pbcopy()),
	  shell(standalone, 0)
	}.

%!	handle(+Action:[pasteboard, standalone]) is det.
%	lualatex pasteboard text with standalone class.
handle([pasteboard, standalone])  --> { shell(standalone, 0)}.

%!	handle(+Action:[description]) is det.
%!	handle(+Action:[description]) is det.
%!	handle(+Action:[description]) is det.
%	Show a LaTeX description environment.
handle([description]) -->  peek([
	"\\begin{description}[style=multiline, labelwidth=1.5cm]",
	"\\item[\\namedlabel{itm:rule1}{Rule 1}] Everything is easy with \\LaTeX",
	"\\item[\\namedlabel{itm:rule2}{Rule 2}] Sometimes it is not that easy\\\\",
	"$\\to$ \\ref{itm:rule1} applies",
	"\\end{description}\n"	]),
	insert("\n"),
	overwrite.


%!	handle(+Action:[list, listing]) is det.
%   Generate a LaTeX listlisting environment.
handle([list, listing]) --> region,
	pred([	X, [	"\\begin{lstlisting}[caption={},label=src:]\n",
			X,
			"\\end{lstlisting}\n"	]]),
	overwrite.
%!	handle(+Action:[begin, align]) is det.
%	Generate a LaTeX align* environment.
handle([begin, align]) --> region,
	pred([	X, [	"\\begin{align*}\n",
			X,
			"\\end{align*}\n"	]]),
	overwrite.
%!	handle(+Action:[vbox]) is det.
%	Generate TeX \vbox template.
handle([vbox]) --> peek([
	"$$\\vbox{\\offinterlineskip",
        "\\halign{\\strut",
        "\\vrule\\vrule\\quad\\textbf{#}\\hfill\\quad & \\vrule\\quad\\hfill #cm \\quad ",
         "& \\vrule\\quad\\hfill #kg \\quad\\vrule\\vrule\\cr",
        "\\noalign{\\hrule\\hrule}",
        "鈴木 一太郎 & 168 & 74 \\cr",
        "\\noalign{\\hrule} ",
        "山田 太郎   & 170 & 72 \\cr",
        "\\noalign{\\hrule} ",
        "渡辺 次郎   & 192 & 103 \\cr",
        "\\noalign{\\hrule\\hrule} ",
        "}}$$"		]),
	insert("\n").
%!	handle(+Action:[halign]) is det.
%	Generate TeX halign
handle([halign]) --> peek([
	"\\halign{",
%	"\\hfill$#$\\hfill\\qquad&\\hfill$#$\\hfill&\\quad\\text{#}\\cr\n",
	"\\hfill$#=\\>$ & $#$ \\hfill & \\qquad \\mbox{#} \\cr\n",
	" &  &   \\cr\n",
      	" &  &   \\cr\n",
	" &  &   \\cr\n",
	"}"	]).
%!	handle(+Action:[case, equation]) is det.
%	Generate a LaTeX cases environment.
%
handle([case, equation]) --> region,
	pred([	Left, [		"$", Left, "= \n",
				"\\begin{cases}\n",
				"     &  \\mbox{} \\\\\n",
				"     &  \\mbox{} \\\\\n",
				"     &  \\mbox{} \n",
				"\\end{cases}$\n"	]]),
	overwrite.
%!	handle(+Action:[eqn, array]) is det.
%	Generate a LaTex eqnarray* environment.
handle([eqn, array]) --> region,
	pred([	_, [	"\\begin{eqnarray*}\n",
			"     &=&          \\\\\n",
			"     &=&          \\\\\n",
			"     &=&          \\\\\n",
			"\\end{eqnarray*}\n"	]]),
	overwrite.

%!	handle(+Action:[eqn, array]) is det.
%	Generate a LaTeX eqnarray* environment.
handle([eqn, array]) --> region, split, remove([]),
	maplist(split(`=`)),
	maplist([[X|Y],	[X, " &=& ", Y]]),
	insert("\\\\\n"),
	pred([Body, [	"\\begin{eqnarray*}\n",
			Body,
			"\n\\end{eqnarray*}\n"	]]),
	overwrite.
%!	handle(+Action:[q, F,X, N]) is det.
%
%
%!	handle(+Action:[q, F,X, N]) is det.
%
%
handle([q, F,X, N]) -->
	peek([F, "(", X, "_1, ", X, "_2, ", "\\ldots ,", X, "_", N, ")"]).
%!	handle(+Action:[q, F, X, N0, N]) is det.
%
%
handle([q, F, X, N0, N]) -->
	peek([F, "(", X, "_", N0, ", ", X, "_1, ", "\\ldots ,", X, "_", N, ")"]).
%!	handle(+Action:[q]) is det.
%
%
handle([q]) --> region, split(` `), remove([]),
	pred(	([[F, X, N], E]:- handle([q, F, X, N], _, E))
		&
		([[F, X, N0|N], E]:- handle([q, F, X, N0, N], _, E))),
	overwrite.
%!	handle(+Action:[ref]) is det.
%
%
handle([ref])	--> region, pred([X, ["\\ref{", X, "}"]]), overwrite.
%!	handle(+Action:[cite]) is det.
%
%
handle([cite])	--> region, pred([X, ["\\cite{", X, "}"]]), overwrite.
%!	handle(+Action:[cs, N]) is det.
%
%
handle([cs, N]) --> region,
	pred([N],[X, ["\\", N, "{", X, "}"]]),
	overwrite.
%
%!	handle(+Action:[parse, bind, context]) is det.
%
%
handle([parse, bind, context])	--> region,	parse_bind_context, overwrite.
%!	handle(+Action:[parse, bind, context, append]) is det.
%
%
handle([parse, bind, context, append])	--> region, parse_bind_context.
%!	handle(+Action:[eval, markup, text]) is det.
%
%
handle([eval, markup, text])		--> region, eval_markup_text, overwrite.
%!	handle(+Action:[eval, markup, text, append]) is det.
%
%
handle([eval, markup, text, append])	--> region, eval_markup_text.
%
%!	handle(+Action:[tag, l]) is det.
%
%
handle([tag, l])	--> {nb_getval(phrase_tag, G), herbrand_opp(G, G0)},  % to list the saved tag
	peek(G0).
%!	handle(+Action:[tag, s]) is det.
%
%
handle([tag, s|P])	--> region_or_line(P),		% to save the tag
	peek(Q),
	{parse_phrase_save(Q)},
	peek("\n the tag saved.\n").
%!	handle(+Action:[tag, a]) is det.
%
%
handle([tag, a])  --> region,
	{ nb_getval(phrase_tag, G) },
	act(G),
	overwrite.
%!	handle(+Action:[t, a]) is det.
%
%

region_debug(X, Y):- var(X), !, region(X, Y).
region_debug(X, X).


handle([t, a])  --> handle([tag,a]).
%!	handle(+Action:[bib]) is det.
%	Generate bibliography commands.
handle([bib]) --> peek(["\\bibliographystyle{plain}\n",
	"\\bibliography{jmukai,mukai}\n"]).

% Convert  delicious 3 data in book/1 to bibtex form.
handle([book, bibtex]) --> region,
		paragraph,
		remove([]),
		maplist(herbrand),
		maplist(book_bibitem),
		insert("\n"),
	    overwrite.

% Convert  csv data to a dict  with keywords.
handle([csv, bibtex])	--> region,
						csv_to_dict,
						peek(key_dict(_, L), L),
						maplist(dict_bibtex).

% Convert  csv data to a dict  with keywords.
% ?- ejockey:csv_to_dict(`a\tb\tc\n1\t2\t3\n4\t5\t6`, R).
%@ R = key_dict([a, b, c], [[a="1", b="2", c="3"], [a="4", b="5", c="6"]]) .

csv_to_dict --> split,
		remove([]),
		maplist(split("\t")),
		peek([H|R], R),
		{	maplist(atom_codes, Keys0, H),
			map_key_tbl(M),
			map_key(Keys0, M, Keys)
		},
		maplist(pred(Keys,
					 ([A, B]:-
						 maplist(pred([K, A0, K=A1]:-
									 string_codes(A1, A0)),
								 Keys, A, B)))),
		peek(D, key_dict(Keys, D)).

%
dict_bibtex(L, BB):-
	maplist(pred([K=V, [K, " = ", "{", V, "}"]]), L, Items),
	insert(",\n", Items, Items0),
	smash(["@book{", "to be filled", ",\n", Items0, "\n}\n"], BB).

%
map_key_tbl([creator-author, 'ISBN'-isbn]).

%
map_key([],_,[]).
map_key([K|R],M,[K0|R0]):- memberchk(K-K0, M), !,
						   map_key(R, M, R0).
map_key([K|R],M,[K|R0]):-  map_key(R, M, R0).


%!	handle(+Action:[thm]) is det.
%	Generate a LaTeX thm environment.
handle([thm|X])--> region_or_line(X),
	peek(Y, ["\\begin{thm}\\label{thm:}\n", Y, "\\end{thm}\n"]),
	overwrite.

%!	handle(+Action:[prop|X]) is det.
%	Generate a LaTeX prop environment.
%
handle([prop|X])--> region_or_line(X),
	peek(Y, ["\\begin{prop}\\label{prop:}\n", Y, "\\end{prop}\n"]),
	overwrite.

%!	handle(+Action:[lem]) is det.
% 	Generate a LaTeX lemma environment.

handle([lem|X])--> region_or_line(X),
	peek(Y, ["\\begin{lemma}\\label{lem:}\n", Y, "\\end{lemma}\n"]),
	overwrite.
%!	handle(+Action:[cor|X]) is det.
% 	Generate a LaTeX cor environment.
handle([cor|X])-->region_or_line(X),
	peek(Y, ["\\begin{cor}\\label{cor:}\n", Y,"\\end{cor}\n"]),
	overwrite.

%!	handle(+Action:[ex|X]) is det.
% 	Generate a LaTeX ex environment.
handle([ex|X])--> region_or_line(X),
	peek(Y, ["\\begin{ex}\n", Y, "\\end{ex}\n"]),
	overwrite.

%!	handle(+Action:[df|X]) is det.
% 	Generate a LaTeX df environment.
handle([df|X])--> region_or_line(X),
	peek(Y, ["\\begin{df}\\label{df:}\n", Y, "\\end{df}\n"]),
	overwrite.

%!	handle(+Action:[rem|X]) is det.
% 	Generate a LaTeX rem environment.
handle([rem|X])--> region_or_line(X),
	peek(Y, ["\\begin{remark}\\label{rem:}\n",Y,"\\end{remark}\n"]),
	overwrite.

%!	handle(+Action:[proof|X]) is det.
% 	Generate a LaTeX proof environment.
handle([proof|X])-->region_or_line(X),
	peek(Y, ["\\begin{Proof}\n",Y,"\\end{Proof}\n"]),
	overwrite.

%!	handle(+Action:[red|X]) is det.
%	Insert red color macro.
handle([red|X])	--> region_or_line(X),
					peek(Y, ["\\Red{", Y, "}"]), overwrite.
%!	handle(+Action:[blue]) is det.
%	Insert blue color macro.
handle([blue|X])	--> region_or_line(X),
						peek(Y, ["\\Blue{", Y, "}"]), overwrite.
%!	handle(+Action:[green]) is det.
%	Insert green color macro.

handle([green|X]) --> region_or_line(X),
					  peek(Y, ["\\Green{", Y, "}"]),
					  overwrite.

% I don't remember what is the purpose of the following handle.
% handle([mkh])	--> region,
% 	split,
% 	maplist(pred( ( [X, Y]:-
% 						html:single_quote(X, X0),
% 						atom_codes(Y, X0)))),
% 	pred([L, ( handle([names]
% 					 ) --> peek(L), insert_nl, ".")]).

%!	handle(+Action:[ms]) is det.
%	Put  font macro \mathscr
handle([ms|X])	--> region_or_line(X),
					peek(Y, ["\\mathscr{", Y, "}"]), overwrite.

%!	handle(+Action:[eq]) is det.
%	Generate a LaTeX euation environment with a label command.
handle([eq|X])	--> region_or_line(X),
	peek(Y, ["\\begin{equation}\\label{eq:}\n",
				Y, "\n",
			"\\end{equation}\n"]),
	overwrite.

%!	handle(+Action:[get, module, name]) is det.
%   get the module name from the source in the
%	current buffer defined by ":- module(<name>, ...)."
handle([get, module, name]) -->
 {
	wait(progn(
		setq(point_saved, point()),
		'goto-char'('point-min'()))),
	line_get(Obj),
	obj_get([line(Line)], Obj),
	string_codes(Line, Codes),
	module_name(Codes, Name),
	wait('goto-char'(point_saved))
    },
    peek(Name).
%!	handle(+Action:[sqc]) is det.
%   Set query context.
handle([sqc]) --> handle([set, query, context]).  % for short

%!	handle(+Action:[set, query, context]) is det.
%   To make Prolog mode to expand queries in the context module
%　 loaded in the current Emacs buffer. Otherwise, the query
%   may cause 'undefined predicate' errors in the query unless
%   the query is fully module prefixed.
handle([set, query, context]) -->
		handle([get, module, name]),
		peek(C, ["% ?- module(", C, ")."]),
		current(X),
		{	smash(X) },
		peek([]),
		{	wait('keyboard-quit'()) }.

		/*******************
		*     directory    *
		*******************/
%!	handle(+Action:[pwd]) is det.
%	Run the shell command pwd.
%
handle([pwd])	-->
	{
		get_string(working_directory, Path)
	},
	peek(Path).

%!	handle(+Action:[set, target, directory]) is det.
%	Set target directory.
handle([set, target, directory])	-->
	{	line_get(Obj),
		obj_get([line(Line)], Obj),
		trim_white(Line, DirPath),
		string_codes(S, DirPath),
		expand_file_name(S, [S0|_]),
		set_string(target_directory, S0)
	}.

%!	handle(+Action:[cwd]) is det.
%	Choose working directory, and set
%	the working_directory to it.
%
handle([cwd]) -->
	{	choose_folder(X),
		set_string(working_directory, X),
		working_directory(_, X)
	},
	peek(X).

%!	handle(+Action:[cd]) is det.
%	Change director to the HOME.
%
handle([cd]) -->
	{	expand_file_name("~/", [D]),
		set_string(working_directory, D),
		working_directory(_, D)
	},
	peek(D).

%!	handle(+Action:[cd, up]) is det.
%   Change directory like "../"
%
handle([cd, up]) -->
	{
		get_string(working_directory, Path),
		change_unix_path(up, Path, New_Path),
		set_string(working_directory, New_Path),
		working_directory(_, New_Path)
	},
	peek(New_Path).

%!	handle(+Action:[open]) is det.
%	Call Finder open.

handle([open])	-->
	{
		line_get(Obj),
		obj_get([line(Line)], Obj),
		trim_white(Line, Line0),
		double_quote(Line0, X),
		(	Line0 = [0'/|_]						% '
		-> 	sh_core(open(X))
		;	string_codes(XStr, X),
			handle_open_relative(XStr)
		)
	}.
%!	handle(+Action:[open, *]) is det.
%	Call Finder open  all files in the region.
handle([open, *]) -->
	{	get_string(working_directory, S),
		S\== ""
	},
	region,
	split,
	remove([]),
	reverse,
	current(L),
	{
	maplist(pred(S, ([X] :-
			double_quote([S, X], SX),
			sh_core(open(SX)))),
		L)
	},
	clear
	;
	peek("**** directory not found. ****\n").

		/*****************************
		*     Accessing Directory    *
		*****************************/
%!	handle(+Action:[finder, open, directory]) is det.
%   Call Finder open for directory.
handle([finder, open, directory])	-->
	{
		line_get(Obj),
		obj_get([line(Line)], Obj),
		first_token_codes(Line, Directory),
		sh(open(-a('Finder'), Directory))
	}.

%
append_slash_code([], [0'/]):-!.	%'
append_slash_code(Line, Line0):- last(Line, C),
	(	C == 0'/ -> Line0 = Line		%'
	;   append(Line, [0'/], Line0)		%'
	).

%!	handle(+Action:[swd]) is det.
%   Set working directory to working_directory.
handle([swd])	-->
	{	line_get(Obj),
		obj_get([line(Line)], Obj),
		trim_white(Line, Line0),
		append_slash_code(Line0, Line1),
		first_token_codes(Line1, Directory),
		string_codes(Dir_string, Directory),
		expand_file_name(Dir_string, [Full_path|_]),
		nb_setval(working_directory, Full_path),
		atom_string(Full_path, S),
		set_string(working_directory,S)
	},
	peek(Full_path).

%!	handle(+Action:[swd, (.)]) is det.
%   Set working directory to the default directory.

handle([swd, (.)])-->
	{	call_lisp_text('default-directory', D),
		string_codes(X, D),
		expand_file_name(X, [Full_path|_]),
		nb_setval(working_directory, Full_path),
		atom_string(Full_path, S),
		set_string(working_directory, S)
	},
	peek([]).

%!	handle(+Action:[finder, default, directory]) is det.
%	Open the default directory.
%
handle([finder, default, directory])	-->
	{	call_lisp_text('default-directory', D),
		string_codes(X, D),
		sh(open(X))
	}.

%
handle([directory, path])	-->
	{	call_lisp_text('default-directory', D)
	},
	peek(D).
%
handle([file, path]) -->
	{ lisp(list('default-directory', 'buffer-name'()), List)
	},
	peek(List).

%!	handle(+Action:[list, files]) is det.
%	List all files in the working directory.

handle([list, files])		--> % ls
	{	get_string(working_directory, S),
		S \== "",
		directory_files(S, Files)
	},
	peek(Files),
	insert("\n")
	;
	peek("**** directory not found. ****\n").

%!	handle(+Action:[dakuten]) is det.
%	Convert the text possibly with ligatures to the normal
%	normal sequences of chars.

handle([dakuten])--> region,
					flip(string_codes),
					dakuten_convert,
					overwrite.
%!	handle(+Action:[dakuten, flip]) is det.
%	Inverse of handle([dakuten]).

handle([dakuten, flip]) --> region,
					flip(string_codes),
					flip(dakuten_convert),
					overwrite.
%!	handle(+Action:[list, files, dakuten]) is det.
%   List files with the ligatures resolved
%	as in handle([dakuten]).
handle([list, files, dakuten])	--> % ls
	{	get_string(working_directory, S),
		S \== "",
		directory_files(S, Files)
	},
	peek(Files),
	maplist(string_codes),
	maplist(dakuten_convert),
	insert("\n")
	;
	peek("**** directory not found. ****\n").

%!	handle(+Action:[list, regex]) is det.
%	List files with a regex filter.

handle([list, regex])	--> region_term,
	current(Regex),
	{	let(Parser, pred(Regex, [X]:- phrase(w(Regex), X, []))),
		get_string(working_directory, S),
		S \== "",
		directory_files(S, Files)
	},
	peek(Files),
	maplist(dakuten_convert),
	maplist(string_codes),
	collect(Parser),
	insert("\n")
	;
	peek("**** directory not found. ****\n").

%!	handle(+Action:[list, timed, pdf]) is det.
%	List all of time-stamped pdf files.
handle([list, timed, pdf])	--> % ls
	{	get_string(working_directory, S),
		S \== "",
		directory_files(S, Files),
		maplist(atom_codes, Files, Codes_list),
		collect(pred([Codes]:- phrase(w("[0-9]+\\.pdf"), Codes,[])),
			Codes_list,
			Pdf_files)
	},
	peek(Pdf_files),
	insert("\n")
	;
	peek("**** directory not found. ****\n").

%!	handle(+Action:[mv]) is det.
%	Move files.
%
handle([mv])	--> % move a file over directories
	{	get_string(working_directory, S),
		S\== "" ,
		get_string(target_directory, T),
		T\== ""
	},
	rename(S, T)
	;
	peek("**** directory not found. ****\n").

%!	handle(+Action:[mv, (.)]) is det.
%	Rename a file.
handle([mv, (.)]) --> % rename a file at a directory
	{	get_string(working_directory, S),
		S \== ""
	},
	rename(S, S)
	;
	peek("**** directory not found. ****\n").

%!	handle(+Action:[mv, *]) is det.
%	Move files at the source directory to the
%	target directory with renaming.

handle([mv, *])	-->
	{	get_string(working_directory, S),
		S\== "",
		get_string(target_directory, T),
		T\== ""
	},
	region,
	paragraph,
	remove([]),
	maplist(trim_nl_mv(S,T)),
	insert("\n"),
	overwrite
	;
	peek("**** directory not found. ****\n").

%!c handle([rename]) is det.
%   Rename a file.

handle([rename]) --> handle([mv, (.)]).

%!	handle(+Action:[rename, *]) is det.
%	Rename multi files.
handle([rename, *]) -->
	{	get_string(working_directory, S),
		S\==""
	},
	region,
	paragraph,
	remove([]),
	maplist(trim_nl_mv(S,S)),
	insert("\n"),
	overwrite
	;
	peek("**** directory not found. ****\n").

%!	handle(+Action:[wd]) is det.
%	Get working directory.
%
handle([wd])	--> {working_directory(X, X)}, peek(X).

%!	handle(+Action:[wd, change]) is det.
%	Change working directory.
handle([wd, change])	--> { line_get(Obj),
		      obj_get([line(D0)], Obj),
		      atom_codes(D, D0),
		      working_directory(_, D)
		    }.

% !! Experimental !!
%	handle([doc, latex]) is det.
%   under debugging.
%
% handle([doc, latex]) --> region,   % @see => C-c-ee
% 	paragraph,
% 	remove([]),
% 	maplist(split),
% 	maplist(remove([])),
% 	pred(([[X, [Y|_]],[X0, Y0]]:-
% 	     maplist(flip(atom_codes), X, X1),
% 	     maplist(expand_file_name, X1, X2),
% 	     append(X2, X0),
% 	     atom_codes(Y1, Y),
% 	     expand_file_name(Y1, [Y0|_]))),
% 	pred(([[X, Y], Y]:-
% 	    doc_latex(X, Y, [public_only(false)]))).

% %!	handle(+Action:[global,set,key]) is det.
% %   Run global-set-key lisp command.
% handle([global,set,key]) --> region, paragraph, maplist(split),
% 	maplist(remove_comment_line),
% 	remove([]),
% 	maplist([[X,Y], done]
% 		:- (herbrand(Y, H),
% 		    elisp:global_set_key(X, H))
% 	       ),
% 	herbrand_opp.

%!	handle(+Action:[global,unset,key]) is det.
%	Run global-unset-key Lisp command.
handle([global,unset,key])	-->
	region_or_line(K),
	{global_unset_key(K)},
	peek(`unset.`).

%!  handle(+Action:[pldoc]) is det.
% View the source in the current buffer
% as an html file generated by the pldoc library.

handle([pldoc]) -->
	{ Doc_html = 'TMPPLDOC.html',
	  atomics_to_string(['~/public_html/', Doc_html], Local_html),
	  expand_file_name(Local_html,[HTML|_]),
	  lisp(list('default-directory', 'buffer-name'()), List),
	  atomics_to_string(List, File_source_name),
	  open(HTML, write, Out_stream),
	  set_output(Out_stream),
	  pldoc_html:doc_for_file(File_source_name,
							  [edit(false),
							  public_only(false)]),
	  close(Out_stream),
  	  getenv(user, User_name),
	  sh(open(-a('Safari'),
				"http://localhost/"
				+ "~"
				+ User_name
			    + "/"
				+ Doc_html))
	}.

		/***************************
		*     make-reftex-label    *
		***************************/

tex_command(Comm, Arg)--> w(".*\\"),
						  w(".*", Comm0),
						  "{",
						  w(".*", Arg0),
						  "}",
						 {	string_codes(Comm, Comm0),
							string_codes(Arg, Arg0)
						  }.
%
reftex_label_prefix("subsection", "sec").
reftex_label_prefix(S, Pref):- string_length(S, L),
							   (	L =< 3
							   ->	Pref = S
							   ;	sub_string(S, 0, 3, _, Pref)
							   ).

%
make_reftex_label("begin", Beg, Rem,
				  ["\n\\label{", Pref, ":", Rem0, "}"]):- !,
		reftex_label_prefix(Beg, Pref),
		trim_white(Rem0, Rem, []).
make_reftex_label(Comm, Arg, _,
				  ["\n\\label{", Pref, ":", Arg, "}"]):-
		reftex_label_prefix(Comm, Pref).

%
handle([reftex, label])--> line,
						tex_command(Comm, Arg),
						make_reftex_label(Comm, Arg).

% \section{world}
% \label{sec:world}

% \begin{thm} ピタゴラスの  定理
% \label{thm:ピタゴラスの  定理}
% \label{thm:ピタゴラスの 定理}

		/*****************************************
		*     helper predicates for handle/4.    *
		*****************************************/

%  \C-l  help  (for help)

%c trim_white(+X:codes, -Y:codes) is det.
%	Trim white codes from both ends of X as long as possible,
%	and unify Y with the remainder of X.
% ?- ejockey:trim_white(` \t/a\tb c/ \t`, P),
% ?- ejockey:trim_white(` \t/a\tb c/ \t`, P), basic:smash(P).
trim_white(X, Y):- trim_white(Y, X, []).
%
trim_white(A) --> wl("[\s\t]*"), w(".*", A), wl("[\s\t]*").
%
trim_white_nl(X, Y):- trim_white_nl(Y, X, []).
%
trim_white_nl(A) --> wl("[\s\t\n]*"), w(".*", A), wl("[\s\t\n]*").

%c trim_white_prefix(+X:codes, -Y:codes) is det.
%	Trim white codes of the prefix of X,
%	and unify Y with the remainder of X.
trim_white_prefix --> wl("[\s\t]*").


%c catch_once(+G:goal, +A:term, -R:term) is det.
%	Unify R with A if G is true, with E if exception E is thrown
%	from a child process of G, and fail if G fails.

catch_once(X, A, R):- catch((once(X), R=A), E, (R = E)), !.
catch_once(_, _, fail).

%c line(_, -L:codes) is det.
%	Get the codes of the current line with  the cursor on.
line(_, Line) :- (line_get(I)), obj_get([line(Line)], I, _).

%c partial_match(As:list, B:atom) is det.
%	True if some atom in As is a prefix atom of B.

% ?- ejockey:partial_match([reg, a], region).
partial_match(Atoms, Fullname):-
	once((
	member(Shortname, Atoms),
	atom(Shortname),
	sub_atom(Fullname, 0, N, _, Shortname),
	N>0)).


% %c select_phrase(+C:cond, +P:phrase, +Q:phrase) is det.
% %	Conditional phrase depending on arguments abbreviation;
% %	Use default unless otherwise being specified.
% select_phrase(Cond, P, _) --> {call(Cond)}, !, phrase(P).
% select_phrase( _, _, Q)	  -->  phrase(Q).

%c region_or_line(As:list, ?X, ?Y) is det.
%	Apply region/2 or line/2 to  X, Y depending on X.

region_or_line([]) --> !, region.
region_or_line(X) --> {partial_match(X, region)}, !, region.
region_or_line(_) --> line.

% region_or_line(X) --> select_phrase(partial_match(X, region), region, line).

%c trim_nl_mv(+S, +T, +X, -Y) is det.
%	Move a file over directories.
trim_nl_mv(S, T) --> trim_nl(L, R),
	handle_mv(S, T),
	peek(X, [L, X, R]).

%c rename(S:directory, T:directory, +X:codes, -Y:codes) is det.
%	Move a file over directories with specified new name.
rename(S, T)-->  set_mark_region,
	region,
	trim_nl_mv(S, T),
	overwrite.

%c indent_region(+C:code, +N:int, +X, -Y) is det.
%	Indent the region by padding the code C  N times.

indent_region(CharCode, N) -->  region,
	split,
	{ 	length(Indent, N),
		maplist(=(CharCode), Indent)
	},
	maplist(pred(Indent,
		     [[], []]
		    &
		     [X, [Indent|X]])),
	insert('\n'),
	overwrite.

%c handle_open_relative(+P:codes) is det.
%	Open the object located at P given as a path relative
%	to the working directory.

handle_open_relative(Line) :-
		get_string(working_directory, Path),
		(	Path \== ""
		->	PathStr = Path
		;	PathStr = ""
		),
		atomics_to_string([PathStr, /, Line], X),
		sh_core(open(X)).

%c remove_comment_line(X:codes, Y:codes) is det.
%	Remove the comment lines from X, and Unify Y with the
%	remaining.

% ?- ejockey:remove_comment_line([`%abc`, `%xyz`, `%hello`], R).
remove_comment_line([],[]).
remove_comment_line([[0'%|_]|R], R0):- !, remove_comment_line(R, R0). %'
remove_comment_line([X|R], [X|R0]):- remove_comment_line(R, R0).

%c handle_mv(+S, +T, +X, -Y) is det.
%	Move a file from directory S to T, whose source and target names
%	are coded in X.

handle_mv(S_dir, T_dir) -->
	trim_nl(Left, Right),
	pred([X, Y]:- foldr(  % : ===>  @
		pred(   [0':,  U, [0'@  | U]]
			&
				[0'/,  U, [0'-  | U]]
			&
				[A,    U, [A|U]] ) ,
		X, [], Y)),
	mv_over_directory(S_dir, T_dir),
	peek(A, [Left, A, Right]).


%c A -->>  B   is a genral form of rules, tentatively called a `DCGX' (DCG extended) rule.
%	Syntactically, A and B must be prolog terms such that A --> B forms a DCG rule.
%	This rule is translated like a DCG rule, but into a predicate H that acts on contextual
%	object of the form (X, E), which is called here a `state'.
%	Procedually, H acts on states as a state transition action, so that we write
%
%	                H
%		(X, E) ~~> (X', E')
%
%	for H((X, E), (X', E')).
%
%	Let H1, ..., Hn be actions for instances of the lefthand side of rules defined
%	by '-->>' rules and (X0, E0) given an initial contextual objects. Then,  a sequence (H1,...,Hn)
%	acts on a state (X0, E0) as an intial state, and then  produce a next state (X1, E1),
%	and does successively so on  like this with a final state (Xn, En).
%
%	                 H1           H2      Hn
%		(X0, E0) ~~> (X1, E1) ~~> ... ~~> (Xn, En).
%

%c mv_at_directory(+L:directory, +S:state, -S0:state) is det.
%	Rename a file under L, whose  source and target names are
%	coded in the state S.

mv_at_directory(L) -->> dcl([dir(L)]),
	paragraph,
	remove([]),
	maplist(split),
	maplist(remove([])),
	remove([]),
	obj(obj_get([dir(F)])),
	maplist(pred(F, ([[X|Y], "renamed."]:-
		maplist(split(` `), Y, Y0),
		maplist(remove([]), Y0, Y1),
		maplist(insert("\\ "), Y1, Y2),
		insert("@", Y2, Y3),
		file_extension(Ext, X, _),
		sh(mv(-i, F + X, F + Y3 + Ext)))
		&
		([P,Q]:- insert("\n", P, Q)))),
	insert("\n").

%c  mv_over_directory(+L:directory, +M:directory, +S:state, -S0:state) is det.
%   Move files over from L to M. The source and target name of a file
%   are in the given state S.

mv_over_directory(L, M) -->> dcl([dir(L), dir_target(M)]),
	paragraph,
	maplist(split),
	maplist(remove([])),
	remove([]),
	obj(obj_get([dir(F), dir_target(G)])),
	maplist(pred([F,G],
			([[X|Y], "Renamed and moved."]:-
				file_name(Y, Y0),
				atomics_to_string([G,/, Y0], Y1),
		 		file_extension(Ext, X, _),
				modify_file_name(Y1, 0, Ext, Y2),
				atom_codes(X0, X),
				atomics_to_string([F,/, X0], X1),
				rename_file(X1, Y2))
			&
			([P,Q]:- insert("\n", P, Q))
		    )
	       ),
	insert("\n").

%c file_name(+X:text, -Y:atom) is det.
%	Concatenate a list X of blocks of codes into an atom Y
%	with '@' as a block separator character.

file_name --> insert(`@`), flatten, flip(string_codes).

%c modify_file_name(+F:file_name, +I:integer, +E:extension, -G:File_name) is det.
%	Modify the file name F to G by adding a minimum integer suffix J >= I
%	to F when F conflicts with an existing one so that G does not so, otherwise,
%	unify G with F.

% ?- ejockey:modify_file_name('emacs-jockey', 0, '.pl', G).
% ?- ejockey:modify_file_name('~/Desktop/test', 0, '.bib', G).
% ?- ejockey:modify_file_name('~/Desktop/test', 1, '.bib', G).

modify_file_name(F, 0, Ext, G):- !,
	atomic_list_concat([F, Ext], F0),
	(	exists_file(F0)
	->	modify_file_name(F, 1, Ext, G)
	;	G = F0
	).
modify_file_name(F, I, Ext, G):- atom_number(A, I),
	atomic_list_concat([F, @, A, Ext], F0),
	(	exists_file(F0)
	->	J is I+1,
		modify_file_name(F, J, Ext, G)
	;	G = F0
	).

%c file_extension(-Ext:atom, +P:codes, -Q:codes) is det.
%	Unify Ext with a file extension codes (including the '.' character) of
%	P, and Q with the remainder prefix of P.  If no extension of P is found,
%	unify Ext and Q with the empty atom '' and P, respective.

% ?- ejockey:file_extension(X, `abc/.efg/a.b.c`, R).
% ?- ejockey:file_extension(X, `abc/.efg/a.b.c/x`, R).

file_extension(Ext) --> w(*(.)),  ".",  wl("[^\\./]*", X), end_of_list, !,
	{ atom_codes(Ext, [0'. | X]) }.		%'
file_extension('') --> [].

%c insert_tab_nl(+N:int, +I:int, +T:list, -T0:list) is det.
%	Insert tab codes or newline codes between each successive elements
%	of T, and unify T0 with it, so that  writing all elements of the list T0
%	in order shows up an array of raws of  N-elements, provided that I = N.

% ?- ejockey:insert_tab_nl(3, 3, [a,b,c,d,e], R).
insert_tab_nl(_, _, [], []).
insert_tab_nl(N, 0, [X|Y], [[X,'\n']|Y0]):- !, insert_tab_nl(N, N, Y, Y0).
insert_tab_nl(N, J, [X|Y], [[X,'\t']|Y0]):- J0 is J-1, insert_tab_nl(N, J0, Y, Y0).

%c insert_nl(+X:list, -Y:list) is det.
%	Shorthand for insert(`\n`, X, Y).

insert_nl --> insert(`\n`).

			/****************************************
			*     listing tex command sequences.    *
			****************************************/


%! 	handle(+Action:[list, tex, cs]) is det.
%   Listing tex command sequeces.
%
handle([list, tex, cs]) --> region,
	texparse,
	list_texcs,
	sort,
	insert("\n").

%
list_texcs_file(File, R):- read_file_to_codes(File, R0, []),
	texparse(R0, R1),
	list_texcs(R2, [], R1, []),
	sort(R2, R).
%
list_texcs(X, Y):-  list_texcs(Y, [], X, []).

%
list_texcs([A|X], Y)	--> [cs(A)], !, list_texcs(X, Y).
list_texcs([F|X], Y)	--> [env(F, B)], !,
	{ list_texcs(X, X0, B, []) },
	 list_texcs(X0, Y).
list_texcs(X, Y)	--> [L], { listp(L) } , !,
	{ list_texcs(X, X0, L, []) },
	list_texcs(X0, Y).
list_texcs(X, Y)	--> [_], !, list_texcs(X, Y).
list_texcs(X, X)	--> [].

% handle([X]) --> phrase(X).


		/************************************************
		*     bi-directional converter for file name    *
		%     with dakuten characters                   *
		************************************************/

%c dakuten_convert(?X:text, ?Y:text) is det and bi-directional.
%	Replace each 'dakuten' and 'semi-dakuten' (voiced sound mark) ligature with
%	the one character in utf8 encoding, and unify Y with the result so that Y is from
%	from such ligatures; and vice versa. Note that copy-paste of Japanese file names
%	of ligature free in Finder may yield codes that has (semi-)dakuten ligatures,
%	which may cause troubles.

% [2013/09, 2014/12]
%  ex. "ば" <==> "ば”   (bi-directional)

% ?- ejockey:dakuten_convert("プロジェクト", Y), ejockey:dakuten_convert(X, Y).
% ?- ejockey:dakuten_convert(`プロジェクト`, Y), ejockey:dakuten_convert(X, Y).
% ?- ejockey:dakuten_convert('プロジェクト', Y), ejockey:dakuten_convert(X, Y).
% ?- ejockey:dakuten_convert("プロジェクトプロジェクト", Y), ejockey:dakuten_convert(X, Y).
% 濁点 '゙'	半濁点 '゚'

dakuten_convert(X, Y):- var(Y), !,
	string_chars(X, U),
	once(convert_chars(U, V)),
	string_chars(Y, V).
dakuten_convert(X, Y):-
	string_chars(Y, V),
	once(convert_chars(U, V)),
	string_chars(X, U).

convert_chars([], []).
convert_chars([X, Y|R], [Z|S]):- conversion_table(Y, D, E),
	chars_table_check(X, D, E, Z),
	convert_chars(R, S).
convert_chars([X|R], [X|S]):- convert_chars(R, S).

%
chars_table_check(X, [X|_], [Z|_], Z).
chars_table_check(X, [_|U], [_|V], Z):- chars_table_check(X, U, V, Z).

% conversion_table(a, X, Y) means that  ba <==> c  for each b in X and c in Y.
conversion_table('゙',
		['か', 'き', 'く', 'け', 'こ',
		 'さ', 'し', 'す', 'せ', 'そ',
		 'た', 'ち', 'つ', 'て', 'と',
		 'は', 'ひ', 'ふ', 'へ', 'ほ'],
		['が', 'ぎ', 'ぐ', 'げ', 'ご',
		 'ざ', 'じ', 'ず', 'ぜ', 'ぞ',
		 'だ', 'ぢ', 'づ', 'で', 'ど',
		 'ば', 'び', 'ぶ', 'べ', 'ぼ']).
conversion_table('゙',
		['ウ',
		 'カ', 'キ', 'ク', 'ケ', 'コ',
		 'サ', 'シ', 'ス', 'セ', 'ソ',
		 'タ', 'チ', 'ツ', 'テ', 'ト',
		 'ハ', 'ヒ', 'フ', 'ヘ', 'ホ'],
		['ヴ',
		 'ガ', 'ギ', 'グ', 'ゲ', 'ゴ',
		 'ザ', 'ジ', 'ズ', 'ゼ', 'ゾ',
		 'ダ', 'ヂ', 'ヅ', 'デ', 'ド',
		 'バ', 'ビ', 'ブ', 'ベ', 'ボ']).
conversion_table('゚',
		['は', 'ひ', 'ふ', 'へ', 'ほ'],
		['ぱ', 'ぴ', 'ぷ', 'ぺ', 'ぽ`']).
conversion_table('゚',
		['ハ', 'ヒ', 'フ', 'ヘ', 'ホ'],
		['パ', 'ピ', 'プ', 'ペ', 'ポ']
	      ).



%c trim_nl(-L:codes, -R:codes, +X:codes, -Y:codes) is det.
%	Trim successive new line codes from both ends of X  as long as possible,
%	and unify Y with the remainder of X.

% ?-ejockey:trim_nl(L, R, `abc`, Y).
% ?-ejockey:trim_nl(L, R, `\n\n\n`, Y).
% ?-ejockey:trim_nl(L, R, `\nabc\n`, Y).
% ?-ejockey:trim_nl(L, R, `\n\nabc\n\n`, Y).
% ?-ejockey:trim_nl(L, R, `\n\n向井\n国昭\nabc\n\n`, Y).

%
trim_nl(L, R) --> wl(*("\n"), L),
	w(*(.), Y),
	wl(*("\n"), R),
	end_of_list,
	peek(Y).

%
end_of_list([], []).

%c meta_handle(?X, -Y) is det.
%	Parse the first line of the region for a handle command,
%	and apply the command to the rest of the region.

meta_handle --> region,
	w("[^\n]*$", L),
	{ parse_line(X, L, []),
	  maplist(atom_codes, A, X)
	},
	pred([A, L], [U, V]:-
		once(find_handle_call(A, L, U, V))).

%c parse_line(+X:list, +Y:codes, -Z:codes) is det.
%	Unify X with a list of (S-expression) tokens that
%	appears in the deference between Y and Z.

% ?- ejockey:parse_line(X, `a b c`, []).
% ?- ejockey:parse_line(X, `a "b c""d e"`, []).
% ?- ejockey:parse_line(X, `a "b c"'d \\"e'`, []).
% ?- ejockey:parse_line(X, `'d\\e'`, []).
% ?- ejockey:parse_line(X, `'d\e'`, []).
% ?- ejockey:parse_line(X, `'d\\\\e'`, []).
% ?- ejockey:parse_line(X, `"d\\\\\e"`, []).
% ?- ejockey:parse_line(X, `"a"`, []).

parse_line(X) --> wl("[\s\t]*"), parse_line0(X).

parse_line0([A|X]) --> token(A), !, parse_line(X).
parse_line0([]) --> [].


% ?-coalgebra:show_am("\"([^\"\\\\]|(\\\\.))*\"" | "'([^'\\\\]|(\\\\.))*'" | "[^ \t\"']+").


%c token(-X:token, +Y:codes, -Z:codes) is det.
%	Unify X with a token in S-expression for the difference betwee Y and Z.

% ?- ejockey:token(X, `abcd  `, Y).
% ?- ejockey:token(X, `"ab\\\"d"`, Y), smash(X).
%@ "ab\"d"
% ?- ejockey:token(X, `"ab\\\"c\\\"d"`, Y), smash(X).
%@ "ab\"c\"d"

token(X) --> wl( "\"([^\"\\\\]|(\\\\.))*\""
	       | "'([^'\\\\]|(\\\\.))*'"
	       | "[^\s\t\"']+",
		X).

%c prolog_identifier(N:, X:codes, Y:codes) is det.
%	Unify N with a list of codes
%	such that N is the longest prolog_identifier prefix of X,
%	and Y with the remaining suffix of X.
%

prolog_identifier(N) --> wl("[a-z][a-zA-Z0-9_]*", N, []).

%c keyword(N:, X:codes, Y:codes) is det.
%	Unify N with a list of codes
%	such that N is the longest keyword prefix of X,
%	and Y with the remaining suffix of X.
%

keyword(N) --> wl("[a-zA-Z][a-zA-Z0-9_]*", N, []).

%c collect_tokens(+W:type, +X:codes, -Y:tokens) is det.
%	Collect tokens in X that satisfies W, and
%	unify Y with it.

collect_tokens(W, X, Y):- collect_tokens(W, Y, [], X, []).

%
collect_tokens(W, X, Y) --> [_], collect_tokens(W, X, Y).
collect_tokens(_, X, X)-->[].

% % [2013/10/07] To escape special characters of the file name
% % in order to pass it to sh/1.
% % ?- ejockey:escape_shell_char(`a : (b)`, R), atom_codes(A, R).
% %@ A = 'a \\@ \\(b\\)' .

escape_shell_char(X, Y):-
  foldr(pred(	[0'(,  U, [0'\\,    0'(  | U]	] &		%'
		[0'),  U, [0'\\,    0')  | U]	] &
		[0'\', U, [0'\\,    0'\' | U]	] &
		[0':,  U, [0'\\,    0'@  | U]	] &
		[0'/,  U, [0'\\,    0'@  | U]	] &
		[A,    U, [A|U]			]
	    ),
	X, [], Y).

% ?- ejockey:remove_leading_comment_chars(`% %@ ?- a, \n %  b.\n`, X).
remove_leading_comment_chars(X, Y) :-
	remove_leading_comment_chars(Y, [], X, []).

%
remove_leading_comment_chars(X, Y) -->
	wl("([% \t]|(%@*)|(\\?-))*"),
	wl("[^\n]*", X, X0),
	remove_leading_comment_chars_continue(X0, Y).
%
remove_leading_comment_chars_continue([0'\n|X], Y) --> "\n",  %' %
	remove_leading_comment_chars(X, Y).
remove_leading_comment_chars_continue(X, X) --> [].

%
comment([], []).
comment(X, ["% "|X]).

uncomment --> wl("%+ ?"|[]).

module_name(Codes, Name):- once(module_name(Name, Codes, _)).

module_name(Name) --> wl("[\s\t]*:-[\s\t]*"),
		      "module(",
		      wl("[^,\s\t]+", Name).
module_name(????) --> [].

%
white_filler --> wl("[\s\t\n]*").
%
non_white_line(X):- \+ white_filler(X, []).

%c to_ascii_space(+X, -Y) is det.
%
%  Convert wide charsspaces in UTF-8 to Ascii sapce chars.

% ?- C = `ab　cd`, ejockey:to_ascii_space(`ab　cd`, R).
%@ C = [97, 98, 12288, 99, 100],
%@ R = [97, 98, 32, 99, 100].

to_ascii_space(X, Y) :- once(to_ascii_space(Y, [], X, [])).

to_ascii_space([0'\s|X], Y) --> "　", to_ascii_space(X, Y).  % '
to_ascii_space([C|X], Y)--> [C], to_ascii_space(X, Y).
to_ascii_space(X, X)--> [].
