:- module(expr, [expr/2, expr/3, expr/4, expr/5, expr/6]).
% expr : a  module for handy parsing on regular expressions.
%       2006-2011
%	2013 September
%	@author  Kuniaki Mukai
%
%	expr(?E:term, +A:list, @B:list) is nondet.
%       True if d-lists A-B is an instance of E.
%
%	expr(?E:term, -X:list, @Y:list, +A:list, @B:list) is nondet.
%       True if both d-lists A-B and X-Y denote a same list
%	which is an instance of E, where the symbol P-Q denotes
%	the list D such that append(D, Q, P) holds.

% 1. Regular expressions.
% 1.1  Basic regular expressons.
%	[....] 		(Prolog list, may be a variable)
%	dot		( . )
%	dot([...])	( [...] )
%	out([...])	( [^...] )
%	*(E)		(Kleene's star)
%	+(E)		(Kleene's star without adding null)
%	E + E0		(binary concatenation)
%	E^N		( E + E +...+ E (N times) )
%	?(E)		(E or null)
%	char(L)		(char_type in L)
%	code(L)		(code_type in L)
%	skip		(skip to the end)
%	nil		(the list is nil)
%	rest		(to feed the rest of the list)
%	E | E0		(disjunction, union)
%	E & E0		(conjunction, intersection)
%	\+ E		(complement of E)
%
% 1.2	special
%	@(A)		(phrase(A))   (deleted)
%
% 2. A query terminates for Kleene's closures *(E) of
%	a regular expression E if E is proper, where E is proper
%	if the null list is not an instance of E.  The user is responsible
%	for making sure that his  Kleene's closures are proper.
%
% 3. For the Kleene closure *(E) of E, the parser works
%	in a greedy way in finding strings that matches *(E).
%
% 4. Very simple, short and straightforward implementation in DCG.
%
% expr/2.  for compatibility)
expr	--> [].
expr	--> [_], expr.

% expr/3,4,5.  for compatibility)
expr(E)		--> kleene(E, min).
expr(E, X)	--> kleene(E, min, X, []).
expr(E, X, Y)	--> kleene(E, min, X, Y).
expr(E, M, X, Y)	--> kleene(E, M, X, Y).

% Sample queries:

% ?- expr:kleene(X, min, [a,b,c], R).
% ?- expr:kleene(X, max, [a,b,c], R).
% ?- expr:kleene( *(dot), max, [a,b,c], R).
% ?- expr:kleene( (X + X + X),min, [a,b,c,a,b,c,a,b,c], R).
% ?- expr:kleene( (X + X + X),max, [a,b,c,a,b,c,a,b,c], R).
% ?- expr:kleene( (X + X)^2, min,[a,a,a,a,a,a], R).
% ?- expr:kleene( (X + X)^2, max,[a,a,a,a,a,a], R).
% ?- expr:kleene( (X + X)^2, max,[a,a,a,a,a,a], R).
% ?- expr:kleene( (*([a])) + X + [c] + (*([c])), min,[a,a,b,b,c,c], R).
% ?- expr:kleene( (*([a])) + X + [c] + (*([c])), max,[a,a,b,b,c,c], R).
% ?- expr:kleene( (+([a,a])) + X + (+([c,c])), min, [a,a,b,c,c], R).
% ?- expr:kleene( (+([a,a])) + X + (+([c,c])), max, [a,a,b,c,c], R).
% ?- expr:kleene( (+dot) + (+dot), min, [a,b,c,d], R).
% ?- expr:kleene( (+dot) + (+dot), max, [a,b,c,d], R).
% ?- expr:kleene( &(*([a]), *([b])), min, [a,b], R).
% ?- expr:kleene( &(*([a]), *([b])), max, [a,b], R).
% ?- expr:kleene( *([a]) + (*(dot)), min, [a,a,a,d,e,f], R).
% ?- expr:kleene( *([a]) + (*(dot)), max, [a,a,a,d,e,f], R).

% kleene/4
kleene([], min)		--> [].
kleene([C|R], min)	--> !, [C], kleene(R, min).
kleene([C|R], max)	--> [C], kleene(R, max).
kleene([], max)		--> !, [].
kleene(X+Y, M)		--> kleene(X, M), kleene(Y, M).
kleene(dot, _)		--> [_].
kleene(dot(L), _)	--> [C], {kleene_dot(C,L)}.
kleene(out(L), _)	--> [C], {\+kleene_dot(C,L)}.
kleene(*(E), M)		--> kleene_star(E, M). % E must be proper.
kleene(+(E), M)		--> kleene(E, M), kleene_star(E, M).
kleene(?(E), M)		--> kleene(E, M); [].
kleene(E^N, M)		--> kleene_repeat(N,E, M).
kleene(char(L), _)	--> [C], {char_type(C,L)}.
kleene(code(L), _)	--> [C], {code_type(C,L)}.
kleene((E|E0), M)	--> kleene(E, M); kleene(E0, M).
kleene(\+(E), M)	-->  \+ kleene(E, M).
kleene(&(E,E0), M, A,B) :-  kleene(E,M,A,B), kleene(E0,M,A,B).
kleene(nil,_,[],[]).
kleene(rest,_,_,[]).
kleene(min(E), _)	--> kleene(E, min).
kleene(max(E), _)	--> kleene(E, max).

%
kleene_star(E, min)	--> !, ([]; kleene(E, min), kleene_star(E, min)).
kleene_star(E, _)	--> ( kleene(E, max), kleene_star(E, max) ; [] ).

% kleene/6
kleene([], min, X,X)	--> [].
kleene([C|R],min,[C|X],Y) --> !, [C], kleene(R,min,X,Y).
kleene([C|R],max,[C|X],Y) --> [C], kleene(R,min,X,Y).
kleene([], max, X, X)	--> !, [].
kleene(E+E0,M,X,Y)	--> kleene(E,M,X,X0), kleene(E0,M,X0,Y).
kleene(*(E),M,X,Y)	--> kleene_star(E,M,X,Y).
kleene(+(E),M,X,Y)	--> kleene(E,M,X,X0), kleene_star(E,M,X0,Y).
kleene(dot,_,[C|X],X)	--> [C].
kleene(dot(L),_,[C|X],X)	--> [C], {kleene_dot(C,L)}.
kleene(out(L),_,[C|X],X)	--> [C], {\+kleene_dot(C,L)}.
kleene(code(L),_,[C|X],X)	--> [C], {code_type(C,L)}.
kleene(char(L),_,[C|X],X)	--> [C], {char_type(C,L)}.
kleene(E^N,M,X,Y)		--> kleene_repeat(N,E,M,X,Y).
kleene(?(E),M,X,Y)	--> kleene(E,M,X,Y); { X = Y }.
kleene((E|E0),M,X,Y)	--> kleene(E,M,X,Y); kleene(E0,M,X,Y).
kleene(\+(E),M,X,Y)	--> \+ kleene(E, M,X,Y).
kleene(max(L),_,X, Y)	--> kleene(L,max,X,Y).
kleene(min(L),_,X,Y)	--> kleene(L,min,X,Y).
kleene(&(E,E0),M,X,Y,A,B) :- kleene(E,M,X,Y,A,B), kleene(E0,M,X,Y,A,B).
kleene(nil,_,X,X,[],[]).
kleene(rest,_,A,B,X,[])   :- append(X, B, A).

%
kleene_star(_, min, X, X) --> [].
kleene_star(E, min, X, Y) --> !, kleene(E, min, X, X0),
	{compound(X)},
	kleene_star(E, min, X0,Y).
kleene_star(E, max, X, Y) --> kleene(E, max, X, X0),
	{compound(X)},
	kleene_star(E, max, X0,Y).
kleene_star(_,max,X,X) --> [].

%  compare for code/char type
kleene_dot(A,[A|_])	:- !.
kleene_dot(A,[B-C|_])	:- B @=< A, A @=< C, !.
kleene_dot(A,[_|L])	:- kleene_dot(A,L).

%
kleene_repeat(0,_,_)	--> [].
kleene_repeat(N,E,M)	--> {N>0}, kleene(E, M),
	{N1 is N-1},
	kleene_repeat(N1,E,M).

%
kleene_repeat(0,_,_,X,X)--> [].
kleene_repeat(N,E,M,X,Y)--> {N>0}, !,
	kleene(E,M,X,X0),
	{N1 is N-1},
	kleene_repeat(N1,E,M,X0,Y).
