:- module(sudoku_strong, []).

% (setq module-query  "qcompile(util('sudoku-strong')), module(sudoku_strong).")
% ?- qcompile(util('sudoku-strong')), module(sudoku_strong).

% ?- module(sudoku_strong).
% Hi all,

% in the most recent git version of SWI 5.7, the all_distinct/1 constraint
% has become significantly stronger. all_distinct/1 was always stronger
% than all_different/1, for example:

% ?-  Vs = [X,Y,Z], Vs ins 1..2, all_different(Vs).
%   %@ Vs = [X, Y, Z],
%   %@ X in 1..2,
%   %@ all_different([X, Y, Z]),
%   %@ Y in 1..2,
%   %@ Z in 1..2.

% whereas all_distinct/1 can detect the inconsistency:

%   :- Vs = [X,Y,Z], Vs ins 1..2, all_distinct(Vs).
%   %@ false.

% Now, all_distinct/1 is even stronger (arc-consistent). As an example,
% consider the recently posted Sudoku program:

:- use_module(library(clpfd)).
:- use_module(util(matrix)).

sudoku(Rows) :-
      length(Rows, 9), maplist(length_(9), Rows),
      append(Rows, Vs), Vs ins 1..9,
%     maplist(all_different, Rows),
      maplist(all_distinct, Rows),
      matrix:transpose_matrix(Rows, Columns),
% 	  maplist(all_different, Columns),
	  maplist(all_distinct, Columns),
      Rows = [A,B,C,D,E,F,G,H,I],
      blocks(A, B, C),
	  blocks(D, E, F),
	  blocks(G, H, I).

length_(L, Ls) :- length(Ls, L).

blocks([], [], []).
blocks([A,B,C|Bs1], [D,E,F|Bs2], [G,H,I|Bs3]) :-
      all_distinct([A,B,C,D,E,F,G,H,I]),
      blocks(Bs1, Bs2, Bs3).

problem(3, P) :-
          P = [[1,_,_,_,_,_,_,_,_],
               [_,_,2,7,4,_,_,_,_],
               [_,_,_,5,_,_,_,_,4],
               [_,3,_,_,_,_,_,_,_],
               [7,5,_,_,_,_,_,_,_],
               [_,_,_,_,_,9,6,_,_],
               [_,4,_,_,_,6,_,_,_],
               [_,_,_,_,_,_,_,7,1],
               [_,_,_,_,_,1,_,3,_]].

problem(4, P) :- % from iPhone App.
          P = [[9,_,2,4,_,_,_,_,7],
               [_,_,4,_,_,_,_,1,_],
               [_,_,_,7,_,_,_,9,2],
               [7,5,_,_,_,8,_,_,_],
               [2,_,_,_,_,_,7,_,_],
               [_,_,_,1,7,_,_,2,_],
               [5,_,_,_,2,_,_,_,_],
               [6,_,_,_,_,_,_,7,_],
               [_,_,_,_,_,4,6,3,1]].

problem(5, P) :- % from iPhone App.
          P = [[_,_,2,4,_,_,_,_,7],
               [_,_,4,_,_,_,_,1,_],
               [_,_,_,7,_,_,_,9,2],
               [7,5,_,_,_,8,_,_,_],
               [2,_,_,_,_,_,7,_,_],
               [_,_,_,1,7,_,_,2,_],
               [5,_,_,_,2,_,_,_,_],
               [6,_,_,_,_,_,_,7,_],
               [_,_,_,_,_,4,6,3,1]].



% ?- qcompile(util('sudoku-strong')), module(sudoku_strong).
% ?- help(labeling).

% ?- problem(4, S), sudoku(S), maplist(writeln, S).

% ?- time((problem(5, S), sudoku(S))), append(S, U),
%	 labeling([], U), maplist(writeln, S).

% ?- time((problem(4, S), sudoku(S))), append(S, U),
%	 labeling([], U), maplist(writeln, S).

% ?- findall(S,  (problem(5, S), sudoku(S), append(S, U),
%	 labeling([], U)), V), length(V, L).

% ?- findall(S,  (problem(4, S), sudoku(S), append(S, U),
%	 labeling([], U)), V), length(V, L).

%  ?- time((problem(4, S), sudoku(S))), append(S, U),

% Document by Markus.
%
% This version is now more than 6 times faster than previously, in fact
% all_distinct/1 is now so strong that the unique solution is found
% without labeling:

% ?- time((problem(3, S), sudoku(S))), maplist(writeln, S).

% The algorithm now used in all_distinct/1 is due to J-C. R辿gin. The price
% to pay for this much smaller search space (compared to all_different/1)
% is slower propagation speed, often significantly slower especially for
% larger sets of variables. I'm currently trying to speed up propagation,
% which mainly consists of finding a maximum matching and the strongly
% connected components of a bipartite graph. Please take a look at
% library(clpfd)'s source if you are interested in helping with this.

% Also, please let me know if you are currently using all_distinct/1 and
% the new filtering algorithm is now too slow for your use case. I will
% restore the previous all_distinct/1 and make the new filtering algorithm
% available as an option for a new all_different/2, or under a different
% predicate, if you consider the current speed a significant regression.

% Independently, I welcome suggestions on how to make options available
% for a new all_different/2. For SICStus compatibility, we could consider
% consistency(domain/bound/value) and on(dom/min/max/minmax/val) to set
% consistency and triggering events. Please let me know what you think.

% Thank you and all the best,
% Markus

% P.S.: Without querying Prolog, how far can you reduce the domains in:

% ?- test(X, Y).
test(Vs,Ds):- length(Vs, 6),
         Ds = [1..2, 2..3, 1\/3, 2\/4, 3..6, 6..7],
         maplist(in, Vs, Ds),
         all_distinct(Vs).
