:- module(turing,[turing/3,turing/4]).
%%
%% A structured programming for Turing machines.
%%
:-dynamic def/2.
:- use_module(util(misc)).
:- use_module(pac(basic)).

run(turing(X,Y),Z):-!,turing(X,Y,Z).
run(turing:history(X,Y),Z):-history(X,Y,Z).



%% For demo.

demo:- prompt(A,''), demo(A,_).

demo(N):- prompt(A,''), demo(A,N).

demo(_,N):- example(N,X), demo_one(N,X), fail.
demo(A,_):- prompt(_,A).

demo_one(N,X):- format("~w  ~w~n",[N,X]),
	atom_to_term(X,Y,V),
	call(Y),
	!,
      (	V=[W, U=H] ->
	writeln(W),nl,
	format("~w=",[U]),
	maplist(writeln,H)
      ; nl, maplist(writeln,V)
      ).

%% Sample Turing machine programs

% example:
def(add, [mark,right,find(@),put(1),find(@),put(space),left,put(@),
	  home,unmark]).

def(q0, q(a,b,r,q1)).
def(q1, halt).

example(1,'turing(q0, [a,a,a], X,H)').
example(2,'turing(while(a, [put(b),right]),[a,a,a,a,a],X,H)').
example(3,'turing([set(x,a), while($x, [put(b),right])],[a,a,a,a,a],X,H)').
example(4,X):- atom_concat('turing(until(space,if(1,[put(0),right],',
			   'if(0,[put(1),right],nop))),[1,1,0,0],X,H)', X).
example(5,'turing(insert_region(a,b),[0,0,0,0,a,1,1,1,b,2,2,2],X,H)').
example(6,'turing([get(x),right,get(y)],[]*[1,0],X,[]-F,H-[])').
example(7,'turing([set(x,a),get(a),right,put($($x))],[]*[1],X,[]-F,H-[])').
example(8,'turing(add,[@,1,1,@,1,1,1,@],X,H)').
example(9,'turing(add,[@,@,@],X,H)').
example(10,'turing(add,[@,1,@,@],X,H)').
example(11,'turing(add,[@,@,1,@],X,H)').
example(12,'turing(shift_right(a),[a,1,1,1]+[2,2,a],X,H)').
example(13,'turing(shift_left(a),[a,1,1,1]+[2,2,a],X,H)').

%%%%
% ?- trace, turing:history(right, [a,b,c], H).

history(X,Y,H):- turing(X, Y,_ , H1),
	insert("<br>\n", H1, H).		% for HTML browser

turing(X,Y,Z) :- turing(X,Y,Z,_).

turing(X,Y,Z,History) :- innerTape(Y,Y1),
	turing(X,Y1,Z,[]-_,U-[]),
	maplist(outerTape, [Y1|U], History).

innerTape(X+Y,X1*Y):-!, reverse(X,X1).
innerTape(X,[]*X):- listp(X),!.
innerTape(X,X).

outerTape(P*Q,(P2 + Q3)):-
	reverse(Q,Q1),
	drop_space(Q1,Q2),
	reverse(Q2,Q3),
	reverse(P,P1),
	drop_space(P1,P2).

drop_space([space|X],Y):-drop_space(X,Y).
drop_space(X,X).

turing(X,Y,FinalTape, A, H):-!, exec(X,Y,P*Q,A,H),!,
	outerTape(P*Q,FinalTape).

movehead(X,0,X,P-P).
movehead(S*[],r,[space|S]*[],[[space|S]*[]|P]-P).
movehead(S*[X|Y],r,[X|S]*Y,[[X|S]*Y|P]-P).
movehead([]*X,l,[]*[space|X],[[]*[space|X]|P]-P).
movehead([X|Y]*Z,l,Y*[X|Z], [Y*[X|Z]|P]-P).

macro_def(q(A,B,r,Q), if(A,[put(B),right,Q])).
macro_def(q(A,B,l,Q), if(A,[put(B),left,Q])).
macro_def(q(A,B,0,Q), if(A,[put(B),Q])).
macro_def(home,until(marked(_),left)).
macro_def(find(A),find_right(A)).
macro_def(find_right(A),until(A,right)).
macro_def(find_left(A),until(A,left)).
macro_def(if(A,T),if(A,T,nop)).
macro_def(unless(A,D),if(A,nop,D)).
macro_def(while(X,Y),if(X,[Y,while(X,Y)])).
macro_def(until(X,Y),unless(X,[Y,until(X,Y)])).
macro_def(insert_region(A,B),
     [left, mark, right,
      until(A, [get(x), mark, find(B), insert($x), home, unmark, right]),
      home, unmark, right
     ]).
macro_def(shift(A),shift_right(A)).
macro_def(shift_right(A),
	  [mark, find_right(A),
	   until(marked(_), [get(shift),right,put($shift),left,left]),
	   unmark,get(shift),right,put($shift),left
	  ]).
macro_def(shift_left(A),
	  [mark, find_left(A),
	   until(marked(_), [get(shift),left,put($shift),right,right]),
	   unmark,get(shift),left,put($shift),right
	  ]).

exec([],X,X,F-F,P-P).
exec([C|R],X,Y,F,H):- exec(C,R,X,Y,F,H).
exec(C,X,Y,F,H):- exec(C,[],X,Y,F,H).

exec(halt,_,X,X,F-F,P-P).
exec(if(A,T,E),R,X,Y,F-G,H):- evalref(A,A1,F), exec_if(A1,T,E,R,X,Y,F-G,H).
exec([],R,X,Y,F,H):- exec(R,X,Y,F,H).
exec([C|D],R,X,Y,F,H):- exec(C,[D|R],X,Y,F,H).
exec(C,R,X,Y,F,H):- macro_def(C,D), exec(D,R,X,Y,F,H).
exec(C,R,X,Y,F,H):- def(C,D),exec(D,R,X,Y,F,H).
exec(C,R,X,Y,F1-F2,P-Q):- exec_basic(C,X,Z,F1-F3,P-P1),
	exec(R,Z,Y,F3-F2,P1-Q).

exec_basic(nop,X,X,F-F,P-P).
exec_basic(left,X,Y,F-F,H):- movehead(X,l,Y,H).
exec_basic(right,X,Y,F-F,H):- movehead(X,r,Y,H).
exec_basic(mark,X,Y,F-F,H):- exec_mark(X,Y,H).
exec_basic(unmark,X,Y,F-F,H):- exec_unmark(X,Y,H).
exec_basic(get(X),Y,Z,F-G,H):- evalref(X,X1,F), exec_get(X1,Y,Z,F-G,H).
exec_basic(put(X),Y,Z,F-G,H):- evalref(X,X1,F), exec_put(X1,Y,Z,F-G,H).
exec_basic(insert(A),X,Y,F-G,H):- evalref(A,A1,F),
	exec_insert(A1,X,Y,F-G,H).
exec_basic(set(A,B),X,X,F-[A1=B1|G],P-P):- evalref(A,A1,F),
	evalref(B,B1,F), delete(F,A=_,G).

evalref($(X),Y,A):-evalref(X,Y1,A), member(Y1=Y,A).
evalref(X,X,_).

exec_if(A,T,E,R,X*[],Y,F,[X*[space]|P]-Q):-
	exec_if(A,T,E,R,X*[space],Y,F,P-Q).
exec_if(A,T,_,R,X*[A|Y],Z,F,H):- exec([T|R],X*[A|Y], Z,F,H).
exec_if(_,_,E,R,X,Y,F,H):- exec([E|R],X,Y,F,H).

exec_get(A, Y*[],Y*[space],F-[A=space|F1],[Y*[space]|P]-P):-
	delete(F,A=_,F1).
exec_get(A, Y*[X|Z],Y*[X|Z],F-[A=X|F1],[Y*[X|Z]|P]-P):- delete(F,A=_,F1).

exec_put(A,Y*[],Y*[A],F-F,[Y*[A]|P]-P).
exec_put(A,Y*[_|Z],Y*[A|Z],F-F,[Y*[A|Z]|P]-P).

exec_insert(A,X*Y,X*[A|Y],F-F,[X*[A|Y]|P]-P).

exec_mark(X*[A|Y],X*[marked(A)|Y],[X*[marked(A)|Y]|P]-P).
exec_mark(X*[],X*[marked(space)],[X*[marked(space)]|P]-P).

exec_unmark(X*[marked(A)|Y],X*[A|Y],[X*[A|Y]|P]-P).
