:- module(gf2, []).
:- use_module(library(apply)).
:- use_module(library(apply_macros)).
:- use_module(zdd('zdd-array')).
:- use_module(util(math)).
:- use_module(util(meta2)).
:- use_module(pac(basic)).
:- use_module(pac(meta)).
:- use_module(util(misc)).
:- use_module(zdd(zdd)).
:- use_module(pac(op)).
:- use_module(pac('expand-pac')).

term_expansion --> pac:expand_pac.

% Via discourse 1969 paper.
% o(2t+1)= q^t PI(i=0, t-1)(q^2t - q^2i)
% o(2t+1)= (q^2t - 1)*o(2t)
% o(2t) = q^iPi(i=1, t-1)(q^2t-q2i)

%!	card_orth(+Q, +N, -V) is det.
%	Q:	2,3,5, ... prime numbers
%	N:	2,3,4,5,...
%	V is unified with the number of O(GF(Q), N).

% ?- time(forall(between(2, 12, N), (card_orth(2, N, V), writeln(orth(N)= V)))).
card_orth(Q, N, V):-
	(	0 =:=  N mod 2 -> card_even_orth(Q, N, V)
	;	card_odd_orth(Q, N, V)
	).
%
card_orth_help(Q, T, I, U, V):- V is U* (Q^(2*T)-Q^(2*I)).
%
card_even_orth(Q, E, V):- T is E//2,
	foldnum(card_orth_help(Q, T), 1-(T-1), 1, U),
	V is Q^T* U.
%
card_odd_orth(Q, O, V):- E is O-1,
	card_even_orth(Q, E, U),
	V is (Q^E-1)*U.

		/***************
		*     Notes    *
		***************/

% For a, b in GF2, 1 and 2 hold.
% 1.  a  = 0  iff  1 + a = 1.
% 2.  ab = 1  iff  a = 1 and b = 1.
% 3.  Orthogonality of matrix A = [ aij ] ( n=< i, j =< n )
% is represented as a system of equations Eij = 0,
% where Eij is a polynomial with indeterminates aij obtained
% from the familiar multiplication of matricies.

% It follows from 1, 2, 3 that the set Orth(n, GF2)
% and the intersection of sol(Eij=0) have the same cardinality.
% where Sol(E) is the set of solutions of E = 0.

		/*****************
		*     Helpers    *
		*****************/

%!	divide_at(+A, +L, -H, -T) is det.
%	A : ZDD atom.
%	L :	sorted list.
%	H,T are unified with lists such that append of reverse(H), [A], and T
%	equals to L.

% ?- divide_at(a, [x,a, y], H, T).
% ?- divide_at(a, [x,y,a], H, T).
divide_at(A, L, H, T):- divide_at(A, L, [], H, T).
%
divide_at(A, [A|T], H, H, T):-!.
divide_at(A, [B|T], H, H0, T0):-
	divide_at(A, T, [B|H], H0, T0).

%!	prefix_pow_rev(+X, +Y, -Z,+S) is det.
%	X : reverse sorted list [a1,..., an].
%	Y : zdd
%	Z is unified with zdd [an, ..., a1]*Y

% ?- #S, prefix_pow_rev([b, a], 1, X, S), psa(X, S).
% ?- #S, prefix_pow_rev([b], 1, X, S), prefix_pow_rev([a], X, Y, S).
prefix_pow_rev([], X, X, _).
prefix_pow_rev([A|As], X, Y, S):-
	cofact(X0, t(A, X, X), S),
	prefix_pow_rev(As, X0, Y, S).

%!	prepare_atom_power(+S) is det.
%!	prepare_atom_power(+Vs, +S) is det.
%	Vs : sorted list of atoms.
%	Create the powerset of Va, Wa for each a in Vs
%	where Va is the set of x in Vs smaller than a,
%	Wa smaller than equal  to a. Afterward, Va and Wa
%	can be accessed by atom_power and atom_power_plus.

% For not repeat to build power set of atom list.
% ?- #S, prepare_atom_power([], X, S).
% ?- #S, prepare_atom_power([a,b,c], X, S), atom_power(b, B, S),
%	atom_power_plus(b, Bplus, S), psa(B, S), psa(Bplus, S).

prepare_atom_power(S):- get_key(vars, Vs, S), prepare_atom_power(Vs, S).

prepare_atom_power(Vs, S):- prepare_atom_power(Vs, _, S).
%
prepare_atom_power([], 1, _).
prepare_atom_power([A|As], P, S):-
	prepare_atom_power(As, Q, S),
	memo(atom_power(A)-Q, S),
	cofact(P, t(A, Q, Q), S),
	memo(atom_power_plus(A)-P, S).

%!	atom_power(+A, -X, +S) is det.
%!	atom_power_plus(+A, -X, +S) is det.
%	A:	atom
%	X is unified with zdd family of subsets
%	 of atoms   less than A			 (for atom_poer)
%				less than equal to A (for atom_power_plus).

atom_power(A, X, S):- memo(atom_power(A)-X, S).
atom_power_plus(A, X, S):- memo(atom_power_plus(A)-X, S).

		/**************************
		*     gf2_system_solver   *
		**************************/

%!	gf2_intern_poly_list_vars(+Ps, -Qs, -D, +S) is det.
%	Ps : list of polynomials.
%	Qs is unified with a list of zdds of polynomials in Ps.
%	D is a difference list for the list of variables appearing in Ps.

% ?- #S, gf2_intern_poly_list_vars([a, b, c], L,  Y-[], S),
%	s_map(psa, L, S).
% ?- #S, gf2_intern_poly_list_vars([*[a,b]], [X],  Y-[], S), psa(X, S).

gf2_intern_poly_list_vars([], [], U-U, _):-!.
gf2_intern_poly_list_vars([P|Ps], [Q|Qs], U-V, S):-!,
	gf2_intern_poly(P, Q, U-W, S),
	gf2_intern_poly_list_vars(Ps, Qs, W-V, S).


%!	preprocess_poly_list(+L, -X) is nondet.
%	L : list of integers.
%	X is unified with a maxmal sublist of L such that 0 is not a member	of X.
%	Fails if 1 is a member of L.

% ?- preprocess_poly_list([0,0,0], X).

preprocess_poly_list([], []).
preprocess_poly_list([0|X], Y):-!, preprocess_poly_list(X, Y).
preprocess_poly_list([X|Y], [X|Z]):- X>1, !, preprocess_poly_list(Y, Z).

		/***************************
		*     orth(GF2, 2) case    *
		***************************/

%!	gf2_count_orth(+N, -C) is det.
%	N: N>0
%	C is unified with the number of orthogonal macricies over GF(2)

% ?- gf2_count_orth(1, C).
% ?- gf2_count_orth(2, C).
% ?- gf2_count_orth(3, C).
% ?- gf2_count_orth(4, C).
% ?- time(gf2_count_orth(5, C)).
% ?- time(gf2_count_orth(6, C)).
% ?- call_with_time_limit(600, time(gf2_count_orth(7, C))).
%@ % 958,026,253 inferences, 119.702 CPU in 122.079 seconds (98% CPU, 8003397 Lips)
%@ C = 1451520.
% Memory 3.08GB

% ?- time(gf2_count_orth(8, C)).
%@ % 77,269,642,066 inferences, 8239.463 CPU in 8300.403 seconds (99% CPU, 9377995 Lips)
%@ C = 185794560.
% 80GB

gf2_count_orth(N, C):- orth_matrix_cond(N, Ps), gf2_solve_poly_system(Ps, C).

%!	gf2_solve_poly(+P, -C) is det.
%!	gf2_solve_poly(+P, -C, -Sols, +S) is det.
%	P: polynomial
%	C is unified with the number of solutions for P = 0.
%	Sols is unified with zdd of the solutions (assignments).

% ?- gf2_solve_poly(a+b, C, Y, S), psa(Y, S).
gf2_solve_poly(Poly, C):- gf2_solve_poly(Poly, C, _, _).
gf2_solve_poly(Poly, C, Sols, S):- gf2_solve_poly_system([Poly], C, Sols, S).

%!	gf2_solve_poly_system(+Ps, -C) is det.
%!	gf2_solve_poly_system(+Ps, -C, -Sols, +S) is det.
%!	gf2_solve_poly_system(+Ps, -Qs, -Vars, -C, -Sols, +S) is det.
%	Ps: a list of polynomials P1, ..., Pn
%	C is unified with the number of solutions of P1=0, ..., Pn=0.
%	Vars is unified with a list of variables appearing in Ps.
%	Qs is unified with a list of zdd of P1, ..., Pn.
%	Sols is unified with a zdd of the solutions (assignments).

gf2_solve_poly_system(Ps, C):- gf2_solve_poly_system(Ps, C, _, _).
%
gf2_solve_poly_system(Ps, C, Sols, S):-gf2_solve_poly_system(Ps, _, _, C, Sols, S).
%
gf2_solve_poly_system(Ps, Qs, Vars, C, Sols, S):-
	#S,
	gf2_intern_poly_list_vars(Ps, Qs, Vars-[], S),
	sort(Vars, Vars0),
	prepare_atom_power(Vars0, S),
	(	preprocess_poly_list(Qs, Rs) ->
		solve_system(Vars0, Rs, Sols, S)
	;	Sols = 0
	),
	card(Sols, C, S).

%!  solve_sytstem(+Vs, +Xs, -Y, +S) is det.
%!  solve_sytstem(+Vs, +Xs, -Y0, -Y1, +S) is det.
%	Vs : a list of variables.
%	Xs : a list of zdd  X1, ..., Xn for polynomimials.
%	Y  is unified with a zdd for the all solutions for X1=0,..., Xn=0.
%	Y0, Y0 are unified with a difference list for the solutions
%	for X1=0,..., Xn=0.

solve_system([], [], 1, _):-!.
solve_system([A|_], [], Pow, S):-!, atom_power_plus(A, Pow, S).
solve_system(Vs, [X|Xs], Y, S):-
	solve_zero(Vs, X, X0, S),
	solve_system(Vs, Xs, X0, Y, S).

%
solve_system(_, _, 0, 0, _):-!.
solve_system(_, [], Y, Y, _):-!.
solve_system(Vs, [X|Xs], Y, Z, S):-
	zdd_slim([Y,X|Xs], [Y0, X0|Xs0], S),
	prepare_atom_power(Vs, S),
	solve_zero(Vs, X0, X1, S),
	zdd_meet(X1, Y0, Y1, S),
	solve_system(Vs, Xs0, Y1, Z, S).

%!	solve_zero(+Vs, +A, +X, -Y, +S) is det.
%	Vs: a list of variables.
%	A: a variable (indeterminate).
%	X: a zdd for a polynomial over Vs.
%	Y is unified with a zdd of all solutions of A*X = 0.

%!	solve_zero(+Vs, +X, -Y, +S) is det.
%	Vs: a list of variables.
%	X: a zdd for a polynomial over Vs.
%	Y is unified with a zdd of all solutions of X = 0.

solve_zero(Vs, A, X, Y, S):-solve_zero(Vs, X, Y0, S),
	atom_power(A, Y1, S),
	cofact(Y, t(A, Y1, Y0), S).
%
solve_zero([], 0, 1, _):-!.
solve_zero([A|_], 0, Y, S):-!, atom_power_plus(A, Y, S).
solve_zero(_, 1, 0, _):-!.
solve_zero(Vs, X, Y, S):- memo(zero(X, Vs)-Y, S),
	(	nonvar(Y) -> true	%, write(.) %
	;	cofact(X, t(A, L, R), S),
		divide_at(A, Vs, Hs, Ts),
		(	L = 0 -> solve_zero(Ts, A, R, Y0, S)  % void left branch.
		;	solve_zero([A|Ts], L, L0, S),
			solve_unit([A|Ts], L, L1, S),
			solve_zero(Ts, A, R, R0, S),
			solve_unit(Ts, A, R, R1, S),
			zdd_meet(L0, R0, U0, S),
			zdd_meet(L1, R1, U1, S),
			zdd_join(U0, U1, Y0, S)
		),
		prefix_pow_rev(Hs, Y0, Y, S)
	).

%!	solve_unit(+Vs, +A, +X, -Y, +S) is det.
%	Vs: a list of variables.
%	A: a variable (indeterminate).
%	X: a zdd for a polynomial over Vs.
%	Y is unified with a zdd of all solutions A*X = 1.

%!	solve_zero(+Vs, +X, -Y, +S) is det.
%	Vs: a list of variables.
%	X: a zdd for a polynomial over Vs.
%	Y is unified with a zdd of all solutions X = 1.
solve_unit(Vs, A, X, Y, S):-solve_unit(Vs, X, Y0, S),
	zdd_cons(Y, A, Y0, S).
%
solve_unit(_, 0, 0, _):-!.
solve_unit([], 1, 1, _):-!.
solve_unit([A|_], 1, Y, S):-!, atom_power_plus(A, Y, S).
solve_unit(Vs, X, Y, S):- memo(unit(X, Vs)-Y, S),
	(	nonvar(Y)-> true	%, write(.);%
	;	cofact(X, t(A, L, R), S),
		divide_at(A, Vs, Hs, Ts),
		(	L = 0 -> solve_unit(Ts, A, R, Y0, S) % void left branch.
		;	solve_zero([A|Ts], L, L0, S),
			solve_unit([A|Ts], L, L1, S),
			solve_zero(Ts, A, R, R0, S),
			solve_unit(Ts, A, R, R1, S),
			zdd_meet(R0, L1, U0, S),
			zdd_meet(R1, L0, U1, S),
			zdd_join(U0, U1, Y0, S)
		),
		prefix_pow_rev(Hs, Y0, Y, S)
	).

		/****************************************************
		*     Read polynomial over GF(2). Count solutions   *
		****************************************************/

%!	gf2_sat_count(+E, -C, +S) is det.
%	C is unified with the number of solutions
%	of E = 0 for GF(2)-polynomical E.

% ?- #S, gf2_sat_count(-(a), N, S).
% ?- #S, gf2_sat_count(a, N, S).
% ?- #S, gf2_sat_count(-(a+a+a+b), N, S).
% ?- #S, N = 1, numlist(1, N, Ns), X<<pow(Ns), gf2_count_solutions(N, X, C, S).
% ?- #S, N = 2, numlist(1, N, Ns), X<<pow(Ns), gf2_count_solutions(N, X, C, S).
% ?- #S, N = 100, numlist(1, N, Ns), X<<pow(Ns), gf2_count_solutions(N, X, C, S).
% ?- #S, N = 10, numlist(1, N, Ns), gf2_sat_count(+Ns, C, S).
% ?- #S, N = 2, findall(*[a(I)], between(1, N, I), L), gf2_sat_count(+L, C, S).
% ?- #S, N = 1000, time((numlist(1, N, Ns), X<<pow(Ns), gf2_count_solutions(N, X, C, S))).

gf2_sat_count(Poly, Card, S):-
	set_key(gf2_atom_count, 0, S),
	time(gf2_intern_poly(Poly, X, S)),
	get_key(gf2_atom_count, N, S),
	format("\n gf2_atom_count = ~w\n", [N]),
	time(gf2_count_solutions(N, X, Card, S)).

%!	gf2_count_solutions(+N, +Poly, -C, +S) is det.
%	N : the number of variables appearing on Poly.
%	Poly : Polynomial over GF2.
%	C is unified with the number of solutions of the equation Poly = 0.

% ?- #S, gf2_intern_poly(a, X, S), gf2_count_solutions(1, X, C, S).
% ?- #S, gf2_intern_poly(a*b, X, S), gf2_count_solutions(2, X, C, S).
% ?- #S, gf2_intern_poly(a+b+c+d, X, S), gf2_count_solutions(4, X, C, S).
% ?- #S, gf2_intern_poly(+[a,b,c,d], X, S), gf2_count_solutions(4, X, C, S).
% ?- #S, N = 12, findall(a(I), between(1, N, I), L), gf2_intern_poly(+L, X, S), gf2_count_solutions(4, X, C, S).
% ?- #S, N = 20, findall(a(I), between(1, N, I), L), gf2_intern_poly(+L, X, S), gf2_count_solutions(N, X, C, S).
% ?- #S, N = 1000, findall(a(I), between(1, N, I), L), time((gf2_intern_poly(+L, X, S), gf2_count_solutions(N, X, C, S))).

gf2_count_solutions(N, X, C, S):- garbage_collect,
	setup_call_cleanup(
	open_memo(S),
	gf2_count_solutions_(N, X, C, S),
	close_memo(S)).

%
gf2_count_solutions_(N, 0, C, _):-!, C is 2^N.
gf2_count_solutions_(_, 1, 0, _):-!.
gf2_count_solutions_(N, X, C, S):- memo(gf2_count_solutions(X, N)-C, S),
	(	nonvar(C) -> true	%,	write(.) % Many hits.
	;	cofact(X, t(_, L, R), S),
		gf2_join(L, R, U, S),
		N0 is N-1,
		gf2_count_solutions_(N0, U, C0, S),
		gf2_count_solutions_(N0, L, C1, S),
		C is C0 + C1
	).

%!	gf2_intern_poly(+E, -V, +S) is det.
%	E: polynomial over GF2.
%	V is unified with E that is read as a ZDD.
%	GF2 rules applied:  e.g. a + a = 0.  a*(b+c)=a*b+a*c.

% ?- #S, X << @gf2_intern_poly(a), psa(X, S).
% ?- #S, X << @gf2_intern_poly(a+a+a+b), psa(X, S).
% ?- #S, X << @gf2_intern_poly(-(a+a+a+b)), psa(X, S).
% ?- #S, X << @gf2_intern_poly(1+ -2*a+b+1), psa(X, S).
% ?- #S, X << @gf2_intern_poly(a*a), psa(X, S).
% ?- #S, X << @gf2_intern_poly((a+b+c)*(c+b+a)), psa(X, S).
% ?- #S, X << @gf2_intern_poly(*[+[a]]), psa(X, S).
% ?- #S, X << @gf2_intern_poly(*[+[a,b], +[c,d]]), psa(X, S).
% ?- #S, X << @gf2_intern_poly(+[a, b] +  *[b, c, d]), psa(X, S).
% ?- #S, X << @gf2_intern_poly(+[a,b]), psa(X, S).
% ?- #S, X << @gf2_intern_poly(*[a,b]), psa(X, S).
% ?- #S, X << @gf2_intern_poly(a*1*(1*a + 1*a)), psa(X, S).

gf2_intern_poly(X, Y, S):- gf2_intern_poly(X, Y, _-[], S).

%
gf2_intern_poly(I, M, V-V,  _)	:- integer(I), !, M is I mod 2.
gf2_intern_poly(-(X), Y, D, S):-!, gf2_intern_poly(X, Y, D, S).
gf2_intern_poly(P-Q, Y, D, S):-!, gf2_intern_poly(P+(-Q), Y, D, S).
gf2_intern_poly(P, Y, U-V, S) :- memo(poly_intern(P)-Y, S),
	(	nonvar(Y) -> U = V	% ,  write(.) % So s0 much hits.
	;	gf2_poly_intern_case(P, Y, U-V, S)
	).
%
gf2_poly_intern_case(X * Y, Z, A-B, S):-!,
	gf2_intern_poly(X, U, A-C, S),
	gf2_intern_poly(Y, V, C-B, S),
	gf2_merge(U, V, Z, S).
gf2_poly_intern_case(X + Y, Z, A-B, S):-!,
	gf2_intern_poly(X, U, A-C, S),
	gf2_intern_poly(Y, V, C-B, S),
	gf2_join(U, V, Z, S).
gf2_poly_intern_case(+(V), Y, D, S):-!, gf2_poly_list(+, V, 0, Y, D, S).
gf2_poly_intern_case(*(V), Y, D, S):-!, gf2_poly_list(*, V, 1, Y, D, S).
gf2_poly_intern_case(X, Y, [X|A]-A, S)	:- gf2_atom_count(X, S),
	zdd_singleton(X, Y, S).
%
gf2_poly_list(_, [], Y, Y, U-U, _):-!.
gf2_poly_list(F, [X|Xs], Y0, Y, U-V, S):-
	gf2_intern_poly(X, A, U-W, S),
	(	F = (*) -> gf2_merge(A, Y0, Y1, S)
	;	gf2_join(A, Y0, Y1, S)
	),
	gf2_poly_list(F, Xs, Y1, Y, W-V, S).

%!	gf2_sat_count_poly(+Poly, -C, +S) is det.
%	Equivalent to gf2_sat_count_poly_list([Poly], C, S).

% ?- #S, Poly = a+b+c, gf2_sat_count_poly(Poly, C, S).
% ?- #S, Poly = 1+(1+a)*(1+b)*(1+c), gf2_sat_count_poly(Poly, C, S).
% ?- #S, P = [1+a, 1+b, 1+c],
%	gf2_intern_poly_list(P, X, N, S),
%	elim_variables(X, Y, K, S),
%	gf2_merge_list(Y, 1, Z, S),
%	gf2_join(1, Z, U, S),
%	gf2_count_solutions(3, U, C, S).

% ?- #S, P = [a, b, c],
%	gf2_intern_poly_list(P, X, N, S),
%	elim_variables(X, Y, K, S),
%	apply_list_with_state(gf2_join(1), Y, Z, S),
%	gf2_merge_list(Z, 1, U, S),
%	gf2_join(1, U, V, S),
%	J is N - K,
%	gf2_count_solutions(J, V, C, S).

%!	gf2_sat_count_poly_list(Ps, C, S) is det.

%!	gf2_sat_count_poly_list(+Ps, -C, +S) is det.
%	C is unified with the number of solutions in GF2 of
%	conjunctions
%		P1 = 0,
%		P2 = 0,
%		...
%		Pn = 0
%	where Ps = [P1, P2, ..., Pn].

% ?- #S, gf2_sat_count_poly_list([a, b, c], C, S).
% ?- #S, Ps = [a],			gf2_sat_count_poly_list(Ps, C, S). %@ C = 1.
% ?- #S, Ps = [a*a, b+b],	gf2_sat_count_poly_list(Ps, C, S). %@ C = 2.
% ?- #S, Ps = [a*b],		gf2_sat_count_poly_list(Ps, C, S). %@ C = 3.
% ?- #S, Ps = [*[a,b]],		gf2_sat_count_poly_list(Ps, C, S). %@ C = 3.
% ?- #S, Ps = [a*b+c],		gf2_sat_count_poly_list(Ps, C, S). %@ C = 4.
% ?- #S, Ps = [a*a+a],		gf2_sat_count_poly_list(Ps, C, S). %@ C = 2.
% ?- #S, Ps = [a*b + b+c],	gf2_sat_count_poly_list(Ps, C, S). %@ C = 4.
% ?- #S, Ps = [b*b + b+c],	gf2_sat_count_poly_list(Ps, C, S). %@ C = 2.
% ?- #S, Ps = [a + b + c + d], gf2_sat_count_poly_list(Ps, C, S). %@ C = 8.
% ?- #S, Ps = [a * b  * c * d], gf2_sat_count_poly_list(Ps, C, S). %@ C = 15.
% ?- #S, Ps = [a + b], gf2_sat_count_poly_list(Ps, C, S). %@ C = 2.
% ?- #S, Ps = [a + b], gf2_sat_count_poly_list(Ps, C, S). %@ C = 2.
% ?- #S, Ps = [a*(1+a)+b+c+d], gf2_sat_count_poly_list(Ps, C, S). %@ C = 8.
% ?- #S, Ps = [a*(1+a)], gf2_sat_count_poly_list(Ps, C, S). %@ C = 2.
% ?- #S, Ps = [1 + a*(1+a) * b*(1+b) * c*(1+c)],
%	gf2_sat_count_poly_list(Ps, C, S). %@ C = 0.
% ?- #S, Ps = [1 + a*(1+a) + b*(1+b) + c*(1+c)],
%	gf2_sat_count_poly_list(Ps, C, S). %@ C = 0.
% ?- #S, Ps = [a+b+1 + c+d+1 + a*c+b*d],
%	gf2_sat_count_poly_list(Ps, C, S). %@ C = 10.
% ?- #S, Ps = [1+1+1+1+1+1], gf2_sat_count_poly_list(Ps, C, S). %@ C = 1.
% ?- #S, Ps = [1+1+1+1+1+1+1], gf2_sat_count_poly_list(Ps, C, S). %@ C = 0.

gf2_sat_count_poly_list(Ps, C, S):-
	gf2_intern_poly_list(Ps, X, N, S),
	elim_variables(X, Y, K, S),
	apply_list_with_state(gf2_join(1), Y, Z, S),
	gf2_merge_list(Z, 1, U, S),
	gf2_join(1, U, V, S),
	writeln(enter_zdd_slim(V)),
	zdd_slim(V, W, S),
	writeln(exit_zdd_slim(V)),
	J is N - K,
	writeln(enter_counting(J, W)),
	gf2_count_solutions(J, W, C, S).
%
apply_list_with_state(F, X, Y, S):-apply_list_flip(X, Y, S, F).
%
apply_list_flip([], [], _, _).
apply_list_flip([X|Xs], [Y|Ys], S, F):-
	call(F, X, Y, S),
	apply_list_flip(Xs, Ys, S, F).

%!	gf2_sat_count_poly(+Poly, -C, S) is det.
%	Poly : polynomial
%	C is unified with the number of solutions for Poly = 0.

% ?- #S, Poly = 1+1+1+1+1+1+1, gf2_sat_count_poly(Poly, C, S).
% ?- #S, Poly = 1+1+1+1+1+1, gf2_sat_count_poly(Poly, C, S).
gf2_sat_count_poly(Poly, C, S):- gf2_sat_count_poly_list([Poly], C, S).

% ?- remove_elem(a, [b,a,a], X).
remove_elem(_, [], []):-!.
remove_elem(A, [A|Ps], Qs):-!, remove_elem(A, Ps, Qs).
remove_elem(A, [B|Ps], [B|Qs]):-!, remove_elem(A, Ps, Qs).


%!	psa_orth(+N, -NV, -K, +S) is det.
%	NV is unified with the number of elements of orthogonal(GF(2), N).
%	K is unified with the number of remaining elements after variable
%	eliminatiton is performed.

% ?- #S, psa_orth(2, NV, K, S).
% ?- #S, psa_orth(3, NV, K, S).

psa_orth(N, NV, K, S):-
	orth_matrix_cond(N, Ps),
	gf2_intern_poly_list(Ps, Qs, NV, S),
	forall(member(M, Qs), psa(M, S)),
	elim_variables(Qs, _Rs, K, S).

%!	gf2_intern_poly_list(+Ps, -Qs, -N, +S) is det.
%	Ps: list of polynomials over GF2.
%   Qs is unified with a list of zdds such that each member of
%	Ps is read in order as a member of Qs.
%	N is unified with the number of atoms which appears in Ps.

% ?- #S, Ps = [*[a,b], +[b,c]], gf2_intern_poly_list(Ps, Qs, N, S),
% forall(member(M, Qs),  psa(M, S)).
% ?- #S, Ps = [a*b, b+c], gf2_intern_poly_list(Ps, Qs, N, S),
% forall(member(M, Qs),  psa(M, S)).
% ?- #S, Ps = [a*b, b+c], gf2_intern_poly_list(Ps, Qs, N, S),
%	elim_variables(Qs, Rs, K, S),
%	forall(member(M, Qs),  psa(M, S)),
%	forall(member(M, Rs),  psa(M, S)).

gf2_intern_poly_list(Ps, Qs, N, S):-
	set_key(gf2_atom_count, 0, S),
	gf2_intern_poly_list(Ps, Qs, S),
	get_key(gf2_atom_count, N, S).
%
gf2_intern_poly_list([],[], _).
gf2_intern_poly_list([X|Ps], [Y|Qs], S):-
	gf2_intern_poly(X, Y, S),
	gf2_intern_poly_list(Ps, Qs, S).
%
gf2_atom_count(X, S):- memo(gf2_atom(X)-B, S),
	(	var(B) ->
		(	get_key(gf2_atom_count, K, S) -> true
		;	K = 0
		),
		K0 is K + 1,
		set_key(gf2_atom_count, K0, S),
		B = true
	;	true
	).

		/***********************************************
		*   Operations on polynomial zdd over GF(2)    *
		***********************************************/

%!	gf2_insert(+X, +Y, -Z, +S) is det.
%	Insert a literal X into each set of literals
%	in a ZDD Y, and unify Z with the result.
%   For short, Z = X*Y.

gf2_insert(X, Y, Z):- @(gf2_insert(X, Y, Z)).
%
gf2_insert(_, 0, 0, _):-!.
gf2_insert(A, 1, J, S):-!, zdd_singleton(A, J, S).
gf2_insert(A, I, J, S):- memo(gf2_insert(I, A)-J, S),
	(	nonvar(J)	-> true
	;	cofact(I, t(B, L, R), S),
		compare(C, A, B),
		( 	C = (=)	 ->  gf2_insert(A, L, L0, S),
						 zdd_cons(R0, A, R, S),
						 gf2_join(L0, R0, J, S)
		;	C = (<)  ->  zdd_cons(J, A, I, S)
		;   C = (>)  ->  gf2_insert(A, R, R0, S),
						 gf2_insert(A, L, L0, S),
						 cofact(J, t(B, L0, R0), S)
		)
	).

%!	gf2_join(+X, +Y, -Z) is det.
%	Unify Z with GF(2)-polynomial addition X +Y.
%	Z = X + Y.

% ?- zdd((X<< (set([a])), gf2_join(X, X, Z), psa(Z))).
% ?- zdd((X<< (set([1])), gf2_join(X, X, Z), psa(Z))).
% ?- zdd((X<< (set([1,a])), gf2_join(X, X, Z), psa(Z))).
% ?- zdd((X<< (set([b])+set([a])), galois_field:gf2_join(X, X, Z), psa(Z))).
% ?- zdd((X<< (1 + set([b])+set([a])), zdd_merge(X, X, Z), psa(Z))).

%!	gf2_join(+X, +Y, -Z, +S) is det.
%	X, Y: zdd of GF(2)-polyminals.
%	Z is unified with GF(2)-polynomial addition X + Y.
gf2_join(X, Y, Z):- @(gf2_join(X, Y, Z)).
%
gf2_join(X, X, 0, _):-!.
gf2_join(0, X, X, _):-!.
gf2_join(X, 0, X, _):-!.
gf2_join(1, X, Y, S):-!, gf2_join_1(X, Y, S).
gf2_join(X, 1, Y, S):-!, gf2_join_1(X, Y, S).
gf2_join(X, Y, Z, S):-
	(	X@<Y -> memo(gf2_join(X,Y)-Z, S)
	;	memo(gf2_join(Y, X)-Z, S)
	),
	(	nonvar(Z) -> true %, write(.)
	; 	cofact(X, t(A, L, R), S),
		cofact(Y, t(A0, L0, R0), S),
		compare(C, A, A0),
		(	C = (<) -> 	gf2_join(L, Y, U, S),
						cofact(Z, t(A, U, R), S)
		;	C = (=)	->	gf2_join(R, R0, U, S),
						gf2_join(L, L0, V, S),
						cofact(Z, t(A, V, U), S)
		; 	gf2_join(L0, X, U, S),
			cofact(Z, t(A0, U, R0), S)
		)
	).

% Switch the left-most leaf (0 or 1).
gf2_join_1(0, 1, _):-!.
gf2_join_1(1, 0, _):-!.
gf2_join_1(X, Y, S):-!, cofact(X, t(A, L, R), S),
	gf2_join_1(L, L0, S),
	cofact(Y, t(A, L0, R), S).

%!	gf2_merge(+X, +Y, -Z, +S) is det.
%	X, Y: GF(2)-polyniomials in zdd.
%	Z is unified with GF(2)-polyniomial addition X * Y.
%	Z = X*Y.

% ?- #S, X<< *[a,b], Y<< *[c,d], gf2_merge(X, Y, Z, S), psa(Z, S).
% ?- #S, X<< *[a,b], Y<< *[b,d], gf2_merge(X, Y, Z, S), psa(Z, S).

% This is faster. Strange.
gf2_merge(X, Y, Z):- shift(gf2_merge(X, Y, Z)).
%
gf2_merge(0, _, 0, _):-!.
gf2_merge(_, 0, 0, _):-!.
gf2_merge(1, X, X, _):-!.
gf2_merge(X, 1, X, _):-!.
gf2_merge(X, X, X, _):-!.
gf2_merge(X, Y, Z, S):-
	(	X@<Y -> memo(gf2_merge(X,Y)-Z, S)
	;	memo(gf2_merge(Y,X)-Z, S)
	),
	(	nonvar(Z) -> true
	;	cofact(X, t(A, L, R), S),
		gf2_merge(L, Y, L0, S),
		gf2_merge(R, Y, R1, S),
		gf2_insert(A, R1, R0, S),
		gf2_join(L0, R0, Z, S)
	).
%
gf2_merge_list([], X, X, _).
gf2_merge_list([A|As], X, Y, S):-
	zdd_slim([X, A|As], Slim, S),
	Slim=[X0, A0|As0],
	gf2_merge(A0, X0, X1, S),
	length([A|As], L),
	writeln(remain(L)),
	writeln(gf2_merge(A0, X0, X1)),
	gf2_merge_list(As0, X1, Y, S).

% ?- zdd((U<<{[a]}, P<<pow([x,y]), @gf2_subst(x, U, P, Q), psa(Q))).
% ?- zdd((U<<{[a], [b]}, P<<pow([x,y]), @gf2_subst(x, U, P, Q), psa(Q))).
% ?- zdd((U<<{[a, b]}, P<<pow([a, b, c, x,y]), @gf2_subst(x, U, P, Q), psa(Q))).

gf2_subst(X, U, P, Q):- @(gf2_subst(X, U, P, Q)).
%
gf2_subst(_, _, X, X, _):- X<2, !.
gf2_subst(X, U, P, Q, S):- memo(gf2_subst(X, U, P)-Q, S),
	(	nonvar(Q) -> true	%, write(.) % So so hits.
	;	cofact(P, t(Y, L, R), S),
		gf2_subst(X, U, L, L0, S),
		compare(C, X, Y),
		(	C = (=) 	-> 	gf2_merge(U, R, R0, S)
		;	C = (<)		-> 	zdd_cons(R0, Y, R, S)
		;	gf2_subst(X, U, R, R1, S),
			gf2_insert(Y, R1, R0, S)
		),
		gf2_join(L0, R0, Q, S)
	).

%!	select_disjunctive_atom(+X, ?A, -Y, S) is det.
%	True if A is a disjunctive atom of X,  and Y is unified with
%	Y0 such that Y = A + Y0. Otherwise, fails.

%	Note that in GF2, u + A = 0  if and if u = A.
%	This property is used to eliminate variables in a system of equations.

% ?- #S, X<< +[*[a,b],c], select_disjunctive_atom(X, A, Y, S), psa(Y, S).
% ?- #S, X<< +[c, *[a,b]], select_disjunctive_atom(X, A, Y, S), psa(Y, S).
% ?- #S, X<< +[*[x,y], u, *[a,b]], select_disjunctive_atom(X, A, Y, S), psa(Y, S).
% ?- #S, X<< +[*[x,y], *[a,b]], select_disjunctive_atom(X, A, Y, S), psa(Y, S). % <== false

select_disjunctive_atom(X, A, L, S):- X > 1, cofact(X, t(A, L, 1), S).

%!	elim_variables(+X, -Y, -N, +S) is det.
%	X: list of polynomials
%	Y is unified with a list of polinomials
%	N is unified with a number of eliminated varialbles.

% ?- #S, elim_variables([], Y, K, S).
% ?- #S, X<< +[a, b], L = [X], psa(X, S), elim_variables([], Y, K, S).
% ?- #S, X<< +[a, b], L = [X], psa(X, S), elim_variables(L, Y, K, S).
% ?- #S, X<< +[a, b], Y<< +[c,d], Z << +[*[ x,y]], L = [X, Y, Z],
%  elim_variables(L, L0, K, S), forall(member(M,L0), psa(M,S)).
% ?- #S, X<< +[a, b], Y<< +[c,d], Z << *[ a, c], L = [X, Y, Z],
%  elim_variables(L, L0, K, S), forall(member(M,L0), psa(M,S)).
% ?- #S, X<< +[a, b], Y<< +[c,d], Z << +[*[ a, c],*[b,d]], L = [X, Y, Z],
%  elim_variables(L, L0, K, S), forall(member(M,L0), psa(M,S)).
% ?- #S, X<< +[a, b], Y<< *[a, b, c,d], Z << +[*[ x,y]], L = [X, Y, Z],
%  elim_variables(L, L0, K, S), forall(member(M,L0), psa(M,S)).
% ?- #S, X<< +[a, b], Y<< +[c,d], Z << +[*[a, x, c, y]], L = [X, Y, Z], psa(X, S), psa(Y, S), elim_variables(L, L0, S),
%	L0 = [W|_], psa(W,S).
% ?- #S, X<< +[a], Y<< +[*[c,d]], Z << +[*[a, x, c, y]],
%	L = [X, Y, Z], elim_variables(L, L0, S),
%	forall(member(M, L0), psa(M,S)).
% ?- #S, X<< +[a], Z << +[*[a, x, c, y]],
%	L = [X, Z], elim_variables(L, L0, S),
%	forall(member(M, L0), psa(M,S)).
% ?- #S, X<< +[a],	Y<< +[a, b], L = [X, Y], elim_variables(L, L0, N, S),
%	forall(member(M, L0), psa(M,S)).

elim_variables(X, Y, N, S):- select(U, X, X0),
	select_disjunctive_atom(U, A, E, S),
	!,
	maplist(gf2_subst_flip(S, A, E), X0, X1),
	elim_variables(X1, Y, N0, S),
	N is N0 + 1.
elim_variables(X, X, 0, _).
%
gf2_subst_flip(S, A, E, X, Y):- gf2_subst(A, E, X, Y, S).

%!	gf2_append(+X, +Y, -Z, +S) is det.
%	X : list of atoms. [A1, A2,..., An]
%	Y : GF(2)-polynomial zdd.
%	Z is unified with GF(2)-polinomial A1 * A2 * ... * An * Y.

% ?- zdd((@gf2_append([b, a, d, c], 1, X), psa(X))).
% ?- zdd((@gf2_append([b, a, d, a, c], 1, X), psa(X))).
% ?- zdd((@gf2_append([1, a, d, a, c], 1, X), psa(X))).
% ?- zdd((@gf2_intern_poly(a*b, A),gf2_append([1, d, c], A, X), psa(X))).
% ?- zdd((@gf2_intern_poly(a*b+a*c, A), gf2_append([ d, c], A, X), psa(X))).
% ?- zdd((@gf2_intern_poly(a*b+1, A), gf2_append([ d, c], A, X), psa(X))).
% ?- zdd((@gf2_intern_poly(a+1, A),gf2_append([ d, c], A, X), psa(X))).
% ?- zdd((@gf2_intern_poly(a+b+a, A),gf2_append([ d, c], A, X), psa(X))).

gf2_append(X, Y, Z):- @gf2_append(X, Y, Z).
%
gf2_append([], X, X, _).
gf2_append([A|As], X, Y, S):- gf2_append(As, X, X0, S),
	gf2_insert(A, X0, Y, S).

		/***********************************************
		*     Counting Orthogonal group O(n, GF(2))    *
		***********************************************/

%!	gf2_sat_count_orth(+N, -C, +S) is det.
%	N>=1.
%	C is unified with the number of all N x N orthogonal matricies
%	over GF2.

% ?- #S, gf2_sat_count_orth(1, C, S).
%@ C = 1.
% ?- #S, gf2_sat_count_orth(2, C, S).
%@ C = 2.
% ?- #S, gf2_sat_count_orth(3, C, S).
%@ C = 6.
% ?- #S, gf2_sat_count_orth(4, C, S).
%@ C = 48.
% ?- #S, gf2_sat_count_orth(5, C, S).
%@ C = 720.
% ?- #S, time(gf2_sat_count_orth(6, C, S)).
%@ % 21,529,533,838 inferences, 2558.382 CPU in 2606.041 seconds (98% CPU, 8415293 Lips)
%@ S = ..,
%@ C = 23040.

gf2_sat_count_orth(N, C, S):-
	orth_matrix_cond(N, PolyList),
	gf2_sat_count_poly_list(PolyList, C, S).

%!	orth_matrix_cond(+N, -List) is det.
%	N >= 1.
%	List is unified with a list of polynomials
%	such that N x N matrix M over GF2 is an orthogonal one
%	if and only if elements of M satifies
%	equations E = 0 for all E in the List.

% ?- orth_matrix_cond(2, X), maplist(writeln, X).
orth_matrix_cond(N, ListOfEq0):- findall( CJK,
	   (	between(1, N, J),
			between(1, N, K),
			K =< J,
			findall(a(J, I) * a(K, I),
				   between(1, N, I),
				   CJK0),
			(	J=K -> CJK = 1 + +(CJK0)
			;	CJK = +(CJK0)
			)
	   ),
	   ListOfEq0).
%
orth_matrix_cond_one_poly(N, 1 + *(ListOfEq0)):- findall( CJK,
	   (	between(1, N, J),
			between(1, N, K),
			K =< J,
			findall(a(J, I) * a(K, I),
				   between(1, N, I),
				   CJK0),
			(	J=K -> CJK = +(CJK0)
			;	CJK = 1+(CJK0)
			)
	   ),
	   ListOfEq0).


		/******************
		*   For debug.    *
		******************/

%!	sharp(+G) is det.
%	Print G after call G, applying psa for integer argument
%	as a zdd.

sharp(X):- call(X),
	writeln("============"),
	get_sharp(S),
	print(X),
	nl,
	X =.. [_|As],
	forall(member(M, As),
		   (	integer(M) -> psa(M, S)
		   ;	true
		   )).

%!	gf2_solve(+F, -A, ?V) is nondet.
%	A naive solver for a polynomial over GF2.
%	F: polynomial.
%	A is unified with an assighment of variables of F.
%	V is unified with the value of F for the assighment A.

% ?- gf2_solve(1 + 3+ 5 + 5*5, X, V).
% ?- collect_atoms(1 + a+ 5 + a*5, F, X),	gf2_solve(F, X, V), V = 0.
% ?- collect_atoms(a + b, F, X), gf2_solve(F, X, V), V = 0.
% ?- P= 1+ ( (a+ b)*(c + d)*(1+a*c+b*d)),
%	collect_atoms(P, F, X), gf2_solve(F, X, V), V = 0.
% ?- P= a+b,
%	collect_atoms(P, F, X), gf2_solve(F, X, V), V = 0.

gf2_solve(F, Asgn, V):- gf2_assignment(Asgn), eval_gf2(F, V).
%
gf2_assignment([]).
gf2_assignment([_-V|R]):- gf2(V),
	gf2_assignment(R).

%!	collect_atoms(+E, -F, -A) is det.
%	E: polynomiaal.
%	F is unified with a polynomial,
%	A is unified with an assoc list such that each variable of E
%	is associated with a new variable, and that F is subsitution of E by A.

% ?- collect_atoms(a, F, A).
%@ A = [a-F].
% ?- collect_atoms(a+b, F, A).
%@ F = _A+_B,
%@ A = [a-_A, b-_B].

collect_atoms(E, F, A):- collect_atoms(E, F, B, []),
	msort(B, C),
	unique_atoms(C, A).
%
collect_atoms(+(X, Y), +(U, V), A, B):-!,
	collect_atoms(X, U, A, C),
	collect_atoms(Y, V, C, B).
collect_atoms(*(X, Y), *(U, V), A, B):-!,
	collect_atoms(X, U, A, C),
	collect_atoms(Y, V, C, B).
collect_atoms(X, X,  A, A):- integer(X), !.
collect_atoms(X, U, [X-U|A], A).

%
unique_atoms([], []):-!.
unique_atoms([X], [X]):-!.
unique_atoms([A-U, A-U|R],  R0):-!,	unique_atoms([A-U|R], R0).
unique_atoms([P|R],  [P|R0]):- unique_atoms(R, R0).

%!	eval_gf2(+P, -V) is det.
%	V is the value of P.
eval_gf2(J, V):- integer(J), !, V is J mod 2.
eval_gf2(+(X, Y), V):-!, eval_gf2(X, X0),
					  eval_gf2(Y, Y0),
					  gf2_table(+, X0, Y0, V), !.
eval_gf2(*(X, Y), V):-!, eval_gf2(X, X0),
					  eval_gf2(Y, Y0),
					  gf2_table(*, X0, Y0, V), !.

%
gf2(0).
gf2(1).

%
gf2_table(+, 0, 0, 0).
gf2_table(+, 0, 1, 1).
gf2_table(+, 1, 0, 1).
gf2_table(+, 1, 1, 0).
gf2_table(*, 1, 1, 1).
gf2_table(*, _, _, 0).

%
gf2_add(X, X, 0):-!.
gf2_add(_, _, 1).

%!	verify_poly(+U, +X, +Sols, +S) is det.
%	U : 0, or 1
%	X : zdd as polynomial
%	Sols: zdd as solutions (assignmets)
%	True if forall assignments in Sols the value of X is U.

verify_poly(_, _, 0, _):-!.
verify_poly(U, X, A, S):-
	forall(set_at(Vs, A, S),
				(	eval_poly(X, Vs, U, S) -> true
				;	writeln(fail_verify_zero_at(X, A)),
					psa(X, S),
					psa(A, S)
				)).


%!	eval_poly(+Y, +Asgn, -V, +S) is det.
%	X: a zdd of polyonimial.
%	Asgn: list of atoms.
%	With X as a polynomial over GF2 and  Asgn as an assignment,
%	V is unified with the value of X w.r.t Asgn, where
%	each member of Asgn interpreted is 1, and 0 for non-members.

eval_poly(X, _, X, _):- X<2, !.
eval_poly(X, Vs, V, S):-
	cofact(X, t(A, L, R), S),
	eval_poly(L, Vs, L0, S),
	eval_poly_right(A, R, Vs, R0, S),
	gf2_add(L0, R0, V).
%
eval_poly_right(A, X, Vs, V, S):-
	(	memberchk(A, Vs) -> eval_poly(X, Vs, V, S)
	;	V = 0
	).

%!	subst_poly(+L, +A, -L0) is det.
%	L : term
%	A : list
%	L0 is unified with a term such that L and L0 differs only
%	at leaves in such a way that each leaf U of L is subtituted with 1
%	if U is member of Asgn, otherwise with 0.

% ?- subst_poly([a,b], [b], X).
% ?- subst_poly(+(a,b), [b], X).
% ?- subst_poly(*[a+b], [b], X).

subst_poly(0, _, 0):-!.
subst_poly(1, _, 1):-!.
subst_poly([], _, []):-!.
subst_poly([A|As], Asgn, [A0|As0]):-!,
	subst_poly(A, Asgn, A0),
	subst_poly(As, Asgn, As0).
subst_poly(T, Asgn, T0):- functor(T, F, _),
	memberchk(F, [*,+]),
	!,
	T =..[F|Args],
	subst_poly(Args, Asgn, Args0),
	T0 =.. [F|Args0].
subst_poly(X, Asgn, Y):-
	(	memberchk(X, Asgn) -> Y = 1
	;	Y = 0
	).
